/*
 * $RCSfile: TIFFYCbCrDecompressor.java,v $
 *
 * 
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met: 
 * 
 * - Redistribution of source code must retain the above copyright 
 *   notice, this  list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in 
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of 
 * contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any 
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND 
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MIDROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL 
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF 
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR 
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES. 
 * 
 * You acknowledge that this software is not designed or intended for 
 * use in the design, construction, operation or maintenance of any 
 * nuclear facility. 
 *
 * $Revision: 1.3 $
 * $Date: 2006/06/23 19:48:28 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.io.ByteArrayInputStream;
import java.io.EOFException;
import java.io.IOException;
import javax.imageio.ImageReader;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.MemoryCacheImageInputStream;
import javax.imageio.stream.ImageInputStream;
import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFDecompressor;
import com.sun.media.imageio.plugins.tiff.TIFFField;

public final class TIFFYCbCrDecompressor extends TIFFDecompressor {
	// Store constants in S15.16 format 
	private static final int FRAC_BITS = 16;
	private static final float FRAC_SCALE = 1 << FRAC_BITS;

	private float _LumaRed = 0.299f;
	private float _LumaGreen = 0.587f;
	private float _LumaBlue = 0.114f;

	private float _referenceBlackY;
	private float _referenceWhiteY = 255.0f;

	private float _referenceBlackCb = 128.0f;
	private float _referenceWhiteCb = 255.0f;

	private float _referenceBlackCr = 128.0f;
	private float _referenceWhiteCr = 255.0f;

	private static final float CONDING_RANGE_Y = 255.0f;

	private int[] _iYTab = new int[256];
	private int[] _iCbTab = new int[256];
	private int[] _iCrTab = new int[256];

	private int[] _iGYTab = new int[256];
	private int[] _iGCbTab = new int[256];
	private int[] _iGCrTab = new int[256];

	private int _chromaSubsampleH = 2;
	private int _chromaSubsampleV = 2;

	private boolean _colorConvert;

	private final TIFFDecompressor _decompressor;

	private BufferedImage _tmpImage;

	//
	// If 'decompressor' is not null then it reads the data from the
	// actual stream first and passes the result on to YCrCr decompression
	// and possibly color conversion.
	//

	public TIFFYCbCrDecompressor(TIFFDecompressor decompressor, boolean colorConvert) {
		this._decompressor = decompressor;
		this._colorConvert = colorConvert;
	}

	private void warning(String message) {
		if (this.reader instanceof TIFFImageReader) {
			((TIFFImageReader)reader).forwardWarningMessage(message);
		}
	}

	//
	// "Chained" decompressor methods.
	//

	@Override
	public void setReader(ImageReader reader) {
		if (_decompressor != null) {
			_decompressor.setReader(reader);
		}
		super.setReader(reader);
	}

	@Override
	public void setMetadata(IIOMetadata metadata) {
		if (_decompressor != null) {
			_decompressor.setMetadata(metadata);
		}
		super.setMetadata(metadata);
	}

	@Override
	public void setPhotometricInterpretation(int photometricInterpretation) {
		if (_decompressor != null) {
			_decompressor.setPhotometricInterpretation(photometricInterpretation);
		}
		super.setPhotometricInterpretation(photometricInterpretation);
	}

	@Override
	public void setCompression(int compression) {
		if (_decompressor != null) {
			_decompressor.setCompression(compression);
		}
		super.setCompression(compression);
	}

	@Override
	public void setPlanar(boolean planar) {
		if (_decompressor != null) {
			_decompressor.setPlanar(planar);
		}
		super.setPlanar(planar);
	}

	@Override
	public void setSamplesPerPixel(int samplesPerPixel) {
		if (_decompressor != null) {
			_decompressor.setSamplesPerPixel(samplesPerPixel);
		}
		super.setSamplesPerPixel(samplesPerPixel);
	}

	@Override
	public void setBitsPerSample(int[] bitsPerSample) {
		if (_decompressor != null) {
			_decompressor.setBitsPerSample(bitsPerSample);
		}
		super.setBitsPerSample(bitsPerSample);
	}

	@Override
	public void setSampleFormat(int[] sampleFormat) {
		if (_decompressor != null) {
			_decompressor.setSampleFormat(sampleFormat);
		}
		super.setSampleFormat(sampleFormat);
	}

	@Override
	public void setExtraSamples(int[] extraSamples) {
		if (_decompressor != null) {
			_decompressor.setExtraSamples(extraSamples);
		}
		super.setExtraSamples(extraSamples);
	}

	@Override
	public void setColorMap(char[] colorMap) {
		if (_decompressor != null) {
			_decompressor.setColorMap(colorMap);
		}
		super.setColorMap(colorMap);
	}

	@Override
	public void setStream(ImageInputStream stream) {
		if (_decompressor != null) {
			_decompressor.setStream(stream);
		} else {
			super.setStream(stream);
		}
	}

	@Override
	public void setOffset(long offset) {
		if (_decompressor != null) {
			_decompressor.setOffset(offset);
		}
		super.setOffset(offset);
	}

	@Override
	public void setByteCount(int byteCount) {
		if (_decompressor != null) {
			_decompressor.setByteCount(byteCount);
		}
		super.setByteCount(byteCount);
	}

	@Override
	public void setSrcMinX(int srcMinX) {
		if (_decompressor != null) {
			_decompressor.setSrcMinX(srcMinX);
		}
		super.setSrcMinX(srcMinX);
	}

	@Override
	public void setSrcMinY(int srcMinY) {
		if (_decompressor != null) {
			_decompressor.setSrcMinY(srcMinY);
		}
		super.setSrcMinY(srcMinY);
	}

	@Override
	public void setSrcWidth(int srcWidth) {
		if (_decompressor != null) {
			_decompressor.setSrcWidth(srcWidth);
		}
		super.setSrcWidth(srcWidth);
	}

	@Override
	public void setSrcHeight(int srcHeight) {
		if (_decompressor != null) {
			_decompressor.setSrcHeight(srcHeight);
		}
		super.setSrcHeight(srcHeight);
	}

	@Override
	public void setSourceXOffset(int sourceXOffset) {
		if (_decompressor != null) {
			_decompressor.setSourceXOffset(sourceXOffset);
		}
		super.setSourceXOffset(sourceXOffset);
	}

	@Override
	public void setDstXOffset(int dstXOffset) {
		if (_decompressor != null) {
			_decompressor.setDstXOffset(dstXOffset);
		}
		super.setDstXOffset(dstXOffset);
	}

	@Override
	public void setSourceYOffset(int sourceYOffset) {
		if (_decompressor != null) {
			_decompressor.setSourceYOffset(sourceYOffset);
		}
		super.setSourceYOffset(sourceYOffset);
	}

	@Override
	public void setDstYOffset(int dstYOffset) {
		if (_decompressor != null) {
			_decompressor.setDstYOffset(dstYOffset);
		}
		super.setDstYOffset(dstYOffset);
	}

	/* Should not need to override these mutators as subsampling
       should not be done by the wrapped decompressor.
    public void setSubsampleX(int subsampleX) {
        if (decompressor != null) {
            decompressor.setSubsampleX(subsampleX);
        }
        super.setSubsampleX(subsampleX);
    }

    public void setSubsampleY(int subsampleY) {
        if (decompressor != null) {
            decompressor.setSubsampleY(subsampleY);
        }
        super.setSubsampleY(subsampleY);
    }
	 */

	@Override
	public void setSourceBands(int[] sourceBands) {
		if (_decompressor != null) {
			_decompressor.setSourceBands(sourceBands);
		}
		super.setSourceBands(sourceBands);
	}

	@Override
	public void setDestinationBands(int[] destinationBands) {
		if (_decompressor != null) {
			_decompressor.setDestinationBands(destinationBands);
		}
		super.setDestinationBands(destinationBands);
	}

	@Override
	public void setImage(BufferedImage image) {
		if (_decompressor != null) {
			ColorModel cm = image.getColorModel();
			_tmpImage =
					new BufferedImage(cm,
							image.getRaster().createCompatibleWritableRaster(1, 1),
							cm.isAlphaPremultiplied(),
							null);
			_decompressor.setImage(_tmpImage);
		}
		super.setImage(image);
	}

	@Override
	public void setDstMinX(int dstMinX) {
		if (_decompressor != null) {
			_decompressor.setDstMinX(dstMinX);
		}
		super.setDstMinX(dstMinX);
	}

	@Override
	public void setDstMinY(int dstMinY) {
		if (_decompressor != null) {
			_decompressor.setDstMinY(dstMinY);
		}
		super.setDstMinY(dstMinY);
	}

	@Override
	public void setDstWidth(int dstWidth) {
		if (_decompressor != null) {
			_decompressor.setDstWidth(dstWidth);
		}
		super.setDstWidth(dstWidth);
	}

	@Override
	public void setDstHeight(int dstHeight) {
		if (_decompressor != null) {
			_decompressor.setDstHeight(dstHeight);
		}
		super.setDstHeight(dstHeight);
	}

	@Override
	public void setActiveSrcMinX(int activeSrcMinX) {
		if (_decompressor != null) {
			_decompressor.setActiveSrcMinX(activeSrcMinX);
		}
		super.setActiveSrcMinX(activeSrcMinX);
	}

	@Override
	public void setActiveSrcMinY(int activeSrcMinY) {
		if (_decompressor != null) {
			_decompressor.setActiveSrcMinY(activeSrcMinY);
		}
		super.setActiveSrcMinY(activeSrcMinY);
	}

	@Override
	public void setActiveSrcWidth(int activeSrcWidth) {
		if (_decompressor != null) {
			_decompressor.setActiveSrcWidth(activeSrcWidth);
		}
		super.setActiveSrcWidth(activeSrcWidth);
	}

	@Override
	public void setActiveSrcHeight(int activeSrcHeight) {
		if (_decompressor != null) {
			_decompressor.setActiveSrcHeight(activeSrcHeight);
		}
		super.setActiveSrcHeight(activeSrcHeight);
	}

	private byte clamp(int floatNumber) {
		if (floatNumber < 0) {
			return (byte)0;
		} else if (floatNumber > 255 * 65536) {
			return (byte)255;
		} else {
			return (byte)(floatNumber >> 16);
		}
	}

	@Override
	public void beginDecoding() {
		if (_decompressor != null) {
			_decompressor.beginDecoding();
		}

		TIFFImageMetadata tmetadata = (TIFFImageMetadata)metadata;
		TIFFField field;

		field = tmetadata.getTIFFField(BaselineTIFFTagSet.TAG_Y_CB_CR_SUBSAMPLING);
		if (field != null) {
			if (field.getCount() == 2) {
				this._chromaSubsampleH = field.getAsInt(0);
				this._chromaSubsampleV = field.getAsInt(1);

				if (_chromaSubsampleH != 1 && _chromaSubsampleH != 2 
						&& _chromaSubsampleH != 4) {
					warning("Y_CB_CR_SUBSAMPLING[0] has illegal value "
							+ _chromaSubsampleH + " (should be 1, 2, or 4), setting to 1");
					_chromaSubsampleH = 1;
				}

				if (_chromaSubsampleV != 1 && _chromaSubsampleV != 2 && _chromaSubsampleV != 4) {
					warning("Y_CB_CR_SUBSAMPLING[1] has illegal value "
							+ _chromaSubsampleV
							+ " (should be 1, 2, or 4), setting to 1");
					_chromaSubsampleV = 1;
				}
			} else {
				warning("Y_CB_CR_SUBSAMPLING count != 2, "
						+ "assuming no subsampling");
			}
		}

		field = tmetadata.getTIFFField(BaselineTIFFTagSet.TAG_Y_CB_CR_COEFFICIENTS);
		if (field != null) {
			if (field.getCount() == 3) {
				this._LumaRed = field.getAsFloat(0);
				this._LumaGreen = field.getAsFloat(1);
				this._LumaBlue = field.getAsFloat(2);
			} else {
				warning("Y_CB_CR_COEFFICIENTS count != 3, "
						+ "assuming default values for CCIR 601-1");
			}
		}

		field =
				tmetadata.getTIFFField(BaselineTIFFTagSet.TAG_REFERENCE_BLACK_WHITE);
		if (field != null) {
			if (field.getCount() == 6) {
				this._referenceBlackY = field.getAsFloat(0);
				this._referenceWhiteY = field.getAsFloat(1);
				this._referenceBlackCb = field.getAsFloat(2);
				this._referenceWhiteCb = field.getAsFloat(3);
				this._referenceBlackCr = field.getAsFloat(4);
				this._referenceWhiteCr = field.getAsFloat(5);
			} else {
				warning("REFERENCE_BLACK_WHITE count != 6, ignoring it");
			}
		} else {
			warning("REFERENCE_BLACK_WHITE not found, assuming 0-255/128-255/128-255");
		}

		this._colorConvert = true;

		float bCb = 2.0f - 2.0f * _LumaBlue;
		float rCr = 2.0f - 2.0f * _LumaRed;

		float gY = (1.0f - _LumaBlue - _LumaRed) / _LumaGreen;
		float gCb = 2.0f * _LumaBlue * (_LumaBlue - 1.0f) / _LumaGreen;
		float gCr = 2.0f * _LumaRed * (_LumaRed - 1.0f) / _LumaGreen;

		for (int i = 0; i < 256; i++) {
			float fY = (i - _referenceBlackY) * CONDING_RANGE_Y / (_referenceWhiteY - _referenceBlackY);
			float fCb = (i - _referenceBlackCb) * 127.0f / (_referenceWhiteCb - _referenceBlackCb);
			float fCr = (i - _referenceBlackCr) * 127.0f / (_referenceWhiteCr - _referenceBlackCr);

			_iYTab[i] = (int)(fY * FRAC_SCALE);
			_iCbTab[i] = (int)(fCb * bCb * FRAC_SCALE);
			_iCrTab[i] = (int)(fCr * rCr * FRAC_SCALE);

			_iGYTab[i] = (int)(fY * gY * FRAC_SCALE);
			_iGCbTab[i] = (int)(fCb * gCb * FRAC_SCALE);
			_iGCrTab[i] = (int)(fCr * gCr * FRAC_SCALE);
		}
	}

	@Override
	public void decodeRaw(byte[] buf,
			int dstOffset,
			int bitsPerPixel,
			int scanlineStride) throws IOException {
		//		 byte[] rows = new byte[3*srcWidth*chromaSubsampleV];

		int elementsPerPacket = _chromaSubsampleH * _chromaSubsampleV + 2;
		byte[] packet = new byte[elementsPerPacket];

		if (_decompressor != null) {
			int bytesPerRow = 3 * srcWidth;
			byte[] tmpBuf = new byte[bytesPerRow * srcHeight];
			_decompressor.decodeRaw(tmpBuf, dstOffset, bitsPerPixel,
					bytesPerRow);
			ByteArrayInputStream byteStream =
					new ByteArrayInputStream(tmpBuf);
			stream = new MemoryCacheImageInputStream(byteStream);
		} else {
			stream.seek(offset);
		}

		for (int y = srcMinY; y < srcMinY + srcHeight; y += _chromaSubsampleV) {
			// Decode chromaSubsampleV rows
			for (int x = srcMinX; x < srcMinX + srcWidth;
					x += _chromaSubsampleH) {
				try {
					stream.readFully(packet);
				} catch (EOFException e) {
					return;
				}

				byte cb = packet[elementsPerPacket - 2];
				byte cr = packet[elementsPerPacket - 1];

				int iCb  = 0, iCr = 0, iGCb = 0, iGCr = 0;

				if (_colorConvert) {
					int cbp = cb & 0xff;
					int crp = cr & 0xff;

					iCb = _iCbTab[cbp];
					iCr = _iCrTab[crp];

					iGCb = _iGCbTab[cbp];
					iGCr = _iGCrTab[crp];
				}

				int yIndex = 0;
				for (int v = 0; v < _chromaSubsampleV; v++) {
					int idx = dstOffset + 3 * (x - srcMinX)
							+ scanlineStride * (y - srcMinY + v);

					// Check if we reached the last scanline
					if (y + v >= srcMinY + srcHeight) {
						break;
					}

					for (int h = 0; h < _chromaSubsampleH; h++) {
						if (x + h >= srcMinX + srcWidth) {
							break;
						}

						byte yY = packet[yIndex++];

						if (_colorConvert) {
							int yp = yY & 0xff;
							int iY = _iYTab[yp];
							int iGY = _iGYTab[yp];

							int iR = iY + iCr;
							int iG = iGY + iGCb + iGCr;
							int iB = iY + iCb;

							byte red = clamp(iR);
							byte green = clamp(iG);
							byte blue = clamp(iB);

							buf[idx] = red;
							buf[idx + 1] = green;
							buf[idx + 2] = blue;
						} else {
							buf[idx] = yY;
							buf[idx + 1] = cb;
							buf[idx + 2] = cr;
						}

						idx += 3;
					}
				}
			}
		}
	}
}
