/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane;

import fr.greyc.pantheon.ariane.model.PlatformSettings;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileInputStream;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.view.ArianeFrame;
import fr.greyc.pantheon.ariane.view.SplashScreen;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import javax.imageio.spi.IIORegistry;
import javax.swing.ImageIcon;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

/**
 * <code>Ariane</code> is the main class of the software.
 * It reads the configuration file and initializes the application.
 * It uses a Splash Screen to pass the time.
 */
public final class Ariane {
	public static final String FRAME_TITLE = "Au fil d'Ariane ... @ GREYC (ENSICAEN & Université de Caen Basse-Normandie)";
	public static final String VERSION_ID = "3.3.4 (Sep 1, 2015)";
	public static final String[] AUTHORS = { "Régis CLOUARD",
		"François RIOULT",
		"Sébastien LEFÈVRE (1999)",
		"Richard LOBRÉAU (1999)",
		"Adrien SCHRÈQUE (2000)",
		"Hicham AMADI (2000)",
		"Mathieu JOETS (2001)",
		"Guillaume CHATELET (2002)",
		"Alban DOUBLET (2002)",
		"Frédéric LAIR (2007)"
	};

	/////
	///// C O N S T A N T S
	/////

	public static final String DEFAULT_ARIANE_FILE_EXT = ".aws";
	public static final String LIBRARY_CONFIG_FILENAME = "ariane.cfg";
	public static final String CONFIG_STATIC_DIRNAME = ".greyc" + File.separator + "ariane";
	public static final String CONFIG_DYNAMIC_DIRNAME = ".ariane";
	public static final String PREFERENCES_FILENAME = "ariane.ini";
	private static final String CONFIG_SUBROUTINE_FILENAME = "subroutines.cfg";
	private static final String CONFIG_OPERATOR_FILENAME = "operators.cfg";
	private static final String ICON_RESOURCE_DIRECTORY = "fr/greyc/pantheon/ariane/resources/images/";
	private static final String STRING_RESOURCE_DIRECTORY = "fr/greyc/pantheon/ariane/resources/locale/";
	private static final String STRING_RESOURCE_FILENAME = "messages";
	public static final int HISTORY_SIZE = 10;

	/**
	 * The List of all error codes used in the project.
	 */
	public enum ErrorCode {
		NOERROR,
		MISSINGPARAMETER,
		MISSINGRESULT,
		MISSINGINPUT,
		RUNTIME,
		INPUTOUTPUT,
		INTERRUPTEDEXECUTION,
		MISSINGFILE,
		UNKNOWNFILE,
		MISSINGFORMULA,
		ARITHMETICS,
	}

	private static String _operatorLibraryPath = ".";
	private static String _configurationDirectoryPath;
	private static ResourceBundle _language;
	private static String _arianeManualURL;

	private static Presenter _presenter;
	private static ArianeFrame _mainFrame;

	private static String _libraryDocumentationPath = "";

	public static ArianeFrame getFrame( ) {
		return _mainFrame;
	}

	public static Presenter getPresenter( ) {
		return _presenter;
	}

	public static ImageIcon getIcon( String filename ) {
		URL location = Ariane.class.getClassLoader().getResource(Ariane.ICON_RESOURCE_DIRECTORY + filename);
		return new ImageIcon(Toolkit.getDefaultToolkit().getImage(location));
	}

	public static String getString( final String name ) {
		return _language.getString(name);
	}

	public static URL getDocumentationURI( String page ) {
		return Ariane.class.getClassLoader().getResource(_arianeManualURL + page);
	}

	public static ResourceBundle getLanguage() {
		return _language;
	}

	public static String getPreferencesFilename() {
		return _configurationDirectoryPath + File.separator + PREFERENCES_FILENAME;
	}

	public static String getSubroutineFilename() {
		return _configurationDirectoryPath + File.separator + CONFIG_SUBROUTINE_FILENAME;
	}

	public static String getPersonalOperatorFilename() {
		return _configurationDirectoryPath + File.separator + CONFIG_OPERATOR_FILENAME;
	}

	public static void setLibraryPath( String path ) {
		_operatorLibraryPath = path;
		System.setProperty("java.library.path", PlatformSettings.getJavaLibraryPath() + File.pathSeparatorChar + _operatorLibraryPath + File.separatorChar + "lib");
	}

	public static String getLibraryPath( ) {
		return _operatorLibraryPath;
	}

	public static void setLibraryDocumentationPath( final String path ) {
		_libraryDocumentationPath = path;
	}

	public static String getLibraryDocumentationPath( ) {
		return _libraryDocumentationPath + "/" + UserPreferences.getCurrentLocale().getLanguage();
	}

	public static void setCurrentLocale( Locale locale ) {
		final String url = STRING_RESOURCE_DIRECTORY + locale.getLanguage() + "/";
		_language = ResourceBundle.getBundle(url + STRING_RESOURCE_FILENAME, locale);
		_arianeManualURL = url + "manual/";
	}

	protected static List<XMLElement> readConfigurationFile( File configFile ) {
		if (configFile == null) {
			return null;
		}
		try {
			return readXMLFile(configFile);
		} catch (IOException e) {
			System.err.println(Ariane.getString("error") + e.getLocalizedMessage());
			return null;
		}
	}

	public static List<XMLElement> readXMLFile( File configFile ) throws IOException {
		try {
			XMLFileInputStream convert = new XMLFileInputStream(configFile);
			convert.parse();
			convert.close();
			return convert.getElements();
		} catch (SAXException | IOException | ParserConfigurationException e) {
			throw new IOException(e);
		}
	}

	protected File locateConfigureFileIn( String directoryName ) {
		// 1. use the specified directory name
		if (directoryName != null) {
			File configurationFile = new File(directoryName + File.separator + PREFERENCES_FILENAME);
			if (configurationFile.exists()) {
				_configurationDirectoryPath = configurationFile.getAbsoluteFile().getParent();
				return configurationFile;
			}
		}
		// 2. The config directory is in the  same directory than the jar file.
		final String currentJarPath = PlatformSettings.getCurrentJarPath();
		final String[] possibleConfigJarFiles = {
				File.separator + CONFIG_DYNAMIC_DIRNAME + File.separator + PREFERENCES_FILENAME,
				File.separator + "ariane" + File.separator + PREFERENCES_FILENAME, // For compatibility
				File.separator + "ariane.cfg" // For compatibility
		};
		for (String configFile : possibleConfigJarFiles) {
			File configurationFile = new File(currentJarPath + configFile);
			if (configurationFile.exists()) {
				_configurationDirectoryPath = configurationFile.getParent();
				return configurationFile;
			}
		}
		// 3. The config directory is in user's config directory
		final String[] possibleConfigFiles = {
				File.separator + CONFIG_STATIC_DIRNAME + File.separator + PREFERENCES_FILENAME,
				File.separator + ".greyc"  + File.separator + "ariane.cfg", // For compatibility
				File.separator + ".ariane" // For compatibility
		};
		for (String configurationFileName : possibleConfigFiles) {
			File configurationFile1 = new File(PlatformSettings.getHomePath() + configurationFileName);
			if (configurationFile1.exists()) {
				_configurationDirectoryPath = PlatformSettings.getHomePath() + File.separator + CONFIG_STATIC_DIRNAME;
				return configurationFile1;
			}
		}
		// 4. default configuration file
		_configurationDirectoryPath = PlatformSettings.getHomePath() + File.separator + CONFIG_STATIC_DIRNAME;
		return null;
	}

	private String locateLibraryPath( List<XMLElement> preferenceHandler) {
		if (preferenceHandler != null) {
			UserPreferences.load(preferenceHandler);
			String libraryFilename = getLibraryPath() + File.separator + "lib" + File.separator + LIBRARY_CONFIG_FILENAME;
			File file = new File(libraryFilename);
			if (file.exists()) { 
				return libraryFilename;
			}
			// Kept for the sake of compatibility
			libraryFilename = getLibraryPath() + File.separator + "lib" + File.separator + "pandore.cfg";
			File file1 = new File(libraryFilename);
			if (!file1.exists()) {
				return null;
			}
		}
		return null;
	}

	private void initializeInterface( boolean doSplash, String operatorLibraryFilename, List<String> graphFilenames, String configDirName ) {
		_mainFrame = new ArianeFrame();
		_mainFrame.installUI();

		SplashScreen splash = new SplashScreen(ICON_RESOURCE_DIRECTORY + "splash.png");
		splash.setVisible(doSplash);

		splash.showProgress("Reading preferences file ...", 0);
		UserPreferences.initializePreferencesWithDefaultValues();
		File preferencesFile = locateConfigureFileIn(configDirName);
		List<XMLElement> preferenceHandler = readConfigurationFile(preferencesFile);
		String defaultLibraryFilename = locateLibraryPath(preferenceHandler);

		splash.showProgress("Reading common operators...", 20);
		List<XMLElement> libraryHandler;
		if (operatorLibraryFilename != null) {
			libraryHandler = readConfigurationFile(new File(operatorLibraryFilename));
		} else if (defaultLibraryFilename != null && UserPreferences.isOperatorLibraryUsed()) {
			libraryHandler = readConfigurationFile(new File(defaultLibraryFilename));
		} else {
			libraryHandler = new ArrayList<XMLElement>();
		}
		ItemTreeModel itemTree = new ItemTreeModel();
		itemTree.setCommonOperators(libraryHandler);

		splash.showProgress("Reading personal operators ...", 40);
		File personalOperatorFile = new File(Ariane.getPersonalOperatorFilename());
		if (personalOperatorFile.exists()) {
			itemTree.addInUserOperators(readConfigurationFile(personalOperatorFile));
		} else if (preferenceHandler != null) { // Keep this for sake of compatibility
			itemTree.addInUserOperators(preferenceHandler);
			File directory = personalOperatorFile.getParentFile();
			directory.mkdirs();
			preferencesFile.renameTo(personalOperatorFile);
		}

		splash.showProgress("Reading subroutines ...", 50);
		File subRoutineFile = new File(Ariane.getSubroutineFilename());
		if (subRoutineFile.exists()) {
			List<XMLElement> subRoutineHandler = readConfigurationFile(subRoutineFile);
			itemTree.addInSubroutines(subRoutineHandler);
		}

		splash.showProgress("Building graphical interfaces ...", 60);
		_presenter = new Presenter();
		_presenter.buildWorkspace(itemTree);

		splash.showProgress("Initializing windows contents ...", 80);
		ImageIcon logo64 = Ariane.getIcon("logo64.png");
		ImageIcon logo32 = Ariane.getIcon("logo32.png");
		_mainFrame.setIconImages(Arrays.asList(new Image[] {logo64.getImage(), logo32.getImage()}));
		_mainFrame.setJMenuBar(_presenter.buildMenuBar());
		_presenter.initContent();
		_mainFrame.initFrame(_presenter);

		if (!graphFilenames.isEmpty()) {
			splash.showProgress("Reading input workspace(s) ...", 90);
			_presenter.display(graphFilenames);
			File file = new File(graphFilenames.get(0));
			UserPreferences.setCurrentDirectory(file.getAbsoluteFile().getParent());
		}

		splash.showProgress("Starting session ...", 100);
		_mainFrame.restoreFrameSize();
		_presenter.restoreSplitPanesLocation();
		_mainFrame.setVisible(true);

		splash.setVisible(false);
		splash.dispose();

		if (isFirstUsed()) {
			_presenter.chooseOperatorLibrary();
		}
	}

	private boolean isFirstUsed() {
		return ".".equalsIgnoreCase(getLibraryPath()) && UserPreferences.isOperatorLibraryUsed();
	}

	private void setTestContext() {
		_presenter = new Presenter();
		_presenter.buildWorkspace(new ItemTreeModel());
	}

	/*
	 * Defined for tests
	 */
	public Ariane() {
		UserPreferences.initializePreferencesWithDefaultValues();
		setTestContext();
	}

	private Ariane( String[] args ) {
		final Map<String, Object> options = new HashMap<>();
		options.put("nosplash", false);
		final List<String> graphFilenames = CommandLineParser.parse(args, options);

		final boolean doSplash = !(Boolean)options.get("nosplash");
		final String configDirName = (String)options.get("config");
		final String operatorLibraryFilename = (String)options.get("lib");

		initializeInterface(doSplash, operatorLibraryFilename, graphFilenames, configDirName);
	}

	private static void installImageIOLibrary() {
		IIORegistry registry = IIORegistry.getDefaultInstance();
		registry.registerServiceProvider(new com.sun.media.imageioimpl.plugins.tiff.TIFFImageReaderSpi());
	}

	public static void main( String[] args ) {
		installImageIOLibrary();
		new Ariane(args);
	}
}
