/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model;

import java.awt.AWTException;
import java.awt.Dimension;
import java.awt.image.ColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageProducer;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Random;

/**
 * Extends ImageProducer to decode Pandore images. This code is based
 * on the book "Java AWT references" by John Zukowski.
 * 
 * Custom usage: PandoreImageDecoder pan= new PandoreImageDecoder();
 * pan.readImage(imagebyte); img = Toolkit.getDefaultToolkit().createImage(pan);
 */
public final class PandoreImageDecoder implements ImageProducer {
	/* List of the Pandore types. */
	/* Do not change the values, they are conventional. */

	//	private static final int Po_Collection = 1;
	//	private static final int Po_Img1duc = 2;
	//	private static final int Po_Img1dsl = 3;
	//	private static final int Po_Img1dsf = 4;

	private static final int PO_IMG2DUC = 5;
	private static final int PO_IMG2DSL = 6;
	private static final int PO_IMG2DSF = 7;

	private static final int PO_IMG3DUC = 8;
	private static final int PO_IMG3DSL = 9;
	private static final int PO_IMG3DSF = 10;

	//	private static final int Po_Reg1d = 11;
	private static final int PO_REG2D = 12;
	private static final int PO_REG3D = 13;

	//	private static final int Po_Graph2d = 14;
	//	private static final int Po_Graph3d = 15;

	private static final int PO_IMC2DUC = 16;
	private static final int PO_IMC2DSL = 17;
	private static final int PO_IMC2DSF = 18;

	private static final int PO_IMC3DUC = 19;
	private static final int PO_IMC3DSL = 20;
	private static final int PO_IMC3DSF = 21;

	//	private static final int Po_Imx1duc = 22;
	//	private static final int Po_Imx1dsl = 23;
	//	private static final int Po_Imx1dsf = 24;
	//	private static final int Po_Imx2duc = 25;
	//	private static final int Po_Imx2dsl = 26;
	//	private static final int Po_Imx2dsf = 27;
	//	private static final int Po_Imx3duc = 28;
	//	private static final int Po_Imx3dsl = 29;
	//	private static final int Po_Imx3dsf = 30;

	private static final byte[] PANDOREMAGIC = { 'P', 'A', 'N', 'D', 'O', 'R', 'E', '0', '4'};
	private int _imageType;
	private boolean _reverseByteOrder;
	private final int _maxSize;
	private ImageConsumer _consumer;
	private int[][] _pixelArray;
	private boolean _loadError;
	private Dimension _imageDimension;

	private static final int PANDORE_HINTS =
			ImageConsumer.TOPDOWNLEFTRIGHT
			| ImageConsumer.COMPLETESCANLINES
			| ImageConsumer.SINGLEPASS
			| ImageConsumer.SINGLEFRAME;

	public PandoreImageDecoder( int maxSize ) {
		_maxSize = maxSize;
	}

	private int readHeader( DataInputStream dis ) throws AWTException, IOException {
		byte[] buf = new byte[9];

		dis.read(buf, 0, 9);	// magic number
		if (!Arrays.equals(buf, PANDOREMAGIC)) {
			throw new AWTException("Invalid Pandore Format.");
		}
		dis.skipBytes(1);		// skip '\n'
		dis.skipBytes(2); 		// skip unused bytes.

		_imageType = dis.readInt();
		if (_imageType > 65535) { // Hint: when typeoj is wrong -> reverse byte order.
			_reverseByteOrder = true;
			_imageType = swap(_imageType);
		}

		dis.read(buf, 0, 8);	// Creator name
		dis.skipBytes(1);		// skip '\0'
		dis.read(buf, 0, 9);	// Date
		dis.skipBytes(1);		// '\0';
		dis.skipBytes(1);		// Skip unused bytes

		return _imageType;
	}

	/*
	 * Assume only one consumer.
	 */
	@Override
	public void addConsumer( ImageConsumer imageConsumer ) {
		_consumer = imageConsumer;
		produce();
		_consumer = null;
	}

	@Override
	public boolean isConsumer( ImageConsumer imageConsumer ) {
		return imageConsumer.equals(_consumer);
	}

	@Override
	public void removeConsumer( ImageConsumer imageConsumer ) {
		if (imageConsumer.equals(_consumer)) {
			_consumer = null;
		}
	}

	@Override
	public void startProduction( ImageConsumer imageConsumer ) {
		addConsumer(imageConsumer);
	}

	@Override
	public void requestTopDownLeftRightResend( ImageConsumer imageConsumer ) {
		// Not needed. The data is always in this format.
	}

	/*
	 * Production Process: Prerequisite: Image already read into store array.
	 * (readImage) props / width / height already set (readImage) Assumes RGB
	 * Color Model - would need to filter to change. Sends Pandore Image data to
	 * consumer. Pixels sent one row at a time.
	 */
	private void produce( ) {
		if (_consumer != null) {
			if (_loadError || _imageDimension == null) {
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
			} else {
				ColorModel colorModel = ColorModel.getRGBdefault();
				_consumer.setDimensions(_imageDimension.width, _imageDimension.height);
				_consumer.setColorModel(colorModel);
				_consumer.setHints(PANDORE_HINTS);
				for (int y = 0; y < _imageDimension.height; y++) {
					_consumer.setPixels(0, y, _imageDimension.width, 1, colorModel, _pixelArray[y], 0, _imageDimension.width);
				}
				_consumer.imageComplete(ImageConsumer.STATICIMAGEDONE);
			}
		}
	}

	private int swap( int value ) {
		int byte0 = value & 0xff;
		int byte1 = (value >> 8) & 0xff;
		int byte2 = (value >> 16) & 0xff;
		int byte3 = (value >> 24) & 0xff;
		return (byte0 << 24) | (byte1 << 16) | (byte2 << 8) | byte3;
	}

	public boolean readImage( String filename ) {
		FileInputStream fileImage = null;
		boolean result;
		try {
			fileImage = new FileInputStream(filename);
			if (fileImage.available() > 0) {
				readImage(fileImage);
			}
			result = true;
		} catch (IOException | AWTException e) {
			result = false;
		} finally {
			if (fileImage != null) {
				try {
					fileImage.close();
				} catch (IOException e) {
				}
			}
		}
		return result;
	}

	private int convertPandorePixelToJavaPixel( byte[] row, int column, int bytes ) {
		int color = 0;
		for (int b = 0; b < bytes; b++) {
			int value = row[column * bytes + b];
			// Problem signed -> unsigned.
			if (value < 0) {
				value += 256;
			}
			int shift = _reverseByteOrder ? b * 8 : (bytes - 1 - b) * 8;
			color |= value << shift;
		}

		return color;
	}

	private Dimension computeImageSize( int height, int width ) {
		Dimension dimension = new Dimension();
		if (width >= height) {
			if (width > _maxSize) {
				int ratio = 1 + width / _maxSize;
				dimension.width = width / ratio;
				dimension.height = height / ratio;
			}
		} else {
			if (height > _maxSize) {
				int ratio = 1 + height / _maxSize;
				dimension.width = width / ratio;
				dimension.height = height / ratio;
			}
		}
		return dimension;
	}

	private int readInt( DataInputStream inputStream ) throws IOException {
		byte[] row = new byte[4];
		for (int b = 0; b < row.length; b++) {
			row[b] = inputStream.readByte();
		}
		return convertPandorePixelToJavaPixel(row, 0, row.length);
	}

	private int[][] read2dGrayData( DataInputStream dataInputStream ) throws IOException {
		// read image size.
		readInt(dataInputStream); // bands = 1 !
		final int height = readInt(dataInputStream);
		final int width = readInt(dataInputStream);

		// read data
		_imageDimension = computeImageSize(height, width);
		final int[][] store = new int[_imageDimension.height][_imageDimension.width];
		final int stepX = width / _imageDimension.width;
		final int stepY = height / _imageDimension.height;

		switch (_imageType) {
		case PO_IMG2DUC:
			read2DGrayIntPixels(dataInputStream, width, store, stepX, stepY, 1);
			break;

		case PO_IMG2DSL:
			read2DGrayIntPixels(dataInputStream, width, store, stepX, stepY, 4);
			break;

		case PO_REG2D:
			int nlabels = readInt(dataInputStream) + 1;
			read2DRegionPixels(dataInputStream, width, store, stepX, stepY, nlabels);
			break;

		case PO_IMG2DSF:
			read2DGrayFloatBand(dataInputStream, width, store, stepX, stepY);
			break;
		default:
			break;
		}
		return store;
	}

	/*
	 * Note that this method is used for displaying
	 * image thumbnail so it does not read all the data but
	 * only the data of the middle slice.
	 */
	private int[][] read3dGrayData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // read the number of bands : always 1 !
		final int depth = readInt(dis);
		final int height = readInt(dis);
		final int width = readInt(dis);

		_imageDimension = computeImageSize(height, width);
		final int[][] store = new int[_imageDimension.height][_imageDimension.width];
		final int stepX = width / _imageDimension.width;
		final int stepY = height / _imageDimension.height;

		// read data
		switch (_imageType) {
		case PO_IMG3DUC:
			dis.skipBytes((depth / 2) * width * height);
			read2DGrayIntPixels(dis, width, store, stepX, stepY, 1);
			break;

		case PO_IMG3DSL:
			dis.skipBytes((depth / 2) * width * 4 * height);
			read2DGrayIntPixels(dis, width, store, stepX, stepY, 4);
			break;

		case PO_IMG3DSF:
			dis.skipBytes((depth / 2) * width * 4 * height);
			read2DGrayFloatBand(dis, width, store, stepX, stepY);
			break;

		case PO_REG3D:
			int nlabels = readInt(dis) + 1;
			dis.skipBytes((depth / 2) * width * height);
			read2DRegionPixels(dis, width, store, stepX, stepY, nlabels);
			break;
		default:
			break;

		}
		return store;
	}

	private int[][] read2dColorData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 3;
		int height = readInt(dis);
		int width = readInt(dis);

		// ColorSpace.
		dis.readInt();

		// read data
		_imageDimension = computeImageSize(height, width);
		int[][] store = new int[_imageDimension.height][_imageDimension.width];
		int stepX = width / _imageDimension.width;
		int stepY = height / _imageDimension.height;

		switch (_imageType) {
		case PO_IMC2DUC:
			// red band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 16);
			dis.skipBytes(width * (height - stepY * _imageDimension.height));
			// green band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 8);
			dis.skipBytes(width * (height - stepY * _imageDimension.height));
			// blue band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 0);
			break;

		case PO_IMC2DSL:
			// red band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 4, 16);
			dis.skipBytes(4 * width * (height - stepY * _imageDimension.height));
			// green band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 4, 8);
			dis.skipBytes(4 * width * (height - stepY * _imageDimension.height));
			// blue band
			read2DColorIntPixels(dis, width, store, stepX, stepY, 4, 0);
			break;

		case PO_IMC2DSF:
			// red band
			read2DColorFloatPixels(dis, width, store, stepX, stepY, 16);
			dis.skipBytes(4 * width * (height - stepY * _imageDimension.height));
			// green band
			read2DColorFloatPixels(dis, width, store, stepX, stepY, 8);
			dis.skipBytes(4 * width * (height - stepY * _imageDimension.height));
			// blue band
			read2DColorFloatPixels(dis, width, store, stepX, stepY, 0);
			break;
		default:
			break;
		}
		return store;
	}

	private int[][] read3dColorData( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 1 !
		int depth = readInt(dis);
		int height = readInt(dis);
		int width = readInt(dis);

		// read data
		_imageDimension = computeImageSize(height, width);
		int[][] store = new int[_imageDimension.height][_imageDimension.width];
		int stepX = width / _imageDimension.width;
		int stepY = height / _imageDimension.height;
		switch (_imageType) {
		case PO_IMC3DUC:
			dis.skipBytes((depth / 2) * width * height);
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 16);
			dis.skipBytes(width * (height - stepY * _imageDimension.height));
			dis.skipBytes((depth / 2) * width * height);
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 8);
			dis.skipBytes(width * (height - stepY * _imageDimension.height));
			dis.skipBytes((depth / 2) * width * height);
			read2DColorIntPixels(dis, width, store, stepX, stepY, 1, 0);
			break;

		case PO_IMC3DSL:
			break;

		case PO_IMC3DSF:
			break;
		default:
			break;
		}

		return store;
	}

	private void readImage( InputStream imageStream ) throws AWTException {
		try {
			DataInputStream dataInputStream = new DataInputStream(new BufferedInputStream(imageStream));
			readHeader(dataInputStream);

			switch (_imageType) {
			case PO_IMG2DUC:
			case PO_IMG2DSL:
			case PO_IMG2DSF:
			case PO_REG2D:
				_pixelArray = read2dGrayData(dataInputStream);
				break;

			case PO_IMC2DUC:
			case PO_IMC2DSL:
			case PO_IMC2DSF:
				_pixelArray = read2dColorData(dataInputStream);
				break;

			case PO_IMG3DUC:
			case PO_IMG3DSL:
			case PO_IMG3DSF:
			case PO_REG3D:
				_pixelArray = read3dGrayData(dataInputStream);
				break;

			case PO_IMC3DUC:
			case PO_IMC3DSL:
			case PO_IMC3DSF:
				_pixelArray = read3dColorData(dataInputStream);
				break;
			default:
				break;
			}
		} catch (IOException io) {
			_loadError = true;
		}
	}

	private int buildRegionLUT( int nlabels, int[] lut ) {
		int bytes;
		if (nlabels < 256) {
			bytes = 1;
		} else if (nlabels < 65536) {
			bytes = 2;
		} else {
			bytes = 4;
		}
		// Colors are set randomly except 0
		final Random rand = new Random(0);
		lut[0] = 0xff << 24;
		for (int i = 1; i < nlabels; i++) {
			lut[i] = (0xff << 24) | (rand.nextInt(256) << 16) | (rand.nextInt(256) << 8) | rand.nextInt(256);
		}
		return bytes;
	}

	private void read2DRegionPixels( DataInputStream dataInputStream, int width, int[][] store, int stepX, int stepY, int nlabels ) throws IOException {
		int[] lut = new int[nlabels];
		int bytes = buildRegionLUT(nlabels, lut);
		byte[] row = new byte[width * bytes];
		for (int y = 0; y < _imageDimension.height; y++) {
			dataInputStream.readFully(row);
			dataInputStream.skipBytes(width * bytes * (stepY - 1));
			for (int x = 0; x < _imageDimension.width; x++) {
				int color = convertPandorePixelToJavaPixel(row, x * stepX, bytes);
				if (color < nlabels) {
					store[y][x] = lut[color];
				}
			}
		}
	}

	private void read2DGrayIntPixels( DataInputStream dataInputStream, int width, int[][] store, int stepX, int stepY, int size ) throws IOException {
		byte[] row = new byte[width * size];
		for (int y = 0; y < _imageDimension.height; y++) {
			dataInputStream.readFully(row);
			dataInputStream.skipBytes(width * size * (stepY - 1));
			for (int x = 0; x < _imageDimension.width; x++) {
				int color = convertPandorePixelToJavaPixel(row, x * stepX, size);
				store[y][x] = (0xff << 24) | (color << 16) | (color << 8) | color;
			}
		}
	}

	private void read2DGrayFloatBand( DataInputStream dataInputStream, int width, int[][] store, int stepX, int stepY ) throws IOException {
		byte[] row = new byte[width * 4];
		for (int y = 0; y < _imageDimension.height; y++) {
			dataInputStream.readFully(row);
			dataInputStream.skipBytes(width * 4 * (stepY - 1));
			for (int x = 0; x < _imageDimension.width; x++) {
				int color = convertPandorePixelToJavaPixel(row, x * stepX, 4);
				color = (int)Float.intBitsToFloat(color); 
				store[y][x] = (0xff << 24) | (color << 16) | (color << 8) | color;
			}
		}
	}

	private void read2DColorFloatPixels( DataInputStream dis, int width, int[][] store, int stepX, int stepY, int shift ) throws IOException {
		byte[] row = new byte[width * 4];
		for (int y = 0; y < _imageDimension.height; y++) {
			dis.readFully(row);
			dis.skipBytes(width * 4 * (stepY - 1));
			for (int x = 0; x < _imageDimension.width; x++) {
				int color = convertPandorePixelToJavaPixel(row, x * stepX, 4);
				color = (int) Float.intBitsToFloat(color);
				store[y][x] |= (0xff << 24) | (color << shift);
			}
		}
	}
	
	private void read2DColorIntPixels( DataInputStream dis, int width, int[][] store, int stepX, int stepY, int size, int shift ) throws IOException {
		byte[] row = new byte[width * size];
		for (int y = 0; y < _imageDimension.height; y++) {
			dis.readFully(row);
			dis.skipBytes(width * size * (stepY - 1));
			for (int x = 0; x < _imageDimension.width; x++) {
				int color = convertPandorePixelToJavaPixel(row, x * stepX, size);
				store[y][x] |= (0xff << 24) | (color << shift);
			}
		}
	}
}