/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.ToolTipManager;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.ariane.view.graph.item.etc.ControlColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.ControlStructureColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.DataFileColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.NoteColor;
import fr.greyc.pantheon.ariane.view.graph.item.etc.OperatorColor;

/**
 * The class <code>UserPreferences</code> stores the current values
 * for all user's preferences.
 * <br>
 * This class sounds like the @link{java.util.prefs.Preferences Preferences class},
 * but the persistence is ensured by our own methods.
 * Each preference is handled using the static get and put methods:
 * <pre>
 * Object value = PreferencesSettings.get("property_name",default_value);
 * PreferencesSettings.set("property_name", value);
 * </pre>
 */
public final class UserPreferences {
	private static final String UNDO_HISTORY_SIZE = "UndoHistorySize";
	private static final String USE_OPERATOR_LIBRARY = "UseOperatorLibrary";
	private static final String PREFERRED_BINARY_DIRECTORY = "BinaryDirectory";
	private static final String LANGUAGE = "Language";
	private static final String FRAME_SIZE = "Frame.Size";
	private static final String TEXT_EDITOR = "TextEditor";
	private static final String USE_TOOLTIP = "UseToolTip";
	private static final String WEB_BROWSER = "WebBrowser";
	private static final String PREFERRED_FILE_DIRECTORY = "FileDirectory";
	private static final String FONT_SIZE = "FontSize";
	private static final String CODE_GENERATOR_PREFERENCE = "CodeGenerator.Preference";
	private static final String TEMPORARY_FILE_EXTENSION = "TemporaryFileExtension";
	private static final String FRAME_MAXIMIZED = "Frame.Maximized";
	private static final String SPLITPANE_H1_RATIO = "SplitPane.H1.Ratio";
	private static final String SPLITPANE_H2_RATIO = "SplitPane.H2.Ratio";
	private static final String SPLITPANE_V1_RATIO = "SplitPane.V1.Ratio";
	private static final String SPLITPANE_V2_RATIO = "SplitPane.V2.Ratio";
	private static final String COMMAND_STATUS = "Command.Status";
	private static final String COMMAND_VISU = "Command.Visu";
	private static final String DATAFILE_COLOR = "Datafile.Color";
	private static final String OPERATOR_COLOR = "Operator.Color";
	private static final String CONTROL_COLOR = "Control.Color";
	private static final String COMPOSITE_COLOR = "Composite.Color";
	private static final String NOTE_COLOR = "Note.Color";
	private static final String DATA_LINK_COLOR = "Image.Link.Color";
	private static final String VALUE_LINK_COLOR = "Value.Link.Color";
	private static final String MASK_LINK_COLOR = "Mask.Link.Color";
	public static final String EMPTY_STRING = "\t";

	private static String _commandStatus = "";
	private static String _commandVisu = "";

	private static String _temporaryDirectory;
	private static String _currentDirectory;
	private static String _temporaryFileExtension = ".tmp";
	private static String _temporaryFilePrefix = "data";

	private static Locale _locale;
	private static Font _defaultFont;

	private static Font _fontDialogPLAIN6;
	private static Font _fontDialogPLAIN8;
	private static Font _fontDialogPLAIN9;
	private static Font _fontDialogPLAIN11;
	private static Font _fontDialogPLAIN10;
	private static Font _fontDialogPLAIN12;
	private static Font _fontDialogPLAIN16;

	/** Stores all the preferences. */
	private static Map<String, Object> _preferences = new HashMap<String, Object>();

	private UserPreferences() { }

	public static void initializePreferencesWithDefaultValues( ) {
		loadFonts();
		putDimension(FRAME_SIZE, new Dimension(1000, 660));
		putInt(SPLITPANE_V1_RATIO, 232);
		putInt(SPLITPANE_V2_RATIO, 584);
		putInt(SPLITPANE_H1_RATIO, 436);
		putInt(SPLITPANE_H2_RATIO, 433);
		putInt(CODE_GENERATOR_PREFERENCE, 0);
		putBoolean(USE_OPERATOR_LIBRARY, true);
		putBoolean(USE_TOOLTIP, true);
		putInt(UNDO_HISTORY_SIZE, 200);
		setDefaultLocale();
		setFontSize(10);
		setCurrentDirectory(System.getProperty("user.dir") + File.separator);
		try {
			setTemporaryDirectoryBase(getSystemTemporaryDirectory());
		} catch (IOException e) { }
	}

	private static void setDefaultLocale() {
		_locale = Locale.getDefault();
		if (_locale.getDisplayLanguage().equals(Locale.FRENCH.getDisplayLanguage())) {
			setLocale("Français"); 
		} else {
			setLocale("English");
		}
	}

	private static void putBoolean( final String key, final boolean value ) {
		_preferences.put(key, value);
	}

	private static boolean getBoolean( String key ) {
		if (_preferences.containsKey(key)) {
			return (Boolean)_preferences.get(key);
		} else {
			return false;
		}
	}

	private static void put( String key, String value ) {
		_preferences.put(key, value);
	}

	private static String get( String key ) {
		return (String)_preferences.get(key);
	}

	private static void putDimension( String key, Dimension value ) {
		_preferences.put(key, value);
	}

	private static Dimension getDimension( String key ) {
		if (_preferences.containsKey(key)) {
			return (Dimension)_preferences.get(key);
		} else {
			return null;
		}
	}

	private static void putInt( String key, int value ) {
		_preferences.put(key, value);
	}

	private static int getInt( String key ) {
		return (Integer)_preferences.get(key);
	}

	/**
	 * Creates a directory unique in the current directory base.
	 * This allows multiple Ariane at the same time.
	 * @param directoryName a regular directory name.
	 */
	private static void createTemporaryFolder( String directoryName ) throws IOException {
		File directory = new File(directoryName);
		if (directory.isDirectory()) {
			if (Files.isWritable(directory.toPath())) {
				File file;
				Long index = 1L;
				do {
					file = new File(directoryName + "ariane" + index++);
				} while (!file.mkdirs());
				_temporaryDirectory = file.getAbsolutePath() + File.separator;
			} else {
				throw new IOException(Ariane.getLanguage().getString("UserPreferences.error2") + ": " + directoryName);
			}
		} else {
			throw new IOException(Ariane.getLanguage().getString("UserPreferences.error1") + ": " + directoryName);
		}
	}

	public static void setTemporaryDirectoryBase( String directory ) throws IOException {
		if (directory.endsWith(File.separator)) {
			createTemporaryFolder(directory);
		} else {
			createTemporaryFolder(directory + File.separator);
		}
	}

	public static String getTemporaryDirectoryBase( ) {
		File file = new File(_temporaryDirectory);
		return file.getParent() + File.separator;
	}

	private static String getSystemTemporaryDirectory( ) {
		return System.getProperty("java.io.tmpdir") + File.separator;
	}

	public static String getTemporaryDirectory( ) {
		return _temporaryDirectory;
	}

	public static String createTemporaryName( int graphIndex, int itemIndex, int slotIndex ) {
		String name = _temporaryFilePrefix
				+ "_"
				+ graphIndex
				+ "_" + itemIndex
				+ "_" + slotIndex + UserPreferences._temporaryFileExtension;
		return name;
	}

	public static String getTemporaryFilePrefix() {
		return _temporaryFilePrefix;
	}

	public static String getTemporaryFileExtension() {
		return _temporaryFileExtension;
	}

	public static void setTemporaryFileExtension( String extension ) {
		_temporaryFileExtension = extension;
	}

	public static Locale getCurrentLocale() {
		return _locale;
	}

	public static void setToolTipState( boolean state ) {
		putBoolean(USE_TOOLTIP, state);
		ToolTipManager.sharedInstance().setEnabled(state);
	}

	public static boolean isToolTipEnabled() {
		return getBoolean(USE_TOOLTIP);
	}

	public static void setCommandVisu( String name ) {
		_commandVisu = name.trim();
	}

	public static String getCommandVisu( ) {
		return _commandVisu;
	}

	public static void setCommandStatus( String name ) {
		_commandStatus = name.trim();
	}

	public static String getCommandStatus( ) {
		return _commandStatus;
	}

	public static void changeLanguage( String language ) {
		String currentLanguage = get(LANGUAGE);
		if (!currentLanguage.equals(language)) {
			setLocale(language);
			Ariane.getPresenter().changeLanguage();
		}
	}

	public static String getLanguage() {
		return get(UserPreferences.LANGUAGE);
	}

	public static int getHistorySize() {
		return getInt(UserPreferences.UNDO_HISTORY_SIZE);
	}

	public static void setHistorySize( int size ) {
		putInt(UNDO_HISTORY_SIZE, size);
	}

	public static String getWebBrowserPath() {
		return get(WEB_BROWSER);
	}

	public static void setWebBrowserPath( String path ) {
		put(WEB_BROWSER, path);
	}

	public static String getTextEditorPath() {
		return get(TEXT_EDITOR);
	}

	public static void setTextEditorPath( String path ) {
		put(TEXT_EDITOR, path);
	}

	public static String getPreferredBinaryDirectory() {
		return get(PREFERRED_BINARY_DIRECTORY);
	}

	public static void setBinaryDirectory( String path ) {
		put(PREFERRED_BINARY_DIRECTORY, path);
	}

	public static boolean isOperatorLibraryUsed() {
		return getBoolean(USE_OPERATOR_LIBRARY);
	}

	public static void setOperatorLibraryState( boolean state ) {
		putBoolean(UserPreferences.USE_OPERATOR_LIBRARY, state);
	}

	public static String getPreferredFileDirectory() {
		return get(PREFERRED_FILE_DIRECTORY);
	}

	public static void setPreferedFileDirectory( String path ) {
		put(PREFERRED_FILE_DIRECTORY, path);
	}

	public static int getCodeGeneratorIndex() {
		return getInt(CODE_GENERATOR_PREFERENCE);
	}

	public static void getCodeGeneratorIndex( int index ) {
		putInt(CODE_GENERATOR_PREFERENCE, index);
	}

	public static Dimension getFrameSize() {
		return getDimension(FRAME_SIZE);
	}

	public static String getCurrentDirectory( ) {
		return _currentDirectory;
	}

	public static void setCurrentDirectory( String directory ) { 
		_currentDirectory = directory;
	}

	public static boolean isFrameMaximized() {
		return UserPreferences.getBoolean(FRAME_MAXIMIZED);
	}

	public static int getSplitPaneH1Ratio() {
		return UserPreferences.getInt(UserPreferences.SPLITPANE_H1_RATIO);
	}

	public static int getSplitPaneH2Ratio() {
		return UserPreferences.getInt(UserPreferences.SPLITPANE_H2_RATIO);
	}

	public static int getSplitPaneV1Ratio() {
		return UserPreferences.getInt(UserPreferences.SPLITPANE_V1_RATIO);
	}

	public static int getSplitPaneV2Ratio() {
		return UserPreferences.getInt(UserPreferences.SPLITPANE_V2_RATIO);
	}

	///// 
	///// F O N T
	///// 

	private static void loadFonts( ) { // Change fonts...
		int screenRes = Toolkit.getDefaultToolkit().getScreenResolution();
		int fontSize = (int)Math.round(12.0 * screenRes / 72.0);
		Font font = new Font("Arial", Font.PLAIN, fontSize);
		_fontDialogPLAIN6 = font.deriveFont(6f);
		_fontDialogPLAIN8 = font.deriveFont(8f);
		_fontDialogPLAIN9 = font.deriveFont(9f);
		_fontDialogPLAIN10 = font.deriveFont(10.0f);
		_fontDialogPLAIN11 = font.deriveFont(11.0f);
		_fontDialogPLAIN12 = font.deriveFont(12f);
		_fontDialogPLAIN16 = font.deriveFont(16f);
	}

	public static Font getFont( int size ) {
		switch (size) {
		case 6: return _fontDialogPLAIN6;
		case 8: return _fontDialogPLAIN8;
		case 9: return _fontDialogPLAIN9;
		case 10: return _fontDialogPLAIN10;
		case 11: return _fontDialogPLAIN11;
		case 12: return _fontDialogPLAIN12;
		case 16: return _fontDialogPLAIN16;
		default: return null;
		}
	}

	public static void setFontSize( int fontSize ) {
		putInt(FONT_SIZE, fontSize);
		switch (fontSize) {
		case 8: _defaultFont = _fontDialogPLAIN8; break;
		case 9: _defaultFont = _fontDialogPLAIN9; break;
		case 10: _defaultFont = _fontDialogPLAIN10; break;
		case 11: _defaultFont = _fontDialogPLAIN11; break;
		case 12: _defaultFont = _fontDialogPLAIN12; break;
		default: _defaultFont = _fontDialogPLAIN10; break;
		}
	}

	public static Font getDefaultFont( ) {
		return _defaultFont;
	}

	public static int getFontSize() {
		return getInt(FONT_SIZE);
	}

	/**
	 * Sets the current language to the specified language.
	 * Loads related resources from the resource files.
	 * 
	 * @param language the name of the language (eg., français, english).
	 */
	private static void setLocale( String language ) {
		put(LANGUAGE, language);
		if ("Français".equals(language)) {
			_locale = new Locale("fr", "FR");
		} else if ("English".equals(language)) {
			_locale = new Locale("en", "US");
		}
		Locale.setDefault(_locale);
		Ariane.setCurrentLocale(_locale);
	}

	public static int deleteTemporaryFiles() {
		File directory = new File(getTemporaryDirectory());
		String[] files = directory.list();
		if (files == null) {
			return 0;
		}

		int numberOfErasedFiles = 0;
		for (int i = 0; i < files.length; i++) {
			if (files[i].startsWith(getTemporaryFilePrefix()) && files[i].endsWith(getTemporaryFileExtension())) {
				File file = new File(getTemporaryDirectory() + files[i]);
				if (file.delete()) {
					numberOfErasedFiles++;
				} else {
					Ariane.getPresenter().printErrorMessage(Ariane.getString("PreferenceChooser.environment.error1") + " " + files[i]);
				}
			}
		}
		directory.delete();
		return numberOfErasedFiles;
	}

	public static boolean saveInUserPreferenceFile() {
		File file = new File(Ariane.getPreferencesFilename());
		File directory = file.getParentFile();
		try {
			directory.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(file, null);
			out.write(convertPreferencesToXml());
			out.write(convertRecentFilesToXml());
			out.close();
		} catch (IOException ex) {
			Ariane.getPresenter().printSystemMessage(Ariane.getString("PreferenceChooser.error.saved") + " " + Ariane.getPreferencesFilename());
			return false;
		}
		return true;
	}

	public static void load( List<XMLElement> xmlTree1 ) {
		// Discards tag <ARIANE> from the xmlTree.
		List<XMLElement> properties = xmlTree1.get(0).getChildren();
		Iterator<XMLElement> iterator = properties.iterator();

		XMLElement element;
		try {
			while (iterator.hasNext()) {
				element = iterator.next();
				if ("Preferences".equalsIgnoreCase(element.getName())) {
					loadPreferences(element.getChildren());
				} else if ("RecentFiles".equalsIgnoreCase(element.getName()) && element.getChildren() != null) {
					convertXMLToRecentFiles(element.getChildren());
				}
			}
		} catch (IOException e) {
		}
	}

	private static String codeAsTriplet( Color color ) {
		return color.getRed() + "," + color.getGreen() + "," + color.getBlue();
	}

	private static Color decodeFromTriplet( String color ) {
		StringTokenizer colorValuePart = new StringTokenizer(color, ",");

		int red = Integer.parseInt(colorValuePart.nextToken());
		int green = Integer.parseInt(colorValuePart.nextToken());
		int blue = Integer.parseInt(colorValuePart.nextToken());
		return new Color(red, green, blue);
	}

	private static XMLElement convertPreferencesToXml( ) throws IOException {
		XMLElement element = new XMLElement("Preferences");
		element.addChildren(new XMLElement("CurrentDirectory", getCurrentDirectory()));
		if (!getSystemTemporaryDirectory().equals(getTemporaryDirectoryBase())) {
			element.addChildren(new XMLElement("TemporaryDirectory", getTemporaryDirectoryBase()));
		}
		element.addChildren(new XMLElement("OperatorLibraryDirectory", Ariane.getLibraryPath()));
		if (!UserPreferences.getCommandVisu().isEmpty()) {
			element.addChildren(new XMLElement(COMMAND_VISU, UserPreferences.getCommandVisu()));
		}
		if (!UserPreferences.getCommandStatus().isEmpty()) {
			element.addChildren(new XMLElement(COMMAND_STATUS, UserPreferences.getCommandStatus()));
		}
		if (!NoteColor.getDefaultColor().equals(NoteColor.getColor())) {
			element.addChildren(new XMLElement(NOTE_COLOR, codeAsTriplet(NoteColor.getColor())));
		}
		if (!OperatorColor.getDefaultColor().equals(OperatorColor.getColor())) {
			element.addChildren(new XMLElement(OPERATOR_COLOR, codeAsTriplet(OperatorColor.getColor())));
		}
		if (!DataFileColor.getDefaultColor().equals(DataFileColor.getColor())) {
			element.addChildren(new XMLElement(DATAFILE_COLOR, codeAsTriplet(DataFileColor.getColor())));
		}
		if (!ControlColor.getDefaultColor().equals(ControlColor.getColor())) {
			element.addChildren(new XMLElement(CONTROL_COLOR, codeAsTriplet(ControlColor.getColor())));
		}
		if (!ControlStructureColor.getDefaultColor().equals(ControlStructureColor.getColor())) {
			element.addChildren(new XMLElement(COMPOSITE_COLOR, codeAsTriplet(ControlStructureColor.getColor())));
		}
		if (!GraphView.getDefaultBackgroundColor().equals(GraphView.getBackgroundColor())) {
			element.addChildren(new XMLElement("Graph.Background.Color", codeAsTriplet(GraphView.getBackgroundColor())));
		}
		if (!GraphView.getDefaultImageLinkColor().equals(GraphView.getImageLinkColor())) {
			element.addChildren(new XMLElement(DATA_LINK_COLOR, codeAsTriplet(GraphView.getImageLinkColor())));
		}
		if (!GraphView.getDefaultValueLinkColor().equals(GraphView.getValueLinkColor())) {
			element.addChildren(new XMLElement(VALUE_LINK_COLOR, codeAsTriplet(GraphView.getValueLinkColor())));
		}
		if (!GraphView.getDefaultMaskLinkColor().equals(GraphView.getMaskLinkColor())) {
			element.addChildren(new XMLElement(MASK_LINK_COLOR, codeAsTriplet(GraphView.getMaskLinkColor())));
		}
		element.addChildren(new XMLElement(LANGUAGE, get(LANGUAGE)));
		element.addChildren(new XMLElement(FONT_SIZE, Integer.toString(getInt(FONT_SIZE))));
		element.addChildren(new XMLElement("ToolTip", Boolean.toString(getBoolean(USE_TOOLTIP))));
		String web = get(WEB_BROWSER);
		if (web != null && !web.trim().isEmpty()) {
			element.addChildren(new XMLElement(WEB_BROWSER, web));
		}
		String editor = get(TEXT_EDITOR);
		if (editor != null && !editor.trim().isEmpty()) {
			element.addChildren(new XMLElement(TEXT_EDITOR, editor));
		}

		Dimension dimension = Ariane.getFrame().getFinalSize();
		element.addChildren(new XMLElement("Frame.Width", Integer.toString(dimension.width)));
		element.addChildren(new XMLElement("Frame.Height", Integer.toString(dimension.height)));
		element.addChildren(new XMLElement(FRAME_MAXIMIZED, Boolean.toString(Ariane.getFrame().getExtendedState() == Frame.MAXIMIZED_BOTH)));
		element.addChildren(new XMLElement("SplitH1.ratio", Integer.toString(Ariane.getPresenter().getSplitH1())));
		element.addChildren(new XMLElement("SplitH2.ratio", Integer.toString(Ariane.getPresenter().getSplitH2())));
		element.addChildren(new XMLElement("SplitV1.ratio", Integer.toString(Ariane.getPresenter().getSplitV1())));
		element.addChildren(new XMLElement("SplitV2.ratio", Integer.toString(Ariane.getPresenter().getSplitV2())));
		element.addChildren(new XMLElement(USE_OPERATOR_LIBRARY, Boolean.toString(getBoolean(USE_OPERATOR_LIBRARY))));
		element.addChildren(new XMLElement(TEMPORARY_FILE_EXTENSION, _temporaryFileExtension));
		element.addChildren(new XMLElement(UNDO_HISTORY_SIZE, Integer.toString(getInt(UNDO_HISTORY_SIZE))));
		element.addChildren(new XMLElement(CODE_GENERATOR_PREFERENCE, Integer.toString(getInt(CODE_GENERATOR_PREFERENCE))));
		if (get(PREFERRED_FILE_DIRECTORY) != null) {
			element.addChildren(new XMLElement(PREFERRED_FILE_DIRECTORY, get(PREFERRED_FILE_DIRECTORY)));
		}
		if (get(PREFERRED_BINARY_DIRECTORY) != null) {
			element.addChildren(new XMLElement(PREFERRED_BINARY_DIRECTORY, get(PREFERRED_BINARY_DIRECTORY)));
		}
		return element;
	}

	private static void loadPreferences( List<XMLElement> xmlTree ) throws IOException {
		Iterator<XMLElement> iterator = xmlTree.iterator();
		while (iterator.hasNext()) {
			XMLElement element = iterator.next();
			// A l'interieur de <Preferences>
			String name = element.getName();
			if ("CurrentDir".equalsIgnoreCase(name) || "CurrentDirectory".equalsIgnoreCase(name)) {
				setCurrentDirectory(element.getValue());
			} else if ("TempDir".equalsIgnoreCase(name) || "TemporaryDirectory".equalsIgnoreCase(name)) {
				setTemporaryDirectoryBase(element.getValue());
			} else if ("PandoreDir".equalsIgnoreCase(name) || "OperatorLibraryDirectory".equalsIgnoreCase(name)) {
				Ariane.setLibraryPath(element.getValue());
			} else if (COMMAND_VISU.equalsIgnoreCase(name)) {
				setCommandVisu(element.getValue());
			} else if (COMMAND_STATUS.equalsIgnoreCase(name)) {
				setCommandStatus(element.getValue());
			} else if (NOTE_COLOR.equalsIgnoreCase(name)) {
				NoteColor.setColor(decodeFromTriplet(element.getValue()));
			} else if (OPERATOR_COLOR.equalsIgnoreCase(name)) {
				OperatorColor.setColor(decodeFromTriplet(element.getValue()));
			} else if (DATAFILE_COLOR.equalsIgnoreCase(name)) {
				DataFileColor.setColor(decodeFromTriplet(element.getValue()));
			} else if (CONTROL_COLOR.equalsIgnoreCase(name)) {
				ControlColor.setColor(decodeFromTriplet(element.getValue()));
			} else if (COMPOSITE_COLOR.equalsIgnoreCase(name)) {
				ControlStructureColor.setColor(decodeFromTriplet(element.getValue()));
			} else if ("Graph.Background.Color".equalsIgnoreCase(name)) {
				GraphView.setBackgroundColor(decodeFromTriplet(element.getValue()));
			} else if (DATA_LINK_COLOR.equalsIgnoreCase(name)) {
				GraphView.setImageLinkColor(decodeFromTriplet(element.getValue()));
			} else if (VALUE_LINK_COLOR.equalsIgnoreCase(name)) {
				GraphView.setValueLinkColor(decodeFromTriplet(element.getValue()));
			} else if (MASK_LINK_COLOR.equalsIgnoreCase(name)) {
				GraphView.setMaskLinkColor(decodeFromTriplet(element.getValue()));
			} else if (LANGUAGE.equalsIgnoreCase(name)) {
				setLocale(element.getValue());
			} else if (FONT_SIZE.equalsIgnoreCase(name)) {
				int size = Integer.parseInt(element.getValue());
				setFontSize(size);
			} else if ("ToolTip".equalsIgnoreCase(name)) {
				setToolTipState(Boolean.parseBoolean(element.getValue()));
			} else if (WEB_BROWSER.equalsIgnoreCase(name)) {
				put(WEB_BROWSER, element.getValue());
			} else if (TEXT_EDITOR.equalsIgnoreCase(name)) {
				put(TEXT_EDITOR, element.getValue());
			} else if ("Frame.Width".equalsIgnoreCase(name)) {
				Dimension dimension = getDimension(FRAME_SIZE);
				dimension.width = Integer.parseInt(element.getValue());
				putDimension(FRAME_SIZE, dimension);
			} else if ("Frame.Height".equalsIgnoreCase(name)) {
				Dimension dimension = getDimension(FRAME_SIZE);
				dimension.height = Integer.parseInt(element.getValue());
				putDimension(FRAME_SIZE, dimension);
			} else if (FRAME_MAXIMIZED.equalsIgnoreCase(name)) {
				putBoolean("Frame.Maximized", Boolean.parseBoolean(element.getValue()));
			} else if ("SplitH1.ratio".equalsIgnoreCase(name)) {
				putInt(SPLITPANE_H1_RATIO, Integer.parseInt(element.getValue()));
			} else if ("SplitH2.ratio".equalsIgnoreCase(name)) {
				putInt(SPLITPANE_H2_RATIO, Integer.parseInt(element.getValue()));
			} else if ("SplitV1.ratio".equalsIgnoreCase(name)) {
				putInt(SPLITPANE_V1_RATIO, Integer.parseInt(element.getValue()));
			} else if ("SplitV2.ratio".equalsIgnoreCase(name)) {
				putInt(SPLITPANE_V2_RATIO, Integer.parseInt(element.getValue()));
			} else if (USE_OPERATOR_LIBRARY.equalsIgnoreCase(name)) {
				putBoolean(USE_OPERATOR_LIBRARY, Boolean.parseBoolean(element.getValue()));
			} else if (TEMPORARY_FILE_EXTENSION.equalsIgnoreCase(name)) {
				_temporaryFileExtension = element.getValue();
			} else if (UNDO_HISTORY_SIZE.equalsIgnoreCase(name)) {
				putInt(UNDO_HISTORY_SIZE, Integer.parseInt(element.getValue()));
			} else if (CODE_GENERATOR_PREFERENCE.equalsIgnoreCase(name)) {
				putInt(CODE_GENERATOR_PREFERENCE, Integer.parseInt(element.getValue()));
			} else if (PREFERRED_FILE_DIRECTORY.equalsIgnoreCase(name)) {
				element.addChildren(new XMLElement(PREFERRED_FILE_DIRECTORY, get(PREFERRED_FILE_DIRECTORY)));
				put(PREFERRED_FILE_DIRECTORY, element.getValue());
			} else if (PREFERRED_BINARY_DIRECTORY.equalsIgnoreCase(name)) {
				element.addChildren(new XMLElement(PREFERRED_BINARY_DIRECTORY, get(PREFERRED_BINARY_DIRECTORY)));
				put(PREFERRED_BINARY_DIRECTORY, element.getValue());
			}
		}
	}

	private static XMLElement convertRecentFilesToXml() throws IOException {
		XMLElement element = new XMLElement("RecentFiles");
		for (String filename : Workspace.getRecentProjectFilenames()) {
			element.addChildren(new XMLElement("File", filename));
		}
		return element;
	}

	private static void convertXMLToRecentFiles( List<XMLElement> xmlTree ) throws IOException {
		Iterator<XMLElement> iterator = xmlTree.iterator();
		List<String> filenames = new ArrayList<String>();
		while (iterator.hasNext()) {
			XMLElement element = iterator.next();
			if ("File".equalsIgnoreCase(element.getName())) {
				filenames.add(element.getValue());
			}
		}
		Workspace.setRecentProjectFilenames(filenames);
	}
}