/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.item.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.item.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraph;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>MsdosGenerator</code> generates MsDos code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */ 
public final class MsdosGenerator extends CodeGenerator {
	private final List<String> _outputImages = new ArrayList<String>();

	protected MsdosGenerator( ) {
		// Only created by the code generator.
	}

	@Override
	protected void generateScriptHeader( String filename, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
		println("@echo off");
		println("rem @(#)" + filename + "				" + getCurrentDate());
		println("rem");
		println("rem This file was automatically generated by Ariane.");
		println("rem Ariane is part of the Pantheon Project.");
		println("rem");
		println("rem GREYC - ENSICAEN & University of Caen Basse-Normandie, France");
		println("");

		println("rem Check argument line consistency.");
		println("if \"%1\"==\"-h\" goto USAGE");
		for (int i = 0; i < arguments.size() + inputs.size() + outputs.size(); i++) {
			println("if \"%1\"==\"\" goto USAGE");
		}
		println("goto SCRIPT");

		print("@echo \"Usage: $0 ");
		for (int i = 0; i < arguments.size(); i++) {
			print(" " + arguments.get(i).getNickName());
		}
		for (int i = 0; i < inputs.size(); i++) {
			print(" input-file" + i);
		}
		for (int i = 0; i < outputs.size(); i++) {
			print(" output-file" + i);
		}
		println("\"");

		print("@echo \"  example: $0");
		for (int i = 0; i < arguments.size(); i++) {
			if (arguments.get(i).getFormula() != null && !arguments.get(i).getFormula().isEmpty()) {
				print(" " + arguments.get(i).getFormula());
			} else {
				print(" " + arguments.get(i).getNickName());
			}
		}
		for (int i = 0; i < inputs.size(); i++) {
			print(" " + inputs.get(i).getNickName());
		}
		for (int i = 0; i < outputs.size(); i++) {
			String name = outputs.get(i).getNickName();
			if ("saver".equals(name)) {
				name += outputs.get(i).getId();
			}
			print(" " + name);
		}
		println("\"");

		println("goto :END");
		println("\n:SCRIPT", 0);
	}

	@Override
	protected void generateScriptFooter( ) {
		println(":END");
	}

	@Override
	protected void generateUtilitySubroutinesDefinition( ) {
		// Do nothing
	}

	private String createImageName( int slotIndex, int itemIndex, int graphId ) {
		String name = UserPreferences.createTemporaryName(graphId, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	private String getInputImage( InputSlot slot, int graphId ) {
		return UserPreferences.createTemporaryName(graphId, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	private String variable( int graph, int item, int slot ) {
		return "var" + graph + "_" + item + "_" + slot;
	}

	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue();
		} else {
			return "%" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%";
		}
	}

	private String getNumericalParameter( GraphItem item, int slotIndex, int graphId ) {
		InputParameterSlot isp = (InputParameterSlot) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null) {
			return ((InputParameterSlot)item.getSlot("InSlotParameter", slotIndex)).getValue();
		} else {
			return "%" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%";
		}
	}

	private String translateLogicalFormula( String formula, String parameter0, String parameter1 ) {
		// Parse the boolean expression into Shell.
		char[] formulaCharacters = formula.toCharArray();
		String commandShell = "";
		int index = 0;
		while (index < formulaCharacters.length) {
			if (formulaCharacters[index] == '&') {
				commandShell += " && ";
				index++;
			} else if (formulaCharacters[index] == '|') {
				commandShell += " || ";
				index++;
			} else if (formulaCharacters[index] == '<') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " <= ";
					index += 2;
				} else {
					commandShell += " < ";
					index++;
				}
			} else if (formulaCharacters[index]  == '>') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " >= ";
					index += 2;
				} else {
					commandShell += " > ";
					index++;
				}
			} else if (formulaCharacters[index] == '=') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " == ";
					index += 2;
				} else {
					index++; // degenerated case!
				}
			} else if (formulaCharacters[index] == '!') {
				if (formulaCharacters[index + 1] == '=') {
					commandShell += " != ";
					index += 2;
				} else {
					commandShell += " ! ";
					index++;
				}
			} else if (formulaCharacters[index] == '(') {
				commandShell += " ";
				index++;
			} else if (formulaCharacters[index] == ')') {
				commandShell += " ";
				index++;
			} else if (formulaCharacters[index] == '$') {
				if (formulaCharacters[index + 1] == '0') {
					commandShell += parameter0;
				} else {
					commandShell += parameter1;
				}
				index += 2;
			} else {
				commandShell += formulaCharacters[index++];
			}
		}
		return commandShell;
	}

	@Override
	protected void generateMainPrologue( List<GraphItem> litems, List<GraphItem> arguments, List<GraphItem> inputs, List<GraphItem> outputs ) {
		println("\nrem\nrem The main program\n");
		println("del " + UserPreferences.getTemporaryFilePrefix() + "?*" + UserPreferences.getTemporaryFileExtension(), 1);
		for (GraphItem item : arguments ) {
			println("set " + variable(1, item.getId(), 0) + "=%1", 1);
			println("shift", 1);
		}
		for (GraphItem item : inputs ) {
			println("copy %1 " + createImageName(0, item.getId(), 1), 1);
			println("shift", 1);
		}
	}

	@Override
	protected void generateMainEpilogue( ) {
	}

	@Override
	protected void generateItem( LoaderItem item, int graphId, int indent ) throws ArianeBadGraphException {
		if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
			String parameters = getParameter(item, 0, graphId);
			println("copy " + parameters + " " + createImageName(0, item.getId(), graphId), indent);
		}
	}

	@Override
	protected void generateItem( SaverItem item, int graphId, int indent ) throws ArianeBadGraphException {
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
				String parameters = getParameter(item, 0, graphId);
				println("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " " + parameters, indent);
			} else  {
				println("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " %1", indent);
				println("shift", indent);
			}
			println("copy "
					+ getInputImage(item.getInputDataSlots().get(0), graphId)
					+ " "
					+ createImageName(0, item.getId(), graphId), indent);
		}
	}

	@Override
	protected void generateItem( ViewerItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String commandShell = item.getViewerCommandName();
		if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
			println(commandShell + " " + getInputImage(item.getInputDataSlots().get(0), graphId), indent);
			println("copy "
					+ getInputImage(item.getInputDataSlots().get(0), graphId)
					+ " "
					+ createImageName(0, item.getId(), graphId), indent);
		}
	}

	@Override
	protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new ArianeBadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		List<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphId));
		}

		// Get mask
		List<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
			maskLine.add(getInputImage(mask, graphId));
		}

		// Get input images.
		List<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, graphId));
		}

		// Get output images.
		List<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), graphId));
		}

		String code = "";
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for (String command : commandLine) {
			code += command + " ";
		}

		// Get output result.
		println(code, indent);
		println("call " + UserPreferences.getCommandStatus(), indent);
		println("call pset " + variable(graphId, item.getId(), 0), indent);
	}

	@Override
	protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName == null) {
			throw new ArianeBadGraphException(Ariane.getString("Generator.error2") + item.getName());
		}

		// Get parameters
		LinkedList<String> parameterLine = new LinkedList<String>();
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameterLine.add(getParameter(item, i, graphId));
		}

		// Get mask
		LinkedList<String> maskLine = new LinkedList<String>();
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
			maskLine.add(getInputImage(mask, graphId));
		}

		// Get input images.
		LinkedList<String> inputLine = new LinkedList<String>();
		for (InputDataSlot image : item.getInputDataSlots() ) {
			if (image.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			inputLine.add(getInputImage(image, graphId));
		}

		// Get output images.
		LinkedList<String> outputLine = new LinkedList<String>();
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			outputLine.add(createImageName(i, item.getId(), graphId));
		}

		String code = "";
		List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for (String command : commandLine) {
			code += command + " ";
		}

		// Get output result.
		println(code, indent);
		println("call " + UserPreferences.getCommandStatus(), indent);
		println("call pset " + variable(graphId, item.getId(), 0), indent);
	}

	private String parseFormula( String formula, String[] parameters ) {
		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("e".equals(etape) || "E".equals(etape)) {
				code += "*10^";
			} else if ("cos".equals(etape)) {
				code += "c";
			} else if ("sin".equals(etape)) {
				code += "s";
			} else if ("ln".equals(etape)) {
				code += "l";
			} else if ("exp".equals(etape)) {
				code += "e";
			} else if ("atan".equals(etape)) {
				code += "a";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += parameters[numberOfParameters];
			} else {
				code += etape;
			}
		}
		return code;
	}

	private String setFunction( String formula ) {
		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		boolean abs = false;
		boolean max = false;
		boolean min = false;
		boolean tan = false;
		boolean log2 = false;
		boolean log10 = false;
		boolean acos = false;
		boolean asin = false;

		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("abs".equals(etape) && !abs) {
				code += " define abs(x){; if (x>=0) return(x) else return (-x)} ";
				abs = true;
			} else if ("max".equals(etape) && !max) {
				code += " define max(x,y){; if (x>y) return(x) else return(y)} ";
				max = true;
			} else if ("min".equals(etape) && !min) {
				code += " define min(x,y){; if (x<y) return(x) else return(y)} ";
				min = true;
			} else if ("tan".equals(etape) && !tan) {
				code += " define tan(x){; return (s(x)/c(x)} ";
				tan = true;
			} else if ("acos".equals(etape) && !acos) {
				code += " define acos(x){; if (x==1) return(0) if (x==-1) return (4*a(1)) return (2*a(1)-a(x/sqrt(1-(x^2))))} ";
				acos = true;
			} else if ("asin".equals(etape) && !asin) {
				code += " define asin(x){; if (x==1) return(2*a(1)) if (x==-1) return (-2*a(1)) return (a(x/sqrt(1-(x^2))))} ";
				asin = true;
			} else if ("log".equals(etape)) {
				etape = stringTokenizer.nextToken();
				if ("2".equals(etape) && !log2) {
					code += " define log2(x){; return l(x)/l(2)} ";
					log2 = true;
				} else if ("1".equals(etape) && !log10) {
					code += " define log10(x){; return l(x)/l(10)} ";
					log10 = true;
				}
			}
		}
		return code;
	}

	@Override
	protected void generateItem( EvalItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphId);
		}

		println("set " + variable(graphId, item.getId(), 0) + "=" + "$(echo \""
				+ setFunction(item.getFormula())
				+ parseFormula(item.getFormula(), parameters)
				+ "\"|bc -l);", indent);
	}

	@Override
	protected void generateItem( FormatItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String[] parameters = new String[item.getInputParameterSlots().size()];
		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getParameter(item, i, graphId);
		}

		String code = "";
		StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
		String etape;
		while (stringTokenizer.hasMoreTokens()) {
			etape = stringTokenizer.nextToken();
			if ("+".equals(etape)) {
				code += "";
			} else if ("quote".equals(etape)) {
				code += "";
			} else if ("\"".equals(etape)) {
				code += "";
			} else if ("(".equals(etape)) {
				code += "\'";
			} else if (")".equals(etape)) {
				code += "\'";
			} else if ("$".equals(etape)) {
				int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
				code += parameters[numberOfParameters];
			} else {
				code += etape;
			}
		}
		println(variable(graphId, item.getId(), 0) + "=\"" + code + "\"", indent);
	}

	@Override
	protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String value = getParameter(item, 0, graphId);
		println("rem Select", indent);
		if (value == null) {
			value = variable(graphId, item.getId(), 0);
			println("read " + value, indent);
			value = "%" + value + "%";
		}
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			println("if [ " + value + " -eq " + i + " ]", indent);
			println("then", indent);
			println("copy "
					+ getInputImage(item.getInputDataSlots().get(i), graphId) + " "
					+ createImageName(0, item.getId(), graphId), indent + 1);
			println("set " + variable(graphId, item.getId(), 0) + "=" + i + ";", indent + 1);
			println("fi", indent);
		}
	}

	@Override
	protected void generateItem( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String[] parameters = new String[2];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			parameters[i] = getNumericalParameter(item, i, graphId);
		}
		String commandShell = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
		println("if [ $( echo \"" + commandShell + "\" |bc -l) -gt 0 ]", indent);
	}

	@Override
	protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		println("then", indent);
		print("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " "
				+ createImageName(0, item.getId(), graphId), indent + 1);
		println("; " + variable(graphId, item.getId(), 0) + "=0");
	}

	@Override
	protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		println("else", indent);
		print("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " "
				+ createImageName(1, item.getId(), graphId), indent + 1);
		println("; " + variable(graphId, item.getId(), 0) + "=1");
	}

	@Override
	protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		println("fi", indent);
	}

	@Override
	protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
		// input image
		InputDataSlot in0;
		in0 = item.getInputDataSlots().get(0);
		int id0;
		id0 = in0.getConnectedSlot().getGraphItem().getId();
		print("if [ -n \"%" + variable(graphId, id0, 0) + "%\" -a ", indent);
		println("-r " + getInputImage(item.getInputDataSlots().get(0), graphId) + " ]");
		println("then", indent);
		print("copy "
				+ getInputImage(item.getInputDataSlots().get(0), graphId)
				+ " "
				+ createImageName(0, item.getId(), graphId), indent + 1);
		println("; " + variable(graphId, item.getId(), 0) + "=0;");
		println("del " + getInputImage(item.getInputDataSlots().get(0), graphId), indent + 1);
		println("else", indent);
		print("copy "
				+ getInputImage(item.getInputDataSlots().get(1), graphId) + " "
				+ createImageName(0, item.getId(), graphId), indent + 1);
		println("; " + variable(graphId, item.getId(), 0) + "=1;");
		println("del " + getInputImage(item.getInputDataSlots().get(1), graphId), indent + 1);
		println("fi", indent);
	}

	@Override
	protected void generateItem( FileNameItem item, int graphId, int indent ) throws ArianeBadGraphException {
		if (!item.isArgumentOfRelatedScript()) {
			println("set " + variable(graphId, item.getId(), 0) + "='" + item.getFormula() + "'", indent);
		}
	}


	@Override
	protected void generateItem( ConstantItem item, int graphId, int indent ) throws ArianeBadGraphException {
		if (!item.isArgumentOfRelatedScript() ) {
			String value = item.getFormula().replaceAll("\"", "\\\\\"");
			println("set " + variable(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
		}
	}

	@Override
	protected void generateItem( ScriptItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String value = item.getFormula().replaceAll("\"", "\\\\\"").replaceAll("\n", "\\\\\n");
		println("set " + variable(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
	}

	@Override
	protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
		String code = "macro" + item.getId() + "_" + item.getGraph().getIdentifier();

		// Get parameters
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			code += " " + getParameter(item, i, graphId);
		}

		// Get mask
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
			code += " " + getInputImage(mask, graphId);
		}

		// Get input images.
		for (InputDataSlot image : item.getInputDataSlots()) {
			if (image.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			code += " " + getInputImage(image, graphId);
		}

		// Get output images.
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			code += " " + createImageName(i, item.getId(), graphId);
		}
		// Get output result.
		println(code);
		println("call " + UserPreferences.getCommandStatus(), indent);
		println("call pset " + variable(graphId, item.getId(), 0), indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that contains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\nrem ");
	}

	private void generateComment( GraphItem item, List<GraphItem> litems ) {
		println("\nrem\nrem " + item.getNickName());
		for (GraphItem note : litems) {
			println("rem " + parseComment(note.getFormula()));
		}
		println("rem");
	}

	@Override
	protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		generateComment(item, subGraph.getListOfGraphItems());

		println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters
		int subgraphId = subGraph.getIdentifier();
		println("# Parameters", 1);
		println("begin=%1; shift", 1);
		println("end=%1; shift", 1);
		println("step=%1; shift", 1);
		for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
			println("set " + variable(subgraphId, 0, i - 2) + "=%1; shift", 1);
		}

		// Get mask
		println("# Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
			println("ln -fs %1 " + createImageName(maskIndex, 0, subgraphId) + "; shift", 1);
		}

		// Get input images
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			println("copy %1 " + createImageName(i, 0, subgraphId) + "; shift", 1);
		}
		println("");
		// Create the output images in case of no iteration
		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			println("touch " + getInputImage(isi, subgraphId), 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			println(variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
		}

		println("");
		println("local i", 1);
		println("for ((i=%begin%; i<=%end%; i+=%step% ))", 1);
		println("do", 1);
		println(variable(subgraphId, 0, 0) + "=%i;", 2);
		generateSubroutineBody(litems, subgraphId, 2);

		// Set intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			String input = getInputImage(isi, subgraphId);
			String output = createImageName(i, 0, subgraphId);
			if (!input.equals(output)) {
				println("copy " + input + " " + output, 2);
			}
		}

		println("done", 1);

		println("");
		// Get output images
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
			println("shift", 1);
		}
		// Output result
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			println("psetstatus SUCCESS", 1);
		} else {
			println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
		}
		println("}");

		generateSubroutinesDefinition(litems);
	}

	@Override
	protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

		// Get and discard the virtual operator
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		generateComment(item, subGraph.getListOfGraphItems());

		println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters
		int subgraphId = subGraph.getIdentifier();
		println("rem Parameters", 1);
		String[] parameters = new String[item.getInputParameterSlots().size()];
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			println("set " + variable(subgraphId, 0, i) + "=%1", 1);
			println("shift", 1);
			parameters[i] = "%" + variable(subgraphId, 0, i) + "%";
		}

		// Get mask.
		println("rem Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
			println("copy %1 " + createImageName(maskIndex, 0, subgraphId), 1);
			println("shift", 1);
		}

		// Get input images.
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			println("copy %1 " + createImageName(i, 0, subgraphId), 1);
			println("shift", 1);
		}

		// Parse the boolean expression.
		String testFormula  = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
		VirtualItem virtualItem = (VirtualItem)virtualOperator;

		println("");
		// Create output images in case of no repetition
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			println("touch " + getInputImage(isi, subgraphId), 1);
		}

		// Initialize the return variable in case of no iteration
		{
			InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
			println(variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
		}

		println("");
		println("while [ $( echo \"" + testFormula + "\" |bc -l) -gt 0 ]", 1);
		println("do", 1);
		
		generateSubroutineBody(litems, subgraphId, 2);

		// Get intermediate images
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			// Discard cp image image (it happens with the output = input)
			String input = getInputImage(isi, subgraphId);
			String output = createImageName(i, 0, subgraphId);
			if (!input.equals(output)) {
				println("copy " + input + " " + output, 2);
			}
		}

		// Get intermediate parameters
		for (int i = 0; i < 2; i++) {
			InputParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
			if (isp.getConnectedSlot() != null) {
				println(variable(subgraphId, 0, i) + "=%" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 2);
			}
		}
		println("done", 1);
		// Get output images
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
			println("shift", 1);
		}

		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			println("psetstatus SUCCESS", 1);
		} else {
			println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
		}
		println("}");

		generateSubroutinesDefinition(litems);
	}

	@Override
	protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
		SubGraph subGraph = item.getSubGraph();
		List<GraphItem> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

		// Get and discard the virtual operator.
		Object virtualOperator = getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.isEmpty()) {
			throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
		}

		generateComment(item, subGraph.getListOfGraphItems());

		println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

		// Get parameters.
		int subgraphId = subGraph.getIdentifier();
		println("rem arguments", 1);
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			println("set " + variable(subgraphId, 0, i) + "=%1", 1);
			println("shift", 1);
		}

		// Get mask.
		println("rem Input data files", 1);
		InputMaskSlot mask = item.getInputMaskSlot();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
			int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
			println("copy %1 " + createImageName(maskIndex, 0, subgraphId), 1);
			println("shift", 1);
		}

		// Get input images.
		List<InputDataSlot> inputImages = item.getInputDataSlots();
		for (int i = 0; i < inputImages.size(); i++) {
			InputDataSlot isi = inputImages.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new ArianeBadGraphException(item.getNickName() + ": " + Ariane.getString("Generator.error.incomplete"));
			}
			println("copy %1 " + createImageName(i, 0, subgraphId), 1);
			println("shift", 1);
		}

		println("");

		generateSubroutineBody(litems, subgraphId, 1);

		VirtualItem virtualItem = (VirtualItem)virtualOperator;
		// Get output images.
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
			println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
			println("shift", 1);
		}

		println("");
		// Output result.
		InputParameterSlot isp = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		if (isp.getConnectedSlot() == null) {
			println("psetstatus SUCCESS", 1);
		} else {
			println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
		}
		println("}");

		generateSubroutinesDefinition(litems);
	}
}