/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.ariane.presenter.Graph;

/**
 * A graph stores a program as an oriented graph of items.
 */
public final class GraphModel {
	private final int _identifier;

	private static int _numberOfCreatedGraphs;
	private int _numberOfCreatedItems;

	private final List<GraphItem> _graphItems = new LinkedList<GraphItem>();
	private File _file;

	public GraphModel( ) {
		_identifier = ++_numberOfCreatedGraphs;
	}

	public int getIdentifier( ) {
		return _identifier;
	}

	public int getNextItemIdentifier( ) {
		return ++_numberOfCreatedItems;
	}

	public void dispose() {
		for (GraphItem item : _graphItems) {
			item.dispose();
		}
		_graphItems.clear();
	}

	public List<GraphItem> getListOfGraphItems() {
		return _graphItems;
	}

	public void setFile( File absoluteFile ) {
		_file = absoluteFile;
	}

	public File getFile( ) {
		return _file;
	}

	public boolean isFirstSave( ) {
		return _file == null;
	}

	public void add( GraphItem item ) {
		_graphItems.add(item);
	}

	public void remove( GraphItem item ) {
		_graphItems.remove(item);
	}

	public void resetAll( ) {
		for (GraphItem item : _graphItems) {
			item.resetAll();
		}
	}

	public void reset( ) {
		for (GraphItem item : _graphItems) {
			item.reset();
		}
	}

	public GraphItem getGraphItem( int identifier ) {
		for (GraphItem item :_graphItems) {
			if (item.isInSameReadPasteSession() && item.getInternalId() == identifier) {
				return item;
			}
		}
		return null;
	}

	public void save( File file, Graph graph ) throws IOException {
		XMLFileOutputStream out = new XMLFileOutputStream(file, null);
		out.write(convertToXMLTree(file, graph));
		out.close();
		_file = file;
	}

	public XMLElement convertToXMLTree( File file, Graph graph ) throws IOException {
		XMLElement element;
		if (graph.isRoot()) {
			element = new XMLElement("Graph");
		} else {
			element = new XMLElement("SubGraph");
		}
		XMLElement items = new XMLElement("ItemList");
		for (GraphItem item : _graphItems) {
			items.addChildren(item.convertToXML(file));
		}
		XMLElement connections = new XMLElement("ConnectionList");
		for (GraphItem item : _graphItems) {
			connections.addChildren(item.convertSlotsToXML());
		}
		element.addChildren(items);
		element.addChildren(connections);
		return element;
	}

	public void createFromXMLTree( List<XMLElement> xmlTree, Graph graph ) throws IOException {
		List<XMLElement> xmlSubTree = null;
		boolean found = false;
		if (graph.isRoot()) {
			xmlSubTree = xmlTree.get(0).getChildren();
			String tagName = xmlSubTree.get(0).getName();
			if ("Graph".equalsIgnoreCase(tagName) || "Graphe".equalsIgnoreCase(tagName)) {
				xmlSubTree = xmlSubTree.get(0).getChildren();
				found = true;
			}
		} else {
			for (XMLElement element : xmlTree) {
				String tagName = element.getName();
				if ("SubGraph".equalsIgnoreCase(tagName) || "SousGraphe".equalsIgnoreCase(tagName)) {
					xmlSubTree = element.getChildren();
					found = true;
					break;
				}
			}
		}

		// Error incorrect file?
		if (!found) {
			throw new IOException("Incorrect XML file.");
		}
		for (XMLElement element : xmlSubTree) {
			String tagName = element.getName();
			if ("ItemList".equalsIgnoreCase(tagName) || "ListeElements".equalsIgnoreCase(tagName)) {
				GraphItem.createFromXML(graph, element.getChildren());
			} else if ("ConnectionList".equalsIgnoreCase(tagName) || "ListeConnexions".equalsIgnoreCase(tagName)) {
				Slot.createFromXML(graph, element.getChildren());
			}
		}
	}

	public void synchronizeItemFilesWithSystemFiles() {
		for (GraphItem item : _graphItems) {
			if (item instanceof LoaderItem) {
				((LoaderItem)item).synchronizeWithFileSystem();
			} else if (item instanceof FileNameItem) {
				((FileNameItem)item).synchronizeWithFileSystem();
			}
		}
	}
}
