/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph;

import java.io.File;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.swing.SwingUtilities;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.graph.item.DataFile;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;

/**
 * A <code>ThreadDirectoryExecution</code> controls the execution of a graph
 * for a whole directory.
 */
public final class ThreadDirectoryExecution extends Thread {
	private GraphItem[] _itemToExecute;
	private final LoaderItem _loader;
	private final File[] _files;
	private final Map<SaverItem, File> _saverDirectories;
	private final File _loaderFile;

	public ThreadDirectoryExecution( List<GraphItem> selectedItems, LoaderItem loader, File loaderFile, File[] files, Map<SaverItem, File> saverDirectories) {
		// We use a copy of the selectItems since user can modify the list
		// while the program is executing it.
		_itemToExecute = selectedItems.toArray(new GraphItem[selectedItems.size()]);
		_loader = loader;
		_files = Arrays.copyOf(files, files.length);
		_loaderFile = loaderFile;
		_saverDirectories = saverDirectories;
	}

	@Override
	public void run( ) {
		Ariane.getPresenter().setExecution(true);
		for (File file : _files) {
			final String basename = getBasename(file.getName());
			_loader.setBinary(file.getAbsolutePath());
			updateItemName(file.getName());
			for (SaverItem item : _saverDirectories.keySet()) {
				String suffix = getSuffix(_saverDirectories.get(item).getName());
				item.setBinary(_saverDirectories.get(item).getParent() + File.separator + basename + suffix);
			}
			_loader.resetAll();
			for (GraphItem item: _itemToExecute ) {
				item.execute();
			}
		}
		_itemToExecute = null;
		Ariane.getPresenter().setExecution(false);

		// Restore old file name to the loader and the savers (modified the execution).
		_loader.setBinary(_loaderFile.getAbsolutePath());

		updateItemName(DataFile.getFileNameWithoutExtension(_loaderFile.getAbsolutePath()));

		for (SaverItem item : _saverDirectories.keySet()) {
			item.setBinary(_saverDirectories.get(item).getAbsolutePath());
		}
		_loader.resetAll();
	}

	private String getSuffix( String filename ) {
		int suffixPosition1 = filename.lastIndexOf('.');
		if (suffixPosition1 > 0) {
			return filename.substring(suffixPosition1);
		} else {
			return "";
		}
	}

	private String getBasename( String filename ) {
		int suffixPosition = filename.lastIndexOf('.');
		if (suffixPosition > 0) {
			return filename.substring(0, suffixPosition);
		} else {
			return filename;
		}
	}

	private void updateItemName( final String name ) {
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				_loader.setTemporaryNickName(DataFile.getFileNameWithoutExtension(name));
				_loader.getGraph().revalidate();
			}
		});
	}
}