/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.Ariane.ErrorCode;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingResultException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.view.graph.item.CompositeItemView;
import fr.greyc.pantheon.ariane.view.graph.item.GraphItemView;
import fr.greyc.pantheon.ariane.view.graph.item.CompositeItemView.StateName;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;

/**
 * An <code>CompositeItem</code> represents an item with
 * an inner subgraph.
 */
public abstract class CompositeItem extends ControlItem {
	private SubGraph _subGraph;

	protected CompositeItem( GraphItemView graphItemView, TreeItem treeItem, Graph graph, String name ) {
		super(graphItemView, treeItem, graph, name);
		((CompositeItemView)getView()).changeState(StateName.Editable);
		createSlots(0, 0, 1, 0, 0);
	}

	protected CompositeItem( GraphItemView graphItemView, Graph graph ) {
		super(graphItemView, graph);
		((CompositeItemView)getView()).changeState(StateName.Editable);
		createSlots(0, 0, 1, 0, 0);
	}

	public CompositeItem( GraphItemView graphItemView, Graph graph, TreeItem treeItem ) {
		super(graphItemView, graph);
		setRelatedTreeItem(treeItem);
		createSlots(0, 0, 1, 0, 0);
		((CompositeItemView)getView()).changeState(StateName.Uneditable);
	}

	protected final void setPrototype( String name, int inputImageCount, int imageParameterCount, int outputImageCount, int interImageCount, int interParameterCount ) {
		setNickName(name);
		adjustSlotsNumber(inputImageCount, imageParameterCount, outputImageCount);
		createSubGraph(name, inputImageCount, imageParameterCount, outputImageCount);
	}

	private void adjustSlotsNumber( int inputImageCount, int inputParameterCount, int outputImageCount ) {
		List<InputDataSlot> inputDataSlots = getInputDataSlots();
		if (inputImageCount > inputDataSlots.size()) {
			for (int i = inputDataSlots.size(); i < inputImageCount; i++) {
				inputDataSlots.add(new InputDataSlot(this, i));
			}
		}
		List<InputParameterSlot> inputParameterSlots = getInputParameterSlots();
		if (inputParameterCount > inputParameterSlots.size()) {
			for (int i = inputParameterSlots.size(); i < inputParameterCount; i++) {
				inputParameterSlots.add(new InputParameterSlot(this, i));
			}
		}
		List<OutputDataSlot> outputDataSlots = getOutputDataSlots();
		if (outputImageCount > outputDataSlots.size()) {
			for (int i = outputDataSlots.size(); i < outputImageCount; i++) {
				outputDataSlots.add(new OutputDataSlot(this, i));
			}
		}
	}

	protected final SubGraph createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount, int interImageCount, int interParameterCount) {
		_subGraph = new SubGraph(this, name, inputImageCount, inputParameterCount, outputImageCount, interImageCount, interParameterCount);
		return _subGraph;
	}

	protected abstract SubGraph createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount );

	@Override
	public final void setNickName( String name ) {
		super.setNickName(name);
		Ariane.getPresenter().renameGraph(_subGraph, name);
	}

	@Override
	public final void removeHook() {
		_subGraph.closeFrame();
	}

	@Override
	protected final void disposeHook() {
		_subGraph.dispose();
		_subGraph = null;
	}

	protected abstract boolean linkItemWithVirtualItem( VirtualItem virtualtem ) throws ArianeException;

	///// 
	///// A C C E S S O R S
	/////

	@Override
	public final SubGraph getSubGraph() {
		return _subGraph;
	}

	public final void viewSubgraph() {
		_subGraph.showContent();
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public final void resetAll( ) {
		super.resetAll();
		if (_subGraph != null) {
			_subGraph.resetAll();
		}
	}

	@Override
	protected final void resetHook() {
		_subGraph.reset();
		resetHookHook();
	}

	protected void resetHookHook() {
		// Empty
	}

	protected final VirtualItem getVirtualItem() {
		return _subGraph.getVirtualItem();
	}

	protected final boolean getInputDataFiles( VirtualItem virtualItem ) throws ArianeMissingInputException {
		Iterator<InputDataSlot> inputDataSlots = getInputDataSlots().iterator();
		Iterator<OutputDataSlot> outputDataSlots = virtualItem.getVirtualInputDataSlots().iterator();
		while (inputDataSlots.hasNext()) {
			InputDataSlot inputDataSlot = inputDataSlots.next();
			if (inputDataSlot.hasValue(true)) {
				OutputDataSlot outputdataSlot = outputDataSlots.next();
				outputdataSlot.setValue(inputDataSlot.getValue());
			} else {
				return false;
			}
		}
		return true;
	}

	protected final void getMask( VirtualItem virtualItem ) throws ArianeRuntimeException {
		InputMaskSlot inputMask = getInputMaskSlot();
		// Note: Mask is optional
		if (inputMask.isConnected() && inputMask.hasValue(false)) {
			OutputDataSlot outputDataSlot = (OutputDataSlot)virtualItem.getVirtualInputMaskSlot();
			outputDataSlot.setValue(inputMask.getValue());
		}
	}

	protected final boolean setOutputResult( VirtualItem virtualItem ) throws ArianeMissingResultException {
		InputParameterSlot inputParameterSlot = (InputParameterSlot)virtualItem.getVirtualOutputResultSlot();
		try {
			if (!inputParameterSlot.hasValue(true)) {
				return false;
			}
			OutputResultSlot outputResultSlot = getOutputResultSlot();
			outputResultSlot.setValue(inputParameterSlot.getValue());
			return true;
		} catch (ArianeMissingParameterException e) {
			throw changeExceptionType(e, "Operator.error8");
		}
	}

	protected final void setOutputData( VirtualItem virtualItem ) throws ArianeMissingResultException {
		Iterator<OutputDataSlot> inputDataSlots = getOutputDataSlots().iterator();
		Iterator<InputDataSlot> outputDataSlots = virtualItem.getVirtualOutputDataSlots().iterator();
		while (inputDataSlots.hasNext()) {
			InputDataSlot inputDataSlot = outputDataSlots.next();
			try {
				if (inputDataSlot.hasValue(true)) {
					OutputDataSlot outputDataSlot = inputDataSlots.next();
					outputDataSlot.setValue(inputDataSlot.getValue());
				} else {
					throw new ArianeMissingResultException(Ariane.getString("Operator.error9"));
				}
			} catch (ArianeMissingInputException e) {
				throw changeExceptionType(e, "Operator.error9");
			}
		}
	}

	protected final ArianeMissingResultException changeExceptionType( ArianeException exception, String message ) {
		ArianeMissingResultException newException = new ArianeMissingResultException(Ariane.getString(message));
		newException.initCause(exception);
		return newException;
	}

	protected final ArianeRuntimeException displayError( Exception exception , ErrorCode error ) {
		if (Graph.isExecutionRunning()) {
			Ariane.getPresenter().printErrorMessage(error, getAbsoluteName(), getId(), exception.getLocalizedMessage());
			getGraph().stopExecution();
			changeStatus(Status.ERROR);
		} else {
			changeStatus(Status.UNEXECUTED);
		}
		return new ArianeRuntimeException(exception);
	}

	///// 
	///// I N P U T   /   O U T P U T
	/////

	@Override
	public final XMLElement convertToXML( File file ) throws IOException {
		XMLElement element = new XMLElement(getTagName());
		List<XMLElement> xmlTree = convertPropertiesToXML(file);
		if (getRelatedTreeItem() instanceof SubroutineTreeItem && ((CompositeItemView)getView()).isUneditable()) {
			xmlTree.add(new XMLElement("SubroutineReference", getRelatedTreeItem().getName()));
		}
		element.addChildren(xmlTree);
		element.addChildren(_subGraph.convertToXMLTree(file));
		return element;
	}

	@Override
	protected final void createAdditionalPropertyFromXML( List<XMLElement> xmlTree ) throws IOException {
		boolean subroutine = false;
		for (XMLElement element: xmlTree) {
			if ("SubroutineReference".equalsIgnoreCase(element.getName())) {
				subroutine = getRelatedTreeItem() instanceof SubroutineTreeItem;
				break;
			}
		}
		SubGraph subGraph = createSubGraph(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
		if (subroutine) {
			((CompositeItemView)getView()).changeState(StateName.Uneditable);
			subGraph.createFromXMLTree(((SubroutineTreeItem)getRelatedTreeItem()).getBody());
		} else {
			subGraph.createFromXMLTree(xmlTree);
		}
	}

	public final void changeEditableStatus( boolean editable ) {
		((CompositeItemView)getView()).changeEditableStatus(editable);
	}

	public abstract TreeItem restoreDefaultTreeItem();

	protected abstract void askProperties() throws InterruptedException;

	protected final void executeBody( VirtualItem virtualItem ) throws ArianeMissingResultException {
		for (int k = 0; k < getInputDataSlots().size(); k++) {
			try {
				InputDataSlot intermediateInputSlot = virtualItem.getVirtualIntermediateDataSlots(k);
				intermediateInputSlot.hasValue(true);
			} catch (ArianeMissingInputException e) {
				throw changeExceptionType(e, "Operator.error10");
			}
		}
	}

	protected final void setIntermediateInputs( VirtualItem virtualItem ) throws ArianeMissingInputException, IOException {
		for (int k = 0; k < getInputDataSlots().size(); k++) {
			InputDataSlot intermediateInputSlot = virtualItem.getVirtualIntermediateDataSlots(k);
			if (intermediateInputSlot.hasValue(true)) {
				OutputDataSlot inputSlot = virtualItem.getVirtualInputDataSlot(k);
				DataFile.copyFile(intermediateInputSlot.getValue(), inputSlot.getTempName());
				inputSlot.setValue();
			}
		}
	}
}