/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.ConstantItemView;

import java.io.IOException;
import java.util.List;

/**
 * A <code>ConstantItem</code> implements a text value that represents
 * a constant.
 */
public final class ConstantItem extends ControlItem implements IScriptArgument {
	public static final String TAGNAME = "Constant";
	private boolean _isArgument;

	public ConstantItem( TreeItem treeItem, Graph graph ) {
		super(new ConstantItemView(), treeItem, graph, treeItem.getName());
		setFormula(""); // IMPERATIVE to signify that formula is used as mandatory property.
		setComment(getFormula());
	}

	public ConstantItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(new ConstantItemView(), graph);
		setFormula(""); // IMPERATIVE to signify that formula is used as mandatory property.
		createFromXML(elements);
		setComment(getFormula());
	}

	@Override
	public boolean isArgumentOfRelatedScript() {
		return _isArgument;
	}

	@Override
	public void setArgumentOfRelatedScript( boolean value ) {
		_isArgument = value;
	}

	@Override
	public void setPrototype( String name, int inputImageCount, int inputParameterCount, int outputImageCount ) {
		// Empty
	}

	@Override
	public boolean changeFormula( String statement ) {
		if (!getFormula().equals(statement)) {
			setFormula(statement);
			setComment(statement);
		}
		return true;
	}

	///// 
	///// A C C E S S O R S
	/////

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			final String constantExpression = getFormula();
			if (constantExpression.isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error19"));
			}
			getOutputResultSlot().setValue(constantExpression);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeArithmeticException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
	}
}