/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.DataFileView;
import fr.greyc.pantheon.ariane.view.graph.item.GraphItemView;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.List;

import javax.swing.JTextField;

/**
 * A <code>DataFile</code> is either a loader, a saver or a viewer
 * of a data files.
 */
public abstract class DataFile extends GraphItem {
	private String _currentFilename;

	protected DataFile( GraphItemView graphItemView, TreeItem treeItem, Graph graph, String name, String binary ) {
		super(graphItemView, treeItem, graph, name, null);
		setBinary(binary);
	}

	protected DataFile( GraphItemView graphItemView, Graph graph ) throws IOException {
		super(graphItemView, graph);
	}

	/**
	 * Creates a <code>Loader</code> or a <code>Saver</code>
	 * on the given graph with the given XML data.
	 * Notice: Loader has 1 input image and 0 output image and
	 * saver has 0 input image and 1 output image.
	 * !!!! This method is kept for sake of compatibility.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @return the item.
	 * @throws IOException XML errors.
	 */
	public static final DataFile getInstance(Graph graph, List<XMLElement> data ) throws IOException {
		int inputDataSlots = 0;
		for (XMLElement element : data) {
			if ("ImagesEntree".equalsIgnoreCase(element.getName())) {
				inputDataSlots = Integer.parseInt(element.getValue());
			}
		}
		if (inputDataSlots > 0) {
			return new SaverItem(graph, data);
		}
		return new LoaderItem(graph, data);
	}

	public static String getFileNameWithoutExtension( String binary ) {
		return binary.substring(binary.lastIndexOf(File.separator) + 1);
	}

	public final void setCurrentFilename( String filename ) {
		_currentFilename = filename;
	}

	public final String getCurrentFilename( ) {
		return _currentFilename;
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 * 
	 * @param filename a file name (absolute or relative).
	 * @param current_directory the name of the current directory or null if the related graph file has no name.
	 */
	public static final String setPathToFilename( String filename, String currentDirectory ) {
		if (filename == null || filename.length() == 1) {
			return "";
		}
		String binary = filename;
		// relative name
		if (binary.charAt(0) == '.') {
			binary = binary.substring(2);
			File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
			return file.getAbsolutePath();
		} else if (binary.startsWith("$FILE_DIRECTORY")) {
			String fileDirectory = UserPreferences.getPreferredFileDirectory();
			binary = binary.substring("$FILE_DIRECTORY".length() + 1);
			File file = new File(fileDirectory, binary.replaceAll("\\\\", "/"));
			return file.getAbsolutePath();
		}
		// absolute name
		File binaryFilename = new File(binary);
		if (binaryFilename.exists()) {
			try {
				return binaryFilename.getCanonicalPath();
			} catch (IOException e) {
			}
		}
		// Keep for the sake of compatibility
		File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
		if (file.exists()) {
			return file.getAbsolutePath();
		}
		// default
		return binaryFilename.getAbsolutePath();
	}

	/**
	 * Returns the name of the file either
	 * with relative path if the file is in the filePath directory
	 * or with an absolute path otherwise.
	 * 
	 * @param dataFileName the name of a data file.
	 * @param graphDirectoryPath the path to the graph directory.
	 * @param baseDirectoryPath the path to the specified base directory.
	 * @return the name.
	 */
	public static final String getRelativePath( String dataFileName, String graphDirectoryPath, String baseDirectoryPath ) {
		// Case of simple cut-copy/paste or no filename or filename with relative path.
		if (dataFileName == null || graphDirectoryPath == null || dataFileName.trim().isEmpty() || dataFileName.trim().charAt(0) != File.separatorChar) {
			return dataFileName;
		}

		File file = new File(dataFileName);
		String filePath = file.getParent();

		if (baseDirectoryPath != null && filePath != null && filePath.startsWith(baseDirectoryPath)) {
			// Use base directory path.
			String subDirectory = filePath.substring(baseDirectoryPath.length());
			if (!subDirectory.isEmpty()) {
				subDirectory = File.separator + subDirectory;
			}
			return "$FILE_DIRECTORY" + subDirectory + File.separator + file.getName();
		} else {
			// Use related path if the file is in the directory, else use the absolute name.
			File directoryPath = new File(graphDirectoryPath);
			if (directoryPath != null && filePath != null && filePath.startsWith(graphDirectoryPath)) {
				return "." + filePath.substring(graphDirectoryPath.length()) + File.separator + file.getName();
			}
		}
		return file.getAbsolutePath();
	}

	@Override
	public final String toString( ) {
		if (getBinary() == null) {
			return "(" + getId() + "):" + getName();
		} else {
			return "(" + getId() + "):" + getBinary();
		}
	}

	public static final void copyFile( String from, String to ) throws IOException {
		if (from.equals(to)) {
			return;
		}
		File src = new File(from);
		File dst = new File(to);
		Files.copy(src.toPath(), dst.toPath(), StandardCopyOption.REPLACE_EXISTING);
	}

	///// 
	///// E X E C U T I O N 
	/////

	public abstract String getViewerCommand();

	public final void changeBinaryName( JTextField textfield1, String binary, boolean addInHistory) {
		((DataFileView)getView()).changeBinaryName(textfield1, binary, addInHistory);
	}
}