/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingFormulaException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.ArithmeticExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.DefaultControlItemView;

import java.io.IOException;
import java.util.List;

/**
 * A <code>EvalItem</code> implements a calculator item.
 */
public final class EvalItem extends ControlItem {
	public static final String TAGNAME = "Eval";

	public EvalItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new DefaultControlItemView(), treeItem, graph, treeItem.getName());
		askProperties(null, 0, 2, 0);
	}

	public EvalItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(new DefaultControlItemView(), graph);
		createFromXML(elements);
		String expression;
		if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
			expression = getBinary();
			setBinary(null);
		} else {
			expression = getFormula();
		}
		try {
			checkExpression(expression);
			setFormula(expression);
			setComment(expression);
		} catch (ArianeArithmeticException e) {
			Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
			setFormula("");
		}
	}

	protected EvalItem( TreeItem treeItem, Graph graph, String name ) { // For test
		super(new DefaultControlItemView(), treeItem, graph, name);
		createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 0, treeItem.getNumberOfOutputDataSlots(), 1);
	}

	@Override
	public void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		createSlots(0, iP, 0, 0, 0);
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			if (getFormula().isEmpty()) {
				throw new ArianeMissingFormulaException(Ariane.getString("arithmetic.error0"));
			}
			int numberOfParameters = getInputParameterSlots().size();
			double[] val = new double[numberOfParameters];
			for (int i = 0; i < numberOfParameters; i++) {
				InputParameterSlot parameter = getInputParameterSlots().get(i);
				if (parameter.hasNumericalValue(true)) {
					val[i] = parameter.getNumericalValue();
				} else {
					return false;
				}
			}
			ArithmeticExpression expression = getCompiledExpression();
			final String result = Double.toString(expression.evaluate(val));
			getOutputResultSlot().setValue(result);
			Ariane.getPresenter().printSystemMessage("eval(" + getFormula() + ")=" + result); // TODO afficher les valeurs des parameter dans la ligne de commande
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
	}

	@Override
	public boolean changeFormula( String arithmeticExpression ) {
		if (getFormula() == null || !getFormula().equals(arithmeticExpression)) {
			try {
				checkExpression(arithmeticExpression);
				setFormula(arithmeticExpression);
				setComment(arithmeticExpression);
			} catch (ArianeArithmeticException e) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				return false;
			}
		}
		return true;
	}

	private ArithmeticExpression getCompiledExpression() {
		try {
			return new ArithmeticExpression(getFormula(), getInputParameterSlots().size());
		} catch (ArianeArithmeticException e) {
			return null;
		}
	}

	private void checkExpression( String textFormula ) throws ArianeArithmeticException {
		new ArithmeticExpression(textFormula, getInputParameterSlots().size());
	}
}