/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.FileNameItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A <code>FileNameItem</code> implements a filename item
 * other than an image file. Such file is used as a parameter
 * of an other item.
 */
public final class FileNameItem extends ControlItem implements IScriptArgument {
	public static final String TAGNAME = "Filename";

	private boolean _isArgument;
	private long _lastModified;

	public FileNameItem( TreeItem treeItem, Graph graph ) {
		super(new FileNameItemView(), treeItem, graph, treeItem.getName());
		super.setFormula(""); // IMPERATIVE to signify that it uses formula.
		setComment(getAbstractFileName());
	}

	public FileNameItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(new FileNameItemView(), graph);
		super.setFormula(""); // IMPERATIVE to signify that it uses formula.
		createFromXML(elements);
		setComment(getAbstractFileName());
	}

	@Override
	public boolean isArgumentOfRelatedScript() {
		return _isArgument;
	}

	@Override
	public void setArgumentOfRelatedScript( boolean value ) {
		_isArgument = value;
	}

	@Override
	public void setPrototype( String name, int iIm, int iP, int oIm ) {
		// Empty
	}

	@Override
	public void setFormula( String formula ) {
		if (!formula.trim().isEmpty()) {
			String currentDirectory = getGraph().getDirectoryName();
			super.setFormula(DataFile.setPathToFilename(formula, currentDirectory));
		}
	}

	@Override
	public boolean changeFormula( String textFormula ) {
		if (getFormula().equals(textFormula)) {
			return true;
		}
		setFormula(textFormula);
		setComment(getAbstractFileName());

		return true;
	}

	@Override
	public String getFormula( String directory ) {
		return DataFile.getRelativePath(super.getFormula(), directory, UserPreferences.getPreferredFileDirectory());
	}

	@Override
	public String getFormulaLabel( ) {
		return Ariane.getString("PropertyViewer.filename.formula");
	}

	@Override
	public String getFormulaText() {
		if (!getFormula().isEmpty()) {
			File file = new File(getFormula());
			return file.getName();
		} else {
			return Ariane.getString("PropertyViewer.browser");
		}
	}

	public void synchronizeWithFileSystem() {
		if (!getFormula().isEmpty()) {
			File file = new File(getFormula());
			if (_lastModified < file.lastModified()) {
				resetAll();
				_lastModified = file.lastModified();
			}
		}
	}

	///// 
	///// A C C E S S O R S
	/////

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	private String getAbstractFileName( ) {
		if (getFormula().trim().length() == 0) {
			return "";
		} else {
			File filename = new File(getFormula());
			return filename.getName();
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			final String filename = getFormula();
			if (filename.isEmpty()) {
				throw new ArithmeticException(Ariane.getString("arithmetic.error18"));
			}
			File file = new File(filename);
			_lastModified = file.lastModified();
			getOutputResultSlot().setValue(filename);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArithmeticException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(Ariane.ErrorCode.ARITHMETICS, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
	}
}