/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.view.graph.item.CompositeItemView;

/**
 * A <code>ControlStructureForItem</code> implements
 * the iterative control structure by a graph item.
 * The sequence of Items to has be iterated is inclosed
 * inside a subgraph.
 * The loop is made from the value of the parameter $0 until
 * the value of the parameter $1 by step of the value of the parameter $2.
 */
public final class ForControlStructureItem extends CompositeItem {
	public static final String TAGNAME = "For";
	private static final String ABSTRACT_NAME = "for";

	public ForControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
		setSignature(getRelatedTreeItem().getSignature());
		setComment(UserPreferences.EMPTY_STRING);
	}

	public ForControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(new CompositeItemView(), graph);
		createFromXML(xmlTree);
		setComment(UserPreferences.EMPTY_STRING);
		// Keep the following lines for the sake of compatibility
		if (getRelatedTreeItem() == null) {
			setRelatedTreeItem(restoreDefaultTreeItem());
		}
		setSignature(getRelatedTreeItem().getSignature());
	}

	public ForControlStructureItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(new CompositeItemView(), graph, treeItem);
		createFromXML(xmlTree);
		setComment(UserPreferences.EMPTY_STRING);
		setSignature(getRelatedTreeItem().getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
	}

	public ForControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		createSlots(inputDataSlotNumber, inputParameterSlotNumber + 3, 0, outputDataSlotNumber, 1);
		setSignature(getRelatedTreeItem().getSignature());
		setComment(UserPreferences.EMPTY_STRING);
	}

	@Override
	public TreeItem restoreDefaultTreeItem() {
		setAbstractName(ABSTRACT_NAME);
		return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
	}

	@Override
	public void setPrototype( String name, int iIm, int iP, int oIm ) {
		setPrototype(name, iIm, 3 + iP, oIm, iIm, 1);
	}

	@Override
	protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP - 2, oIm, iIm, 0);
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public void resetHookHook() {
		setComment(UserPreferences.EMPTY_STRING);
	}

	@Override
	protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
		// 1. Get and check the 3 parameter values used in the conditional expression
		int[] loop = new int[3];
		if (!getLoopParameters(loop)) {
			return false;
		}

		// 2. Set the first loop value to the virtual operator
		OutputResultSlot result1 = virtualItem.getVirtualInputParameterSlots().get(0);
		result1.setValue(Integer.toString(loop[0]));

		// 3. Get and check the rest of the parameter values
		for (int i = loop.length; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			if (parameter.hasValue(true)) {
				OutputResultSlot result = virtualItem.getVirtualInputParameterSlots().get(1 + i - loop.length);
				result.setValue(parameter.getValue());
			} else {
				return false;
			}
		}

		getMask(virtualItem);

		return getInputDataFiles(virtualItem);
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			VirtualItem virtualItem = getVirtualItem();
			if (!linkItemWithVirtualItem(virtualItem)) {
				return false;
			}

			int[] loop = new int[3];
			if (!getLoopParameters(loop)) {
				return false;
			}

			Ariane.getPresenter().printSystemMessage("> For loop from " + loop[0] + " to " + loop[1] + " step " + loop[2] + " do");

			if (loop[0] > loop[1] && loop[2] > 0 || loop[0] < loop[1] && loop[2] < 0) {
				Ariane.getPresenter().printSystemMessage("> end loop for");
				return false;
			}

			changeStatus(Status.RUNNING);
			// Iteration
			for (int loopv = loop[0]; loopv <= loop[1] && Graph.isExecutionRunning(); loopv += loop[2]) {
				reset();
				virtualItem.changeStatus(Status.EXECUTED);

				Ariane.getPresenter().printSystemMessage("> loop indice " + loopv);
				setComment(Integer.toString(loopv));
				changeStatus(Status.RUNNING);

				// Set the value of the inner parameter with the current loop value.
				virtualItem.getVirtualInputParameterSlots().get(0).setValue(Integer.toString(loopv));

				// The body execution is split in 2 steps
				// to avoid problems when two operators use the same entry.
				executeBody(virtualItem);
				setIntermediateInputs(virtualItem);
			}

			Ariane.getPresenter().printSystemMessage("> end loop for");

			setOutputData(virtualItem);

			if (!setOutputResult(virtualItem)) {
				return false;
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			throw displayError(e, e.getErrorCode());
		} catch (IOException e) {
			throw displayError(e, Ariane.ErrorCode.INPUTOUTPUT);
		}
	}

	private boolean getLoopParameters( int[] loop ) throws ArianeMissingParameterException {
		for (int i = 0; i < 3; i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			if (parameter.hasNumericalValue(true)) {
				loop[i] = (int)parameter.getNumericalValue();
			} else {
				return false;
			}
		}
		return true;
	}

	///// 
	///// P R O P E R T I E S
	/////

	@Override
	protected void askProperties( ) throws InterruptedException {
		askProperties(null, getInputDataSlots().size(), 0, getOutputDataSlots().size());
	}
}
