/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeDataFileIOException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.LoaderItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.JTextField;

/**
 * A <code>LoaderItem</code> manages data files loading
 * within a graph.
 */
public final class LoaderItem extends DataFile {
	public static final String TAGNAME = "Loader";

	private long _dateOfLastModification;

	public LoaderItem( TreeItem relatedTreeItem, Graph graph ) {
		super(new LoaderItemView(), relatedTreeItem, graph, relatedTreeItem.getName(), null);
		createSlots(0, 1, 0, 1, 1);
	}

	public LoaderItem( Graph graph, List<XMLElement> data ) throws IOException {
		super(new LoaderItemView(), graph);
		createSlots(0, 0, 0, 0, 1);
		createFromXML(data);
		if (getInputParameterSlots().isEmpty()) { // For the sake of compatibility
			createSlots(0, 1, 0, 0, 0);
		}
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	/*
	 * Returns the name of the file with either
	 * - relative path if the file is in the specified directory
	 * - or with an absolute path.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory, UserPreferences.getPreferredFileDirectory());
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory name
	 * of the project.
	 * 
	 * @param filename a file name (absolute or relative)
	 */
	@Override
	public void setBinary( String filename ) {
		if (filename != null && !filename.trim().isEmpty()) {
			String currentDirectory = getGraph().getDirectoryName();
			super.setBinary(DataFile.setPathToFilename(filename, currentDirectory));
		}
	}

	/**
	 * Resynchronizes the item with the related file.
	 * if the related file has been changed, resets the item.
	 */
	public void synchronizeWithFileSystem() {
		if (getBinary() != null) {
			File file = new File(getBinary());
			if (_dateOfLastModification < file.lastModified()) {
				resetAll();
				_dateOfLastModification = file.lastModified();
			}
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public String getViewerCommand() {
		return GraphItem.getAbsoluteBinary(UserPreferences.getCommandVisu());
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			setCurrentFilename(null);
			final InputParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
			final File file;
			if (parameter != null && parameter.hasValue(false)) {
				changeBinaryName((JTextField)getView().getNameEditor(), null, false);
				file = new File(parameter.getValue());
			} else if (getBinary() == null || getBinary().isEmpty()) {
				throw new ArianeDataFileIOException(Ariane.ErrorCode.MISSINGFILE, Ariane.getString("ImageFile.error3"));
			} else {
				file = new File(getBinary());
				_dateOfLastModification = file.lastModified();
			}
			if (!file.exists()) {
				throw new ArianeDataFileIOException(Ariane.ErrorCode.UNKNOWNFILE, Ariane.getString("ImageFile.error4")
						+ " \""
						+ file.getAbsolutePath()
						+ "\"");
			}
			final String filename = file.getAbsolutePath();
			getOutputDataSlots().get(0).setValue(filename);
			setCurrentFilename(filename);
			if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
				getOutputResultSlot().setValue(filename);
			}
		} catch (ArianeDataFileIOException | ArianeMissingParameterException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}

		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	@Override
	public void fireConnectionHook( InputSlot slot, boolean connection ) {
		if (slot == getInputParameterSlots().get(0)) {
			((LoaderItemView)getView()).setButtonEnabled(!connection);
		}
	}
}