/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.view.graph.item.CompositeItemView;

/**
 * A <code>MacroOperatorItem</code> is a graph item that is composed by
 * a subgraph of items.
 */
public final class MacroOperatorItem extends CompositeItem {
	public static final String TAGNAME = "MacroOperator";
	private static final String ABSTRACT_NAME = "macro-operator";

	public MacroOperatorItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		askProperties(ABSTRACT_NAME + getId(), 1, 1, 1);
	}

	public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(new CompositeItemView(), graph);
		createFromXML(xmlTree);
		// Keep the following lines for the sake of compatibility
		if (getRelatedTreeItem() == null) {
			setRelatedTreeItem(restoreDefaultTreeItem());
		}
		if (getInputMaskSlot() != null) {
			createSlots(0, 0, 1, 0, 0);
		}
		if (getRelatedTreeItem() != null) {
			setSignature(getRelatedTreeItem().getSignature());
		}
	}

	public MacroOperatorItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(new CompositeItemView(), graph, treeItem);
		createFromXML(xmlTree);
		setSignature(getRelatedTreeItem().getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		// Keep the following lines for the sake of compatibility
		if (getInputMaskSlot() != null) {
			createSlots(0, 0, 1, 0, 0);
		}
	}

	public MacroOperatorItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		createSlots(inputDataSlotNumber, inputParameterSlotNumber, 0, outputDataSlotNumber, 1);
	}

	@Override
	public TreeItem restoreDefaultTreeItem() {
		setAbstractName(ABSTRACT_NAME);
		return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
	}

	@Override
	public void setPrototype( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
		setPrototype(name, inputDataCount, parameterCount, outputDataCount, 0, 0);
	}

	@Override
	protected SubGraph createSubGraph( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
		return createSubGraph(getNickName(), inputDataCount, parameterCount, outputDataCount, 0, 0);
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
		Iterator<InputParameterSlot> its1 = getInputParameterSlots().iterator();
		Iterator<OutputResultSlot> itd1 = virtualItem.getVirtualInputParameterSlots().iterator();
		while (its1.hasNext()) {
			InputParameterSlot parameter = its1.next();
			if (parameter.hasValue(true)) {
				OutputResultSlot result = itd1.next();
				result.setValue(parameter.getValue());
			} else {
				return false;
			}
		}

		getMask(virtualItem);

		return getInputDataFiles(virtualItem);
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			VirtualItem virtualItem = getVirtualItem();
			if (!linkItemWithVirtualItem(virtualItem)) {
				return false;
			}

			changeStatus(Status.RUNNING);

			setOutputData(virtualItem);

			if (!setOutputResult(virtualItem)) {
				return false;
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			throw displayError(e, e.getErrorCode());
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	@Override
	protected void askProperties() throws InterruptedException {
		askProperties(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
	}
}