/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.Ariane.ErrorCode;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.model.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.OperatorItemView;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

/**
 * An <code>OperatorItem</code> is a graph item that encapsulates
 * a Shell command. It uses inputs to produce outputs with
 * specific parameter values.
 */
public final class OperatorItem extends GraphItem {
	public static final String TAGNAME = "Operator";
	public static final String SEPARATOR = "&";

	public OperatorItem( TreeItem treeItem, Graph graph ) {
		super(new OperatorItemView(), treeItem, graph, treeItem.getName(), treeItem.getCategory());
		createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 1, treeItem.getNumberOfOutputDataSlots(), 1);
		setSignature(treeItem.getSignature());
	}

	public OperatorItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(new OperatorItemView(), graph);
		// Creates mask and result because they are not saved in the xmlTree.
		createSlots(0, 0, 1, 0, 1);
		createFromXML(xmlTree);
		if (getRelatedTreeItem() != null) {
			setSignature(getRelatedTreeItem().getSignature());
		} else {
			Ariane.getPresenter().printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName(), getId(), Ariane.getString("Operator.error7"));
		}
	}

	@Override
	public String toString( ) {
		String commandLine = "(" + getId() + "): " + getName();

		for (InputParameterSlot isp : getInputParameterSlots()) {
			if (isp.getValue() != null) {
				commandLine += " " + isp.getValue();
			}
		}
		return commandLine;
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	///// 

	private String getRelativeFilePathName( String filename ) {
		String directory = UserPreferences.getTemporaryDirectory();
		if (filename.startsWith(directory)) {
			return filename.substring(directory.length());
		}
		return filename;
	}

	public static String getCommandPart( String binary ) {
		if (binary == null ) {
			return null;
		}
		String[] format =  binary.split(OperatorItem.SEPARATOR);
		if (format.length > 0) {
			return format[0];
		}
		return null;
	}

	public static String getFormatPart( String binary ) {
		if (binary != null) {
			int index = binary.indexOf(OperatorItem.SEPARATOR);
			if (index > 0 && index + 1 < binary.length()) {
				return binary.substring(index + 1);
			} else {
				return "[$p] [-m $m] [$i] [$o]";
			}
		}
		return null;
	}

	@Override
	public String getBinary( ) {
		return (getRelatedTreeItem() == null) ? super.getBinary() : getRelatedTreeItem().getBinaryPath();
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  $p for parameter; $p0 for the first parameter
	 *  $m for the mask
	 *  $i for input; $i0 for the first input data
	 *  $o for output; $o0 for the first output data.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse(getCommandPart(getBinary()), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 *  %p for parameter; %p0 for the first parameter
	 *  %m for the mask
	 *  %i for input; %i0 for the first input data file.
	 *  %o for output; %o0 for the first output data file.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public List<String> buildCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse(GraphItem.getAbsoluteBinary(getCommandPart(getBinary())), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			if (getBinary() == null) {
				throw new ArianeRuntimeException(Ariane.getString("Operator.error7"));
			}

			List<String> parameterLine = new LinkedList<String>();
			if (!getParameters(parameterLine)) {
				return false;
			}

			List<String> maskLine = new LinkedList<String>();
			getOptionalMask(maskLine);

			List<String> inputLine = new LinkedList<String>();
			if (!getInputDataFile(inputLine)) {
				return false;
			}
			
			List<String> outputLine = new LinkedList<String>();
			createOutputDatFiles(outputLine);

			List<String> commandLine = buildCommandLine(maskLine, parameterLine, inputLine, outputLine);
			String consoleMessage = getName();
			for (int i = 1; i < commandLine.size(); i++ ) {
				consoleMessage += " " + getRelativeFilePathName(commandLine.get(i));
			}
			Ariane.getPresenter().printSystemMessage(consoleMessage);

			changeStatus(Status.RUNNING);
			executeBinary(commandLine);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			throw displayError(e, e.getErrorCode());
		} catch (IOException e) {
			throw displayError(e, Ariane.ErrorCode.INPUTOUTPUT);
		} catch (InterruptedException e) {
			throw displayError(e, Ariane.ErrorCode.INTERRUPTEDEXECUTION);
		}
	}

	private void createOutputDatFiles( List<String> outputLine ) {
		for (OutputDataSlot slot : getOutputDataSlots()) {
			outputLine.add(slot.getTempName());
		}
	}

	private void executeBinary( List<String> commandLine ) throws IOException,
			InterruptedException, ArianeRuntimeException {
		OperatorProcessBuilder processBuilder = new OperatorProcessBuilder(commandLine);
		long start = System.currentTimeMillis(); // NOPMD by regis on 5/6/15 5:51 PM
		Process proc1 = processBuilder.start();
		getGraph().setCurrentProcess(proc1);

		BufferedReader reader = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
		String line;
		while ((line = reader.readLine()) != null) {
			Ariane.getPresenter().printOutputMessage(line);
		}
		int exitValue = proc1.waitFor();
		getGraph().setCurrentProcess(null);

		if (exitValue == 0) {
			// Get the status value
			if (!UserPreferences.getCommandStatus().isEmpty()) {
				processBuilder = new OperatorProcessBuilder(GraphItem.getAbsoluteBinary(UserPreferences.getCommandStatus()));
				Process proc2 = processBuilder.start();
				InputStream resultStream = proc2.getInputStream();
				proc2.waitFor();

				// Exploit the results
				byte[] resultByte = new byte[resultStream.available() + 1];
				resultStream.read(resultByte);
				String resultString = new String(resultByte);
				getOutputResultSlot().setValue(resultString.trim());
			} else {
				getOutputResultSlot().setValue("SUCCESS");
			}
		}

		// I mean if an execution fails than the error code !=0.
		if (exitValue != 0) {
			throw new ArianeRuntimeException(Ariane.getString("Operator.error2"));
		}
		// Everything is OK -> validate the result of all outputs.
		for (OutputDataSlot slot : getOutputDataSlots()) {
			slot.setValue();
		}
		long end = System.currentTimeMillis();
		Ariane.getPresenter().printSystemMessage("(" + (end - start) / 1000.0f + "s)");
	}

	private boolean getInputDataFile( List<String> inputLine )
			throws ArianeMissingInputException {
		for (InputDataSlot slot: getInputDataSlots()) {
			if (slot.hasValue(true)) {
				inputLine.add(slot.getValue());
			} else {
				return false;
			}
		}
		return true;
	}

	private void getOptionalMask( List<String> maskLine ) throws ArianeRuntimeException {
		if (getInputMaskSlot().isConnected() && getInputMaskSlot().hasValue(false)) {
			String temp = getInputMaskSlot().getValue();
			maskLine.add(temp);
		}
	}

	private boolean getParameters( List<String> parameterLine )
			throws ArianeMissingParameterException {
		// Get the parameter values
		for (InputParameterSlot slot : getInputParameterSlots()) {
			if (slot.hasValue(true)) {
				parameterLine.add(slot.getValue());
			} else {
				return false;
			}
		}
		return true;
	}

	private ArianeRuntimeException displayError( Exception exception, ErrorCode inputoutput ) {
		if (Graph.isExecutionRunning()) {
			Ariane.getPresenter().printErrorMessage(inputoutput, getAbsoluteName(), getId(), exception.getLocalizedMessage());
			getGraph().stopExecution();
			changeStatus(Status.ERROR);
		} else {
			changeStatus(Status.UNEXECUTED);
		}
		return new ArianeRuntimeException(exception);
	}
}