/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.SaverItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.JTextField;

/**
 * A <code>saverItem</code> manages saving
 * of a data file in a graph.
 */
public final class SaverItem extends DataFile {
	public static final String TAGNAME = "Saver";

	public SaverItem( TreeItem relatedTreeItem, Graph graph ) {
		super(new SaverItemView(), relatedTreeItem, graph, relatedTreeItem.getName(), null);
		createSlots(1, 1, 0, 1, 1);
	}

	public SaverItem( Graph graph, List<XMLElement> data ) throws IOException {
		super(new SaverItemView(), graph);
		createSlots(0, 0, 0, 0, 1);
		createFromXML(data);
		if (getOutputDataSlots().isEmpty()) { // For the sake of compatibility
			createSlots(0, 0, 0, 1, 0);
		}
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	/*
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 */
	@Override
	public void setBinary( String binary ) {
		if (binary != null && !binary.trim().isEmpty()) {
			String currentDirectory = getGraph().getDirectoryName();
			super.setBinary(DataFile.setPathToFilename(binary, currentDirectory));
		}
	}

	/*
	 * Returns the name of the file with either
	 * - relative path if the file is in the specified directory
	 * - or with an absolute path.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory, UserPreferences.getPreferredFileDirectory());
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public String getViewerCommand() {
		return GraphItem.getAbsoluteBinary(UserPreferences.getCommandVisu());
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			setCurrentFilename(null);
			final InputDataSlot isi = getInputDataSlots().get(0);
			if (isi.hasValue(true)) {
				File file = getFilename();
				copyFile(isi.getValue(), file.getAbsolutePath());
				String exec = Ariane.getString("ImageFile.save") + ": " + file.getAbsolutePath();
				Ariane.getPresenter().printSystemMessage(exec);
				final String filename = file.getAbsolutePath();
				setCurrentFilename(filename);
				getOutputDataSlots().get(0).setValue(filename);
				if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
					getOutputResultSlot().setValue(filename);
				}
			} else {
				return false;
			}
		} catch (ArianeException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		} catch (IOException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName(), getId(), Ariane.getString("ImageFile.error6") + e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	private File getFilename() throws ArianeMissingParameterException, ArianeMissingInputException {
		File file;
		InputParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
		if (parameter != null && parameter.hasValue(false)) {
			changeBinaryName((JTextField)getView().getNameEditor(), null, false);
			file = new File(parameter.getValue());
		} else if (getBinary() == null || getBinary().isEmpty()) {
			throw new ArianeMissingInputException(Ariane.getString("ImageFile.error7"));
		} else {
			file = new File(getBinary());
		}
		return file;
	}

	@Override
	public void fireConnectionHook( InputSlot slot, boolean connection ) {
		if (!getInputParameterSlots().isEmpty() && slot == getInputParameterSlots().get(0)) {
			((SaverItemView)getView()).setButtonEnabled(!connection);
		}
	}
}