/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.ControlStructureItemView;

/**
 * A <code>ControlStructureSelectItem</code> implements the a
 * selector of multiple inputs. The selected input
 * is given by the parameter value.
 */
public final class SelectControlStructureItem extends ControlItem {
	public static final String TAGNAME = "Select";

	public SelectControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new ControlStructureItemView(), treeItem, graph, treeItem.getName());
		askProperties(null, 1, 0, 0);
		setSignature(treeItem.getSignature());
	}

	public SelectControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(new ControlStructureItemView(), graph);
		createFromXML(xmlTree);
	}

	public SelectControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { // For tests
		super(new ControlStructureItemView(), treeItem, graph, treeItem.getName());
		createSlots(inputDataSlotNumber, inputParameterSlotNumber, 0, outputDataSlotNumber, 1);
		setSignature(treeItem.getSignature());
	}

	@Override
	public void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		createSlots(iIm, 1, 0, 1, 0);
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	/////  E X E C U T I O N 
	/////

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			// Gets and checks the parameter (-> executes forward operators if needed).
			InputParameterSlot parameter = getInputParameterSlots().get(0);
			int inputRank;
			if (parameter.hasNumericalValue(true)) {
				inputRank = (int)parameter.getNumericalValue();
			} else {
				return false;
			}

			// Checks consistency of the parameter.
			if (inputRank < 0 || inputRank >= getInputDataSlots().size()) {
				throw new ArithmeticException(Ariane.getString("controlstructure.error1")); 
			}

			// Gets and checks the input image (-> executes forward operators if needed).
			InputDataSlot isi = getInputDataSlots().get(inputRank);
			if (isi != null && isi.hasValue(true)) {
				// Validate the output.
				(getOutputDataSlots().get(0)).setValue(isi.getValue());
				getOutputResultSlot().setValue(Float.toString(inputRank));
			} else {
				return false;
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
	}

	///// 
	/////  P R O P E R T I E S
	/////

	protected void askProperties() throws InterruptedException {
		askProperties(null, 1, 0, 0);
	}
}