/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.Ariane.ErrorCode;
import fr.greyc.pantheon.ariane.model.PlatformSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.model.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.SystemItemView;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

/**
 * A <code>SystemItem</code> is a graph item that
 * executes an operating system command. It stores the command line
 * and gets its parameters, inputs and outputs from slots.
 */
public final class SystemItem extends ControlItem {
	public static final String TAGNAME = "System";

	public SystemItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new SystemItemView(), treeItem, graph, treeItem.getName());
		askProperties(null, 1, 1, 1);
		setBinary("");
	}

	public SystemItem( Graph graph, List<XMLElement> elements ) throws IOException {
		super(new SystemItemView(), graph);
		createFromXML(elements);
	}

	/*
	 * For tests
	 */
	public SystemItem( TreeItem treeItem, Graph graph, int inputs, int parameters, int outputs ) {
		super(new SystemItemView(), treeItem, graph, treeItem.getName());
		setPrototype(treeItem.getName(), inputs, parameters, outputs);
		setBinary("");
	}

	@Override
	public void setPrototype( String name, int inputImageNumber, int inputParameterNumber, int outputImageNumber ) {
		setAbstractName(name);
		createSlots(inputImageNumber, inputParameterNumber, 0, outputImageNumber, 1);
	}

	///// 
	///// A C C E S S O R S
	/////

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/*
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  $p for parameter; $p0 for the first parameter
	 *  $m for the mask
	 *  $i for input; $i0 for the first input image
	 *  $o for output; $o0 for the first output image.
	 */
	public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
		return ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine);
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			List<String> parameterLine = new LinkedList<String>();
			if (!getParameterValues(parameterLine) ) {
				return false;
			}

			List<String> inputLine = new LinkedList<String>();
			if (!getInputData(inputLine)) {
				return false;
			}

			List<String> outputLine = new LinkedList<String>();
			createOutputDataAsTemporaryFiles(outputLine);

			List<String> maskLine = new LinkedList<String>();
			List<String> commandLine = ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine );

			String consoleMessage = "";
			for (int i = 0; i < commandLine.size(); i++ ) {
				consoleMessage += " " + commandLine.get(i);
			}
			Ariane.getPresenter().printSystemMessage(consoleMessage);
			
			commandLine.clear();
			if (PlatformSettings.isWindows()) {
				commandLine.add("cmd");
				commandLine.add("/C");
				commandLine.add(consoleMessage);
			} else {
				String shell = System.getenv("SHELL");
				if (shell.length() == 0) {
					shell = "/bin/sh";
				}
				commandLine.add(shell);
				commandLine.add("-c");
				commandLine.add(consoleMessage);
			}

			// Execution of the operator binary
			changeStatus(Status.RUNNING);
			OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
			Process proc1 = pb.start();
			getGraph().setCurrentProcess(proc1);

			BufferedReader br = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
			String line;
			while ((line = br.readLine()) != null) {
				Ariane.getPresenter().printOutputMessage(line);
			}
			int exitValue = proc1.waitFor();
			getGraph().setCurrentProcess(null);

			if (exitValue == 0) {
				getOutputResultSlot().setValue("SUCCESS");
			} else {
				throw new ArianeRuntimeException(Ariane.getString("Operator.error2"));
			}
			// Everything is OK -> validate the result of all outputs.
			for (OutputDataSlot slot : getOutputDataSlots()) {
				slot.setValue();
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			errorMessage(e, e.getErrorCode());
			throw new ArianeRuntimeException(e);
		} catch (IOException e) {
			errorMessage(e, Ariane.ErrorCode.INPUTOUTPUT);
			throw new ArianeRuntimeException(e);
		} catch (InterruptedException e) {
			errorMessage(e, Ariane.ErrorCode.INTERRUPTEDEXECUTION);
			throw new ArianeRuntimeException(e);
		}
	}

	private void errorMessage( Exception error, ErrorCode errorCode  ) {
		if (Graph.isExecutionRunning()) {
			Ariane.getPresenter().printErrorMessage(errorCode, getAbsoluteName(), getId(), error.getLocalizedMessage());
			getGraph().stopExecution();
			changeStatus(Status.ERROR);
		} else {
			changeStatus(Status.UNEXECUTED);
		}
	}

	private void createOutputDataAsTemporaryFiles( List<String> outputLine ) {
		for (OutputDataSlot slot : getOutputDataSlots()) {
			outputLine.add(slot.getTempName());
		}
	}

	private boolean getInputData( List<String> inputLine ) throws ArianeMissingInputException {
		for (InputDataSlot slot: getInputDataSlots()) {
			if (slot.hasValue(true)) {
				inputLine.add(slot.getValue());
			} else {
				return false;
			}
		}
		return true;
	}

	private boolean getParameterValues( List<String> parameterLine ) throws ArianeMissingParameterException {
		// Get the parameter values
		for (InputParameterSlot slot : getInputParameterSlots()) {
			if (slot.hasValue(true)) {
				parameterLine.add(slot.getValue());
			} else {
				return false;
			}
		}
		return true;
	}
}