/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.view.graph.item.ViewerItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A <code>ViewerItem</code> manages displaying
 * of a data file in a graph.
 */
public final class ViewerItem extends DataFile {
	public static final String TAGNAME = "Viewer";

	public ViewerItem( TreeItem treeItem, Graph graph ) {
		super(new ViewerItemView(), treeItem, graph, treeItem.getName(), null);
		createSlots(1, 0, 0, 1, 0);
	}

	public ViewerItem( Graph graph, List<XMLElement> data ) throws IOException {
		super(new ViewerItemView(), graph);
		createFromXML(data);
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public String getViewerCommand() {
		if (getBinary() == null) {
			return GraphItem.getAbsoluteBinary(UserPreferences.getCommandVisu());
		} else {
			return getBinary();
		}
	}

	public String getViewerCommandName() {
		File file = new File(getViewerCommand());
		return file.getName();
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			setCurrentFilename(null);
			InputDataSlot isi = getInputDataSlots().get(0);
			if (isi.hasValue(true)) {
				executeCoreAction(isi);
			} else {
				return false; 
			}
			((ViewerItemView)getView()).displayImage();
			// Validate the result.
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeMissingInputException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		} catch (ArianeRuntimeException | InterruptedException | IOException e) {
			if (Graph.isExecutionRunning()) {
				Ariane.getPresenter().printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName(), getId(), e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			throw new ArianeRuntimeException(e);
		}
	}

	protected void executeCoreAction( InputDataSlot isi ) {
		final String filename = isi.getValue();
		setCurrentFilename(filename);
		getOutputDataSlots().get(0).setValue(filename);
	}

	public void changeCommandName( String command, boolean addInHistory ) {
		((ViewerItemView)getView()).changeCommandName(command, addInHistory);
	}
}