/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.SubGraph;
import fr.greyc.pantheon.ariane.view.graph.item.CompositeItemView;

import java.io.IOException;
import java.util.List;

/**
 * A <code>ControlStructureWhileItem</code> implements the
 * repetitive structure control by a graph item.
 * The sequence of Items to has be repeated is inclosed
 * inside a subgraph.
 */
public final class WhileControlStructureItem extends CompositeItem {
	public static final String TAGNAME = "While";
	private static final String ABSTRACT_NAME = "while";

	public WhileControlStructureItem( TreeItem treeItem, Graph graph ) throws InterruptedException {
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		askProperties(null, 1, 1, 1);
		setSignature(getRelatedTreeItem().getSignature());
		setComment(UserPreferences.EMPTY_STRING); 
	}

	public WhileControlStructureItem( Graph graph, List<XMLElement> xmlTree ) throws IOException {
		super(new CompositeItemView(), graph);
		createFromXML(xmlTree);
		String expression;
		if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
			expression = getBinary();
			setBinary(null);
		} else {
			expression = getFormula();
		}
		try {
			checkExpression(expression);
			setFormula(expression);
			setComment(expression);
		} catch (ArianeArithmeticException e) {
			Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
			setFormula("");
		}

		setComment(UserPreferences.EMPTY_STRING);
		if (getRelatedTreeItem() == null) {
			setRelatedTreeItem(restoreDefaultTreeItem());
		}
		setSignature(getRelatedTreeItem().getSignature());
	}

	public WhileControlStructureItem( Graph graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(new CompositeItemView(), graph, treeItem);
		createFromXML(xmlTree);
		String expression;
		if (getFormula() == null && getBinary() != null) { // Keep this for sake of compatibility
			expression = getBinary();
			setBinary(null);
		} else {
			expression = getFormula();
		}
		try {
			checkExpression(expression);
			setFormula(expression);
			setComment(expression);
		} catch (ArianeArithmeticException e) {
			Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
			setFormula("");
		}
		// used to display the current loop value
		setComment(UserPreferences.EMPTY_STRING);
		setSignature(getRelatedTreeItem().getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
	}

	public WhileControlStructureItem( TreeItem treeItem, Graph graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { //For test
		super(new CompositeItemView(), treeItem, graph, ABSTRACT_NAME);
		createSlots(inputDataSlotNumber, inputParameterSlotNumber + 3, 0, outputDataSlotNumber, 1);
		setSignature(getRelatedTreeItem().getSignature());
		setComment(UserPreferences.EMPTY_STRING);
	}

	@Override
	public TreeItem restoreDefaultTreeItem() {
		setAbstractName(ABSTRACT_NAME);
		return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
	}

	@Override
	public void setPrototype( String name, int iIm, int iP, int oIm ) {
		super.setPrototype(name, iIm, 2 + iP, oIm, iIm, 2);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
	}

	@Override
	protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP, oIm, iIm, 2);
	}

	@Override
	public String getTagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	@Override
	public void resetHookHook() {
		setComment(UserPreferences.EMPTY_STRING);
	}

	@Override
	protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
		// 1. Get the parameter values
		for (int i = 0; i < getInputParameterSlots().size(); i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			if (parameter.hasValue(true)) {
				OutputResultSlot result = virtualItem.getVirtualInputParameterSlots().get(i);
				result.setValue(parameter.getValue());
			} else {
				return false;
			}
		}

		getMask(virtualItem);

		return getInputDataFiles(virtualItem);
	}

	@Override
	protected boolean primitiveExecute( ) throws ArianeRuntimeException {
		try {
			if (getFormula().isEmpty()) {
				throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0"));
			}

			VirtualItem virtualItem = getVirtualItem();
			if (!linkItemWithVirtualItem(virtualItem)) {
				return false;
			}

			double[] whilevariables = new double[2];
			if (!getLoopParameters(whilevariables)) {
				return false;
			}

			changeStatus(Status.RUNNING);

			// Repetition
			int iteration = 0;
			LogicalExpression condition = getCompiledExpression();

			Ariane.getPresenter().printSystemMessage("> While loop (" + condition.toHtml() + ") do");
			while (condition.evaluate(whilevariables[0], whilevariables[1]) && Graph.isExecutionRunning()) {
				iteration++;

				reset();
				virtualItem.changeStatus(Status.EXECUTED); // Prevent the rexecution of the virtual operator.

				Ariane.getPresenter().printSystemMessage("> cycle " + iteration);
				setComment(Integer.toString(iteration));
				changeStatus(Status.RUNNING);

				// Set the value of the 2 inner parameters.
				for (int i = 0; i < 2; i++) {
					virtualItem.getVirtualInputParameterSlot(i).setValue(Float.toString((float)whilevariables[i]));
				}

				// The body execution is split in 2 steps
				// to avoid problems when two operators use the same entry.
				executeBody(virtualItem);
				setIntermediateInputs(virtualItem);

				setLoopParameterValues(virtualItem, whilevariables);
			}
			Ariane.getPresenter().printSystemMessage("end while loop");

			setOutputData(virtualItem);

			if (!setOutputResult(virtualItem)) {
				return false;
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArianeException e) {
			throw displayError(e, e.getErrorCode());
		} catch (IOException e) {
			throw displayError(e, Ariane.ErrorCode.INPUTOUTPUT);
		}
	}

	private boolean getLoopParameters( double[] whilevariables )
			throws ArianeMissingParameterException {
		for (int i = 0; i < 2; i++) {
			InputParameterSlot parameter = getInputParameterSlots().get(i);
			if (parameter.hasNumericalValue(true)) {
				whilevariables[i] = (int)parameter.getNumericalValue();
			} else {
				return false;
			}
		}
		return true;
	}

	private void setLoopParameterValues( VirtualItem virtualItem, double[] whilevariables ) throws ArianeMissingParameterException {
		for (int i = 0; i < 2; i++) {
			InputParameterSlot parameter = virtualItem.getVirtualIntermediateResultSlot(i);
			if (parameter.isConnected() && parameter.hasNumericalValue(true)) {
				whilevariables[i] = (int)parameter.getNumericalValue();
			}
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	@Override
	public boolean changeFormula( String condition ) {
		if (getFormula() != null && getFormula().equals(condition)) {
			return true;
		}
		try {
			checkExpression(condition);
			setFormula(condition);
			return true;
		} catch (ArianeArithmeticException e) {
			Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
		}
		return false;
	}

	private LogicalExpression getCompiledExpression() {
		try {
			return new LogicalExpression(getFormula());
		} catch (ArianeArithmeticException e) {
			return null;
		}
	}

	private void checkExpression( String textFormula ) throws ArianeArithmeticException {
		new LogicalExpression(textFormula);
	}

	@Override
	protected void askProperties( ) throws InterruptedException {
		askProperties(null, 1, 0, 0);
	}
}