/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;

/**
 * A <code>ArithmeticExpression</code> is a calculus formula that includes
 * arithmetic operators, constant values and parameters values ($0, $1, etc.).
 */
public final class ArithmeticExpression {
	private static final byte PLUS = -1;
	private static final byte MINUS = -2;
	private static final byte TIMES = -3;
	private static final byte DIVIDE = -4;
	private static final byte MODULO = -5;
	private static final byte POWER = -6;
	private static final byte MIN = -11;
	private static final byte MAX = -12;
	private static final byte SIN = -21;
	private static final byte COS = -22;
	private static final byte TAN = -23;
	private static final byte ASIN = -24;
	private static final byte ACOS = -25;
	private static final byte ATAN = -26;
	private static final byte EXP = -27;
	private static final byte LN = -28;
	private static final byte LOG10 = -29;
	private static final byte LOG2 = -30;
	private static final byte ABS = -31;
	private static final byte SQRT = -32;
	private static final byte UNARYMINUS = -33;
	private final byte[] _variables;
	private final int _nbParameters;

	private final String[] _functionNamesUnary = {
			"sin", "cos", "tan", "asin",
			"acos", "atan", "exp",
			"ln", "log10", "log2", "abs", "sqrt" };

	private final String[] _functionNamesBinary = { "min", "max" };

	private final String _initialExpression;
	private byte[] _codedExpression;
	private double[] _stack;
	private double[] _constantStack;

	private int _pos;
	private int _constantStackIndex;
	private int _codeSize;

	public ArithmeticExpression( String formula, int nbParameters ) throws ArianeArithmeticException {
		_nbParameters = nbParameters;
		_variables = new byte[_nbParameters];
		_initialExpression = formula;
		if (formula == null || formula.trim().isEmpty()) {
			return;
		}
		_codedExpression = new byte[formula.length()];
		_constantStack = new double[formula.length()];
		for (int i = 0; i < _nbParameters; i++) {
			_variables[i] = (byte) (-34 - i);
		}

		// Convert formula to computable code
		_pos = 0;
		_constantStackIndex = 0;
		_codeSize = 0;
		parseFormula();
		skipWhiteSpaces();
		if (nextCharacter() != 0) {
			error(Ariane.getString("arithmetic.error1"));
		}
		int stackSize = computeStackUsage();
		_stack = new double[stackSize];
		byte[] code = new byte[_codeSize];
		System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
		_codedExpression = code;
		double[] copy = new double[_constantStackIndex];
		System.arraycopy(_constantStack, 0, copy, 0, _constantStackIndex);
		_constantStack = copy;
	}

	public double evaluate( double[] values ) throws ArianeArithmeticException {
		int top = 0;
		for (int i = 0; i < _codeSize; i++) {
			int var = 0;
			boolean flag = false;
			if (_codedExpression[i] >= 0) { // constants
				_stack[top++] = _constantStack[_codedExpression[i]];
			} else if (_codedExpression[i] >= POWER) { // binary operators
				double operand2 = _stack[--top];
				double operand1 = _stack[--top];
				double ans = performBasicBinaryOperation(_codedExpression[i], operand1, operand2);
				if (Double.isNaN(ans)) {
					return ans;
				}
				_stack[top++] = ans;
			} else {
				// parameters
				while (!flag && var < _nbParameters) {
					if (_codedExpression[i] == _variables[var]) {
						flag = true;
						_stack[top++] = values[var];
					}
					var++;
				}
				// otherwise
				if (!flag) {
					byte function = _codedExpression[i];
					double ans;
					if (isBinaryFunction(function)) {
						double operand2 = _stack[--top];
						double operand1 = _stack[--top];
						ans = performBinaryFunction(function, operand1, operand2);
					} else {
						double operand1 = _stack[--top];
						ans = performUnaryFunction(function, operand1);
					}
					if (Double.isNaN(ans)) {
						throw new ArianeArithmeticException(Ariane.getString("arithmetic.error17"));
					}
					_stack[top++] = ans;
				}
			}
		}
		if (Double.isInfinite(_stack[0])) {
			throw new ArianeArithmeticException(Ariane.getString("arithmetic.error17"));
		} else {
			return _stack[0];
		}
	}

	private boolean isBinaryFunction( byte function ) {
		return function == MIN || function == MAX;
	}

	private double performBinaryFunction( byte function, double operand1, double operand2 ) {
		switch (function) {
		case MIN:
			return operand1 < operand2 ? operand1 : operand2;
		case MAX:
			return  operand1 > operand2 ? operand1 : operand2;
		default:
			return Double.NaN;
		}
	}

	private double performUnaryFunction( byte function, double operand1 ) {
		switch (function) {
		case SIN: return Math.sin(operand1);
		case COS: return Math.cos(operand1);
		case TAN: return Math.tan(operand1);
		case ASIN: return Math.asin(operand1);
		case ACOS: return Math.acos(operand1); 
		case ATAN: return Math.atan(operand1);
		case EXP: return Math.exp(operand1);
		case LN: return Math.log(operand1);
		case LOG2: return Math.log(operand1) / Math.log(2);
		case LOG10: return Math.log(operand1) / Math.log(10);
		case ABS: return Math.abs(operand1);
		case SQRT: return Math.sqrt(operand1);
		case UNARYMINUS: return -operand1;
		default: return Double.NaN;
		}
	}

	private double performBasicBinaryOperation( byte operator, double operande1, double operande2 ) {
		switch (operator) {
		case PLUS: return operande1 + operande2;
		case MINUS: return operande1 - operande2;
		case TIMES: return operande1 * operande2;
		case DIVIDE: return operande1 / operande2;
		case MODULO: return operande1 % operande2;
		case POWER: return Math.pow(operande1, operande2);
		default:
			return Double.NaN;
		}
	}


	private void error( String message ) throws ArianeArithmeticException {
		throw new ArianeArithmeticException(Ariane.getString("arithmetic.error0.1") + " " + message + " " + Ariane.getString("arithmetic.error0.2") + " " + _pos + ")");
	}

	private int computeStackUsage( ) {
		int size = 0;
		int max = 0;
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0 || isVariable(i)) {
				size++;
				if (size > max) {
					max = size;
				}
			} else if (_codedExpression[i] >= POWER) {
				size--;
			}
		}
		return max;
	}

	private boolean isVariable( int rank ) {
		return _codedExpression[rank] <= -34 && _codedExpression[rank] > -34 - _nbParameters;
	}

	private char nextCharacter() {
		return _pos >= _initialExpression.length() ? 0 : _initialExpression.charAt(_pos);
	}

	private void skipWhiteSpaces() {
		while (Character.isWhitespace(nextCharacter())) {
			_pos++;
		}
	}

	/**
	 * Parses a mathematical expression such as:
	 * [+|-]Term[+|-]Term.
	 */
	private void parseFormula( ) throws ArianeArithmeticException {
		boolean negative = false;
		skipWhiteSpaces();
		// Signe plus ou moins au debut de l'expression
		if (nextCharacter() == '+' || nextCharacter() == '-') {
			negative = nextCharacter() == '-';
			_pos++;
			skipWhiteSpaces();
		}
		parseTerm();
		if (negative) {
			_codedExpression[_codeSize++] = UNARYMINUS;
		}
		skipWhiteSpaces();
		while (nextCharacter() == '+' || nextCharacter() == '-') {
			char op = nextCharacter();
			_pos++;
			parseTerm();
			_codedExpression[_codeSize++] = (op == '+') ? PLUS : MINUS;
			skipWhiteSpaces();
		}
	}

	/**
	 * Parses an expression such as :
	 * Primary [*|/|^] Primary.
	 */
	private void parseTerm( ) throws ArianeArithmeticException {
		parsePrimary();
		skipWhiteSpaces();
		while (nextCharacter() == '*' || nextCharacter() == '/' || nextCharacter() == '^' || nextCharacter() == '%') {
			char op = nextCharacter();
			_pos++;
			parsePrimary();
			if (op == '*') {
				_codedExpression[_codeSize++] = TIMES;
			} else if (op == '/') {
				_codedExpression[_codeSize++] = DIVIDE;
			} else if (op == '^') {
				_codedExpression[_codeSize++] = POWER;
			} else if (op == '%') {
				_codedExpression[_codeSize++] = MODULO;
			}
			skipWhiteSpaces();
		}
	}

	/**
	 * Parses an expression such as:
	 * [VARIABLE1 | VARIABLE2| FUNCTION | NUMBER | (FORMULA) ].
	 */
	private void parsePrimary( ) throws ArianeArithmeticException {
		skipWhiteSpaces();
		char ch = nextCharacter();
		if (ch == '$') {
			_pos++;
			if ( Integer.valueOf(nextCharacter()) < (48 + _nbParameters) && Integer.valueOf(nextCharacter()) >= 48) {
				_codedExpression[_codeSize++] = _variables[-48 + Integer.valueOf(nextCharacter())];
			} else { //rajouter des conditions. Du genre while
				error(Ariane.getString("arithmetic.error2") + " " + ch + nextCharacter());
			}
			_pos++;
		} else if (Character.isLetter(ch)) {
			parseFunction();
		} else if (Character.isDigit(ch) || ch == '.') {
			parseNumber();
		} else if (ch == '(') {
			_pos++;
			parseFormula();
			skipWhiteSpaces();
			if (nextCharacter() != ')') {
				error(Ariane.getString("arithmetic.error3"));
			}
			_pos++;
		} else if (ch == ')') {
			error(Ariane.getString("arithmetic.error4"));
		} else if (ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '^') {
			error(Ariane.getString("arithmetic.error5.1") + ch + Ariane.getString("arithmetic.error5.2"));
		} else if (ch == 0) {
			error(Ariane.getString("arithmetic.error6"));
		} else {
			error(Ariane.getString("arithmetic.error7.1") + ch + Ariane.getString("arithmetic.error7.2"));
		}
	}

	private void parseFunction( ) throws ArianeArithmeticException {
		String fonctionName = "";
		while (Character.isLetterOrDigit(nextCharacter())) {
			fonctionName += nextCharacter();
			_pos++;
		}
		fonctionName = fonctionName.toLowerCase();
		for (int i = 0; i < _functionNamesUnary.length; i++) {
			if (fonctionName.equals(_functionNamesUnary[i])) {
				skipWhiteSpaces();
				if (nextCharacter() != '(') {
					error(Ariane.getString("arithmetic.error8.1`") + fonctionName + Ariane.getString("arithmetic.error8.2"));
				}
				_pos++;
				parseFormula();
				skipWhiteSpaces();
				if (nextCharacter() != ')') {
					error(Ariane.getString("arithmetic.error9") + fonctionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(SIN - i);
				return;
			}
		}
		for (int i = 0; i < _functionNamesBinary.length; i++) {
			if (fonctionName.equals(_functionNamesBinary[i])) {
				skipWhiteSpaces();
				if (nextCharacter() != '(') {
					error(Ariane.getString("arithmetic.error10.1") + fonctionName + Ariane.getString("arithmetic.error10.2"));
				}
				_pos++;
				parseFormula();
				skipWhiteSpaces();
				if (nextCharacter() != ',') {
					error(Ariane.getString("arithmetic.error11") + fonctionName + "'.");
				}
				_pos++;
				parseFormula();
				skipWhiteSpaces();
				if (nextCharacter() != ')') {
					error(Ariane.getString("arithmetic.error12") + fonctionName + "'.");
				}
				_pos++;
				_codedExpression[_codeSize++] = (byte)(MIN - i);
				return;
			}
		}
		error(Ariane.getString("arithmetic.error13") + fonctionName );
	}

	private String parseInteger() {
		String numerical = "";
		while (Character.isDigit(nextCharacter())) {
			numerical += nextCharacter();
			_pos++;
		}
		return numerical;
	}

	private String parseFloat() {
		String numerical = parseInteger();
		if (nextCharacter() == '.') {
			numerical += '.';
			_pos++;
			numerical += parseInteger();
		}
		return numerical;
	}

	private void parseNumber() throws ArianeArithmeticException {
		String number = parseFloat();
		if (".".equals(number)) {
			error(Ariane.getString("arithmetic.error14"));
		}
		if (nextCharacter() == 'E' || nextCharacter() == 'e') {
			number += nextCharacter();
			_pos++;
			if (nextCharacter() == '+' || nextCharacter() == '-') {
				number += nextCharacter();
				_pos++;
			}
			if (!Character.isDigit(nextCharacter())) {
				error(Ariane.getString("arithmetic.error15"));
			}
			number += parseInteger();
		}
		double value;
		try {
			value = Double.parseDouble(number);
		} catch (NumberFormatException e) {
			value = Double.NaN;
			error(Ariane.getString("arithmetic.error16.1") + number + Ariane.getString("arithmetic.error16.2"));
		}
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = value;
	}

	@Override
	public String toString() {
		return _initialExpression;
	}
}