/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;

/**
 * A <code>LogicalExpression</code> is defined from a
 * boolean formula that includes logical operators and
 * and the two parameters values ($0 and $1).
 */
public final class LogicalExpression {
	private static final byte GT = -1;
	private static final byte LT = -2;
	private static final byte GE = -3;
	private static final byte LE = -4;
	private static final byte EQ = -5;
	private static final byte NEQ = -6;
	private static final byte AND = -7;
	private static final byte OR = -8;
	private static final byte XOR = -9;
	private static final byte NOT = -10;
	private static final byte VARIABLE1 = -22;
	private static final byte VARIABLE2 = -23;

	private final String _initialExpression;
	private byte[] _codedExpression;
	private final boolean[] _stack;
	private double[] _constantStack;

	private int _position;
	private int _constantStackIndex;
	private int _codeSize;

	public LogicalExpression( String formula ) throws ArianeArithmeticException {
		_initialExpression = formula;
		if (formula == null || formula.trim().isEmpty()) {
			_stack = null;
			return;
		}

		_codedExpression = new byte[formula.length()];
		_constantStack = new double[formula.length()];
		_position = 0;
		_constantStackIndex = 0;
		_codeSize = 0;

		parseFormula();
		skipWhiteSpaces();
		if (nextCharacter() != 0) {
			error(Ariane.getString("boolean.error1"));
		}
		int stackSize = computeStackUsage();
		_stack = new boolean[stackSize];
		byte[] code = new byte[_codeSize];
		System.arraycopy(_codedExpression, 0, code, 0, _codeSize);
		_codedExpression = code;
		double[] tmp = new double[_constantStackIndex];
		System.arraycopy(_constantStack, 0, tmp, 0, _constantStackIndex);
		_constantStack = tmp;
	}

	public boolean evaluate( double value1, double value2 ) throws ArianeArithmeticException {
		double[] doubleStack = new double[3];
		int doubleTop = 0;
		int top = 0;
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0) {
				doubleStack[doubleTop++] = _constantStack[_codedExpression[i]];
			} else if (_codedExpression[i] == NOT) {
				boolean operand1 = _stack[--top];
				_stack[top++] = !operand1;
			} else if (_codedExpression[i] > XOR && _codedExpression[i] <= AND) {
				boolean operand2 = _stack[--top];
				boolean operand1 = _stack[--top];
				_stack[top++] = performBinaryLogicalConnector(_codedExpression[i], operand1, operand2);
			} else if (_codedExpression[i] >= NEQ && _codedExpression[i] <= GT) {
				double operand2 = doubleStack[--doubleTop];
				double operand1 = doubleStack[--doubleTop];
				_stack[top++] = performBinaryLogicalOperator(_codedExpression[i], operand1, operand2);
			} else if (_codedExpression[i] == VARIABLE1) {
				doubleStack[doubleTop++] = value1;
			} else if (_codedExpression[i] == VARIABLE2) {
				doubleStack[doubleTop++] = value2;
			}
		}
		return _stack[0];
	}

	private boolean performBinaryLogicalOperator( byte operator, double operand1, double operand2 ) {
		switch (operator) {
		case GT: return operand1 > operand2;
		case GE: return operand1 >= operand2;
		case LT: return operand1 < operand2;
		case LE: return operand1 <= operand2;
		case EQ: return operand1 == operand2;
		case NEQ: return operand1 != operand2;
		default: return false;
		}
	}

	private boolean performBinaryLogicalConnector( int connector, boolean operand2, boolean operand1 ) {
		switch (connector) {
		case AND: return operand1 && operand2;
		case OR: return operand1 || operand2;
		case XOR: return (operand1 || operand2) && !(operand1 && operand2);
		default: return false;
		}
	}

	private void error( String message ) throws ArianeArithmeticException { 
		throw new ArianeArithmeticException(Ariane.getString("boolean.error0.1") + " " + message + Ariane.getString("boolean.error0.2") + " " + _position + ")");
	}

	private int computeStackUsage( ) {
		int size = 0;
		int max = 0;
		for (int i = 0; i < _codeSize; i++) {
			if (_codedExpression[i] >= 0 || _codedExpression[i] == VARIABLE1 || _codedExpression[i] == VARIABLE2) {
				size++;
				if (size > max) {
					max = size;
				}
			}
		}
		return max;
	}

	private char nextCharacter( ) { 
		return _position >= _initialExpression.length() ? 0 : _initialExpression.charAt(_position);
	}

	private void skipWhiteSpaces( ) {
		while (Character.isWhitespace(nextCharacter())) {
			_position++;
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * [!]Term [ '&' | '|' | '^' ] [!]Term.
	 */
	private void parseFormula( ) throws ArianeArithmeticException {
		boolean negative = false;
		skipWhiteSpaces();
		// Signe not au debut de l'expression
		if (nextCharacter() == '!' ) {
			negative = true;
			_position++;
			skipWhiteSpaces();
		}
		parseTerm();
		if (negative) {
			_codedExpression[_codeSize++] = NOT;
		}
		skipWhiteSpaces();
		while (nextCharacter() == '&' || nextCharacter() == '|' || nextCharacter() == '^') {
			char code = nextCharacter();
			_position++;
			negative = false;
			if (nextCharacter() == '!')  {
				negative = true;
				_position++;
				skipWhiteSpaces();
			}
			parseTerm();
			if (negative) {
				_codedExpression[_codeSize++] = NOT;
			}
			if (code == '&') {
				_codedExpression[_codeSize++] = AND;
			} else if (code == '|') {
				_codedExpression[_codeSize++] = OR;
			} else if (code == '^') {
				_codedExpression[_codeSize++] = XOR;
			}
			skipWhiteSpaces();
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * TERM :: NUMERIC | (FORMULA).
	 *</pre>
	 */
	private void parseTerm( ) throws ArianeArithmeticException {
		skipWhiteSpaces();
		char ch = nextCharacter();
		if (ch == '$') { // VARIABLE
			parsePrimary();
		} else if (Character.isDigit(ch) || ch == '.') { // NUMBER
			parsePrimary(); 
		} else if (ch == '(') { // (FORMULA)
			_position++;
			parseFormula();
			skipWhiteSpaces();
			if (nextCharacter() != ')') {
				error(Ariane.getString("boolean.error2"));
			}
			_position++;
		} else if (ch == ')') {
			error(Ariane.getString("boolean.error3"));
		} else if (ch == '&' || ch == '|' || ch == '^' || ch == '=' || ch == '!') {
			error(Ariane.getString("boolean.error4.1") + ch + Ariane.getString("boolean.error4.2"));
		} else if (ch == 0) {
			error(Ariane.getString("boolean.error5"));
		} else {
			error(Ariane.getString("boolean.error6.1") + ch + Ariane.getString("boolean.error6.2"));
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * PRIMARY :: NUMERIC '>' | '<' | '<=' | '>=' | '==' | '!='  NUMERIC.
	 *</pre>
	 */
	private void parsePrimary( ) throws ArianeArithmeticException {
		parseNumeric();
		skipWhiteSpaces();
		if (nextCharacter() == '<' || nextCharacter() == '>' || nextCharacter() == '=' || nextCharacter() == '!') {
			char op = nextCharacter();
			byte code = LE;
			_position++;
			if (nextCharacter() == '=') {
				_position++;
				if (op == '<') {
					code = LE;
				} else if (op == '>') {
					code = GE;
				} else if (op == '=') {
					code = EQ;
				} else { // if (op == '!')
					code = NEQ;
				}
			} else {
				if (op == '<') {
					code = LT;
				} else if (op == '>') {
					code = GT;
				} else {
					error(Ariane.getString("boolean.error7"));
				}
			}
			parseNumeric();
			_codedExpression[_codeSize++] = code;
		} else {
			error(Ariane.getString("boolean.error8"));
		}
	}

	/**
	 * Parses a boolean expression such as:
	 * <pre>
	 * NUMERIC :: VARIABLE1 | VARIABLE2| NUMBER
	 * ou VARIABLE :: $0 | $1.
	 *</pre>
	 */
	private void parseNumeric( ) throws ArianeArithmeticException {
		skipWhiteSpaces();
		char ch = nextCharacter();
		// VARIABLE
		if (ch == '$') {
			_position++;
			if (nextCharacter() == '0') {
				_codedExpression[_codeSize++] = VARIABLE1;
			} else if (nextCharacter() == '1') {
				_codedExpression[_codeSize++] = VARIABLE2;
			} else {
				error(Ariane.getString("boolean.error9") + " " + ch + _position);
			}
			_position++;
		} else if (Character.isDigit(ch) || ch == '.' || ch == '-') { // NUMBER
			parseNumber();
		} else {
			error(Ariane.getString("boolean.error10"));
		}
	}

	private void parseNumber( ) throws ArianeArithmeticException {
		char x = nextCharacter();
		String number;
		if (x == '-') {
			number = "-";
			_position++;
		} else {
			number = "";
		}
		number += parseFloat();
		if (".".equals(number)) {
			error(Ariane.getString("boolean.error11"));
		}
		if (nextCharacter() == 'E' || nextCharacter() == 'e') {
			number += nextCharacter();
			_position++;
			if (nextCharacter() == '+' || nextCharacter() == '-') {
				number += nextCharacter();
				_position++;
			}
			if (!Character.isDigit(nextCharacter())) {
				error(Ariane.getString("boolean.error12"));
			}
			number += parseInteger();
		}
		double value;
		try {
			value = Double.parseDouble(number);
		} catch (NumberFormatException e) {
			value = Double.NaN;
		}
		if (Double.isNaN(value)) {
			error(Ariane.getString("boolean.error13.1")
					+ number
					+ Ariane.getString("boolean.error13.2"));
		}
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = value;
	}

	private String parseFloat( ) {
		String number = parseInteger();
		if (nextCharacter() == '.') {
			number += nextCharacter();
			_position++;
			number += parseInteger();
		}
		return number;
	}

	private String parseInteger( ) {
		String number = "";
		while (Character.isDigit(nextCharacter())) {
			number += nextCharacter();
			_position++;
		}
		return number;
	}

	public String toHtml() {
		return _initialExpression.replaceAll("<", "&lt;");
	}

	@Override
	public String toString() {
		return _initialExpression;
	}
}