/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.view.graph.slot.InputMaskSlotView;

import java.io.IOException;

/**
 * The class <code>InputMaskSlot</code> defines an input mask connection slot.
 * It can only be connected to an <code>OutputDataSlot</code>.
 */
public final class InputMaskSlot extends InputSlot {
	
	public InputMaskSlot( GraphItem item ) {
		super(new InputMaskSlotView(), item, 0);
	}

	@Override
	protected Link createLink() {
		return new MaskLink(this);
	}

	@Override
	public String toString( ) {
		return "InputMaskSlot";
	}

	///// 
	/////  R E S U L T
	/////

	/**
	 * Tries to get the data of the connected slot.
	 * Result exists if:
	 * <ul>
	 * <li>the slot is connected;
	 * <li>the result of the connected slot is validate.
	 * </ul>
	 * 
	 * @param isEmptyFatal is an empty value a fatal error?
	 * @return true if the result is validate.
	 */
	public boolean hasValue( boolean isEmptyFatalError ) throws ArianeRuntimeException {
		if (isConnected()) {
			return ((OutputDataSlot)getConnectedSlot()).hasValue();
		} else if (isEmptyFatalError) {
			throw new ArianeRuntimeException(Ariane.getString("Operator.error3") + getRank());
		}
		return false;
	}

	public String getValue( ) {
		return ((OutputDataSlot)getConnectedSlot()).getValue();
	}

	@Override
	public boolean isAcceptableDropSite( Slot slot ) {
		return slot.getGraphItem() != this.getGraphItem() && slot instanceof OutputDataSlot;
	}

	@Override
	public XMLElement convertToXML( ) throws IOException {
		if (isConnected()) {
			XMLElement element = new XMLElement("Connection");
			element.addChildren(new XMLElement("EndItemId", Integer.toString(getGraphItem().getId())));
			element.addChildren(new XMLElement("EndSlotType", "InSlotMask"));
			element.addChildren(new XMLElement("EndSlotIndex", Integer.toString(getRank())));
			getConnectedSlot().convertToXML(element);
			if (!getLink().getPath().isEmpty()) {
				element.addChildren(new XMLElement("UserDefined", Boolean.toString(getLink().isUserDefined())));
				XMLElement pa = new XMLElement("LinkPath");
				for (int i = 0; i < getLink().getPath().size(); i++) {
					XMLElement po = new XMLElement("LinkPoint");
					po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
					po.addChildren(new XMLElement("X-coordinate", Integer.toString((int)getLink().getPath().get(i).getX())));
					po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int)getLink().getPath().get(i).getY())));
					pa.addChildren(po);
				}
				element.addChildren(pa);
			}
			return element;
		}
		return null;
	}
}