/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.command.ChangeParameterValueEdit;
import fr.greyc.pantheon.ariane.view.graph.slot.InputParameterSlotView;

import java.io.IOException;

/**
 * The class <code>InputParameterSlot</code> defines an input parameter connection slot.
 * An input parameter slot can be connected to a OutputResultSlot
 * or can hold its own value (either float or String value).
 */
public final class InputParameterSlot extends InputSlot {
	private String _value;
	private String _name;

	public InputParameterSlot( GraphItem item, int rank ) {
		super(new InputParameterSlotView(), item, rank);
	}

	@Override
	protected Link createLink() {
		((InputParameterSlotView)getView()).setEnabledTextfield(false);
		return new ValueLink(this);
	}

	@Override
	public void disposeLink() {
		_value = null;
		((InputParameterSlotView)getView()).setEnabledTextfield(true);
	}

	@Override
	public String toString() {
		return "Parameter $" + getRank();
	}

	///// 
	/////  V A L U E 
	/////

	/**
	 * Tests if the parameter has a validate value.
	 * If it is connected, tests if the connected slot
	 * has a validate value.
	 * 
	 * @param isEmptyFatal is an empty value a fatal error?
	 * @return true if the parameter has a validate value.
	 */
	public boolean hasValue( boolean isEmptyFatalError ) throws ArianeMissingParameterException {
		if (isConnected()) {
			// Gets the value of the connected slot.
			OutputResultSlot outputSlot = (OutputResultSlot)getConnectedSlot();
			if (outputSlot.hasValue()) {
				setValue(outputSlot.getValue());
			} else {
				return false;
			}
		} else if (isEmptyFatalError && _value == null) {
			throw new ArianeMissingParameterException(Ariane.getString("Operator.error1") + getRank());
		}
		return _value != null;
	}

	/**
	 * Tests if the parameter has a float value and is validated.
	 * If it is connected tests if the connected slot
	 * has a float value and is validated.
	 * 
	 * @param isEmptyFatal is an empty value a fatal error?
	 * @return true if the parameter has a validate value and is a float.
	 */
	public boolean hasNumericalValue( boolean isEmptyFatalError ) throws ArianeMissingParameterException {
		if (hasValue(isEmptyFatalError)) {
			try {
				Float.valueOf(getValue()); 
				return true;
			} catch (NumberFormatException e) {
				return false;
			}
		} else {
			return false;
		}
	}

	public void setName( String name ) {
		_name = name;
	}

	public String getName() {
		if (_name == null) {
			return Ariane.getString("PropertyViewer.parameter") + " $" + getRank();
		} else {
			return _name;
		}
	}

	public boolean hasName() {
		return _name != null;
	}

	public void setValue( String value ) {
		if (value != null && !value.isEmpty()) {
			_value = value.trim();
		} else {
			_value = null;
		}
		((InputParameterSlotView)getView()).updateTextfield(_value);
	}

	public String getValue() {
		return _value;
	}

	public float getNumericalValue() { 
		try {
			return Float.valueOf(getValue());
		} catch (NumberFormatException e) {
			Ariane.getPresenter().printErrorMessage(Ariane.getString("Slot.error4"));
			return 0.0f;
		}
	}

	public void changeValue( String value, boolean addInHistory) {
		if (value == null && getValue() == null) {
			return;
		}
		if (value == null || !value.equals(getValue())) {
			final String oldValue = getValue();
			setValue(value);
			getGraphItem().resetAll();
			getGraphItem().getView().updatePropertyViewer();
			getGraph().setModified(true);
			if (addInHistory) {
				getGraph().validateOneAction(new ChangeParameterValueEdit(this, value, oldValue));
			}
		}
	}

	@Override
	public boolean isAcceptableDropSite( Slot slot ) {
		return slot instanceof OutputResultSlot;
	}

	@Override
	public XMLElement convertToXML() throws IOException {
		XMLElement element = new XMLElement("Connection");
		element.addChildren(new XMLElement("EndItemId", Integer.toString(getGraphItem().getId())));
		element.addChildren(new XMLElement("EndSlotType", "InSlotParameter"));
		element.addChildren(new XMLElement("EndSlotIndex", Integer.toString(getRank())));
		if (isConnected()) {
			getConnectedSlot().convertToXML(element);
			if (!getLink().getPath().isEmpty()) {
				element.addChildren(new XMLElement("UserDefined", Boolean.toString(getLink().isUserDefined())));
				XMLElement pa = new XMLElement("LinkPath");
				for (int i = 0; i < getLink().getPath().size(); i++) {
					XMLElement po = new XMLElement("LinkPoint");
					po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
					po.addChildren(new XMLElement("X-coordinate", Integer.toString((int)getLink().getPath().get(i).getX())));
					po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int)getLink().getPath().get(i).getY())));
					pa.addChildren(po);
				}
				element.addChildren(pa);
			}
		} else {
			element.addChildren(new XMLElement("ParameterValue", getValue()));
		}
		return element;
	}
}