/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeSlotException;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.command.ConnectItemEdit;
import fr.greyc.pantheon.ariane.presenter.command.DisconnectItemEdit;
import fr.greyc.pantheon.ariane.view.graph.slot.SlotView;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * The class <code>InputSlot</code> defines input connection slot.
 * Only one <code>OutputSlot</code> can be connected to one
 * <code>InputSlot</code>. It is a drop site for an <code>OutputSlot</code>.
 * Three types of input slots are available:
 * <ul>
 * <li>@see InputdataSlot
 * <li>@see InputMaskSlot
 * <li>@see InputParameterSlot
 * </ul>
 */
public abstract class InputSlot extends Slot {
	private transient OutputSlot _connectedSlot;
	private transient Link _link;

	public InputSlot( SlotView slotView, GraphItem item, int rank ) {
		super(slotView, item, rank);
	}

	public final Link getLink() {
		return _link;
	}

	protected final void setLink( Link link ) {
		_link = link;
	}

	///// 
	///// C O N N E C T I O N   /   D I S C O N N E C T I O N 
	/////

	@Override
	public final void connectWith( Slot from ) {
		try {
			OutputSlot outputSlot = (OutputSlot)from;
			// check if the slot are in the same graph!
			if (getGraph() != outputSlot.getGraph()) {
				throw new ArianeSlotException(Ariane.getString("Slot.error3"));
			}
			// check slots consistency.
			boolean isConnectable = this instanceof InputParameterSlot && outputSlot instanceof OutputResultSlot
					|| this instanceof InputDataSlot && outputSlot instanceof OutputDataSlot
					|| this instanceof InputMaskSlot && outputSlot instanceof OutputDataSlot;
			if (isConnectable) {
				getGraph().startActionEdition();
				outputSlot.connectTo(this, null, true);
				getGraph().endActionEdition();
			} else {
				throw new ArianeSlotException(Ariane.getString("Slot.error2"));
			}
		} catch (ArianeSlotException e) {
			Ariane.getPresenter().printErrorMessage(Ariane.getString("Slot.error1") + e.getMessage());
		} catch (ClassCastException e) {
			Ariane.getPresenter().printErrorMessage(Ariane.getString("Slot.error1") + Ariane.getString("Slot.error2"));
		}
	}

	protected final void connectFrom( OutputSlot outputSlot ) {
		if (!outputSlot.equals(_connectedSlot)) {
			disconnect();
			outputSlot.connectTo(this, null, true);
		}
	}

	public final OutputSlot getConnectedSlot() {
		return _connectedSlot;
	}

	@Override
	public final List<? extends Slot> getConnectedSlots() {
		List<Slot> slots = new ArrayList<Slot>();
		slots.add(_connectedSlot);
		return slots;
	}

	@Override
	public final boolean isConnected() {
		return _connectedSlot != null;
	}

	protected final void setConnectedSlot( OutputSlot slot, Link link, boolean isAddedToHistory ) {
		getGraphItem().resetAll();
		getGraphItem().getGraph().setModified(true);
		boolean isDeconnection = slot == null;
		if (isDeconnection) {
			getView().setDisconnectedIcon();
			if (isAddedToHistory) {
				getGraph().validateAction(new DisconnectItemEdit(getGraph(), this, _connectedSlot));
			}
			_link.dispose();
			getGraphItem().fireConnection(this, false);
		} else {
			getView().setConnectedIcon();
			if (link != null ) {
				_link = link;
			} else {
				_link = createLink();
			}
			if (isAddedToHistory) {
				getGraph().validateAction(new ConnectItemEdit(getGraph(), this, slot));
			}
			getGraphItem().fireConnection(this, true);
		}
		_connectedSlot = slot;
	}

	public final void setConnectedSlot( OutputSlot slot ) { // For tests
		_connectedSlot = slot;
	}

	@Override
	public final void disconnectAll() {
		disconnect();
	}

	public final void disconnect() {
		if (_connectedSlot != null) {
			_connectedSlot.disconnectFrom(this, true);
			disposeLink();
		}
	}

	protected abstract Link createLink();

	public void disposeLink() { 
		/* empty */
	}

	public abstract XMLElement convertToXML() throws IOException;
}