/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.tree; // NOPMD by regis on 4/25/15 5:34 PM

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;

import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

/**
 * The class <code>Tree</code> contains all the available graph items
 * organized in a tree.
 * This tree is displayed in the explorer from where the user can drag and
 * drop items into the graph editor.
 */
public final class ItemTreeModel extends DefaultTreeModel {
	private final MutableTreeItem _builtinSubtree = new MutableTreeItem(null, true);
	private final MutableTreeItem _commonOperatorSubtree = new MutableTreeItem(null, true);
	private MutableTreeItem _userOperatorSubtree = new MutableTreeItem(null, true);
	private MutableTreeItem _subRoutineSubtree = new MutableTreeItem(null, true);
	private String _pathToHTMLLibraryDocumentationFile;

	public ItemTreeModel( ) {
		super(new MutableTreeItem("Library", true));
		initContents();
	}

	private void initContents( ) {
		createBuiltinItems();
		MutableTreeItem rootNode = (MutableTreeItem)getRoot();
		rootNode.add(_builtinSubtree);
		rootNode.add(_commonOperatorSubtree);
		rootNode.add(_subRoutineSubtree);
		rootNode.add(_userOperatorSubtree);
	}

	private void createBuiltinItems( ) {
		for (TreeItem.Item itemType : TreeItem.Item.values()) {
			if (itemType != TreeItem.Item.OPERATOR) {
				TreeItem treeItem = getItem(itemType);
				MutableTreeItem fileItem = new MutableTreeItem(treeItem, false);
				_builtinSubtree.add(fileItem);
			}
		}
	}

	public static TreeItem getItem( TreeItem.Item itemType ) {
		switch (itemType) {
		case LOADER: return new OperatorTreeItem(TreeItem.Item.LOADER, "loader", "loader.html", 0, 0, 1, "");
		case SAVER: return new OperatorTreeItem(TreeItem.Item.SAVER, "saver", "saver.html", 1, 0, 0, "");
		case VIEWER: return new OperatorTreeItem(TreeItem.Item.VIEWER, "viewer", "viewer.html", 1, 0, 0, "");
		case FILENAME: return new OperatorTreeItem(TreeItem.Item.FILENAME, "filename", "filename.html", 0, 0, 0, "");
		case CONSTANT: return new OperatorTreeItem(TreeItem.Item.CONSTANT, "constant", "constant.html",  0, 0, 0, "");
		case SCRIPT: return new OperatorTreeItem(TreeItem.Item.SCRIPT, "script", "script.html", 0, 0, 0, "");
		case EVAL: return new OperatorTreeItem(TreeItem.Item.EVAL, "eval", "eval.html", 0, 2, 0, "");
		case FORMAT: return new OperatorTreeItem(TreeItem.Item.FORMAT, "format", "format.html", 0, 2, 0, "");
		case SYSTEM: return new OperatorTreeItem(TreeItem.Item.SYSTEM, "system", "system.html", 0, 0, 0, "");
		case IF_CONTROL: return new OperatorTreeItem(TreeItem.Item.IF_CONTROL, "if", "if.html", 1, 2, 2, "");
		case SELECT_CONTROL: return new OperatorTreeItem(TreeItem.Item.SELECT_CONTROL, "select", "select.html", 0, 0, 1, "$0:index");
		case MERGE_CONTROL: return new OperatorTreeItem(TreeItem.Item.MERGE_CONTROL, "merge", "merge.html", 2, 0, 1, "");
		case FOR_CONTROL: return new OperatorTreeItem(TreeItem.Item.FOR_CONTROL, "for", "for.html", 3, 0, 0, "$0:initial value;$1:final value;$2:increment");
		case WHILE_CONTROL: return new OperatorTreeItem(TreeItem.Item.WHILE_CONTROL, "while", "while.html", 2, 0, 0, ""); // TODO ajouter la signature.
		case MACRO_OPERATOR: return new OperatorTreeItem(TreeItem.Item.MACRO_OPERATOR, "macro-operator", "macro.html", 0, 0, 0, "");
		case NOTE: return new OperatorTreeItem(TreeItem.Item.NOTE, "note", "note.html", 0, 0, 0, "");
		default: return null;
		}
	}

	public TreeItem getTreeItem( String name, String category ) {
		return getNode((MutableTreeItem)getRoot(), name, category);
	}

	private TreeItem getNode( MutableTreeItem subtree, String name, String category ) {
		for (final Enumeration<?> e = subtree.children(); e.hasMoreElements();) {
			MutableTreeItem node = (MutableTreeItem)e.nextElement();
			if (node.isLeaf()) {
				TreeItem treeItem = node.getTreeItemObject();
				if (name.equals(treeItem.getName()) && (category == null || category.equals(treeItem.getCategory()))) {
					return treeItem;
				}
			} else {
				TreeItem item = getNode(node, name, category);
				if (item != null) {
					return item;
				}
			}
		}
		return null;
	}

	public void setInternationalText( ) {
		_subRoutineSubtree.setUserObject(Ariane.getString("Tree.subroutines"));
		_userOperatorSubtree.setUserObject(Ariane.getString("Tree.personalOperators"));
		_commonOperatorSubtree.setUserObject(Ariane.getString("Tree.operators"));
		_builtinSubtree.setUserObject(Ariane.getString("Tree.builtinItems"));
		reload();
	}

	public String getPathToLibraryDocumentationFile( ) {
		return _pathToHTMLLibraryDocumentationFile;
	}

	private List<TreeItem> subtreeToList( MutableTreeItem subtree ) {
		ArrayList<TreeItem> treeList = new ArrayList<TreeItem>();
		Enumeration<?> iterator = subtree.depthFirstEnumeration();
		while (iterator.hasMoreElements()) {
			MutableTreeItem node = (MutableTreeItem)iterator.nextElement();
			if (node.getUserObject() instanceof TreeItem) {
				treeList.add(node.getTreeItemObject());
			}
		}
		return treeList;
	}

	public List<TreeItem> getUserOperatorTreeAsList( ) {
		return subtreeToList(_userOperatorSubtree);
	}

	public List<TreeItem> getSubroutineTreeAsList( ) {
		return subtreeToList(_subRoutineSubtree);
	}

	public MutableTreeItem getUserOperatorTree( ) {
		return _userOperatorSubtree;
	}

	public MutableTreeItem getSubroutineTree( ) {
		return _subRoutineSubtree;
	}

	private MutableTreeItem getCommonOperatorTree( ) {
		return _commonOperatorSubtree;
	}

	private MutableTreeItem changeSubtree( MutableTreeItem oldSubtree, MutableTreeItem newSubtree ) {
		MutableTreeItem rootNode = (MutableTreeItem)getRoot();
		int index = getIndexOfChild(rootNode, oldSubtree);
		removeNodeFromParent(oldSubtree);
		insertNodeInto(newSubtree, rootNode, index);
		return newSubtree;
	}

	public void setUserOperatorTree( MutableTreeItem newTree ) {
		_userOperatorSubtree = changeSubtree(_userOperatorSubtree, newTree);
	}

	public void setSubroutineTree( MutableTreeItem newTree ) {
		_subRoutineSubtree = changeSubtree(_subRoutineSubtree, newTree);
	}

	public void setCommonOperators( List<XMLElement> readItems ) {
		addInCommonOperators(loadItemsFromXMLTree(readItems));
	}

	public void addInCommonOperators( List<TreeItem> operatorList ) {
		_commonOperatorSubtree.removeAllChildren();
		addItemsInSubtree(operatorList, _commonOperatorSubtree);
	}

	public void addInUserOperators( List<XMLElement> operatorList ) {
		addItemsInSubtree(loadItemsFromXMLTree(operatorList), _userOperatorSubtree);
	}

	public MutableTreeItem addInSubroutines( List<XMLElement> routinesList ) {
		addItemsInSubtree(loadItemsFromXMLTree(routinesList), _subRoutineSubtree);
		return _subRoutineSubtree;
	}

	public void addSubroutine( TreeItem item ) {
		ArrayList<TreeItem> treeItems = new ArrayList<TreeItem>();
		treeItems.add(item);
		addItemsInSubtree(treeItems, _subRoutineSubtree);
	}

	public String[] getSubroutineCategories( ) {
		String[] categories = new String[_subRoutineSubtree.getChildCount()];
		for (int i = 0; i < _subRoutineSubtree.getChildCount(); i++) {
			categories[i] = _subRoutineSubtree.getChildAt(i).toString();
		}
		return categories;
	}

	private void addItemsInSubtree( List<TreeItem> itemList, MutableTreeItem subtree ) {
		for (TreeItem treeItem : itemList) {
			String categoryName = treeItem.getCategory();
			MutableTreeItem category = subtree.getNodeWithSpecifiedName(categoryName);

			if (category == null) { // create a new category.
				category = new MutableTreeItem( categoryName, true);
				subtree.add(category);
			} else if (category.getNodeWithSpecifiedName(treeItem.getName()) != null) {
				category.remove(category.getNodeWithSpecifiedName(treeItem.getName()));
			}
			MutableTreeItem item = new MutableTreeItem(treeItem, false);
			category.add(item); // replace existing subroutine
		}
		reload(subtree);
	}

	public List<TreePath> getItemsWithTheSpecifiedNameInTree( final String name ) {
		return getItemsWithTheSpecifiedNameInSpecifiedSubtree((MutableTreeItem)getRoot(), name);
	}

	public List<TreePath> getItemsWithTheSpecifiedNameInSpecifiedSubtree( MutableTreeItem subtree, String name ) {
		ArrayList<TreePath> paths = new ArrayList<TreePath>();
		for (final Enumeration<?> e = subtree.children(); e.hasMoreElements();) {
			MutableTreeItem node = (MutableTreeItem)e.nextElement();
			if (node.isLeaf()) {
				String operatorName = node.getTreeItemName();
				if (operatorName.contains(name)) {
					paths.add(new TreePath(node.getPath()));
				}
			} else {
				paths.addAll(getItemsWithTheSpecifiedNameInSpecifiedSubtree(node, name));
			}
		}
		return paths;
	}

	public TreeItem getItemEqualsTo( String name ) {
		for (MutableTreeItem subtree : getSubTrees()) {
			List<TreePath> paths = getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, name);
			for (TreePath treePath : paths) {
				MutableTreeItem leaf = (MutableTreeItem)treePath.getLastPathComponent();
				TreeItem item = leaf.getTreeItemObject();
				// Ensure that the operator BEGINS with the pattern.
				if (item.getName().equals(name)) {
					return leaf.getTreeItemObject();
				}
			}
		}
		return null;
	}

	/**
	 * The list of all operators that can be added to the graph.
	 */
	private MutableTreeItem[] getSubTrees() {
		return new MutableTreeItem[]{ getCommonOperatorTree(),
				getUserOperatorTree(),
				getSubroutineTree()};
	}

	public List<String> getItemsWithNameStartingWith( String pattern ) {
		List<String> result = new ArrayList<String>();
		for (MutableTreeItem subtree : getSubTrees()) {
			for (TreePath path : getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, pattern)) {
				MutableTreeItem leaf = (MutableTreeItem)path.getLastPathComponent();
				TreeItem item = leaf.getTreeItemObject();
				// Ensure that the operator BEGINS with the pattern.
				if (item.getName().startsWith(pattern)) {
					result.add(item.getName());
				}
			}
		}
		return result;
	}

	///// 
	///// I N P U T   /   O U T P U T
	/////

	protected List<TreeItem> loadItemsFromXMLTree( List<XMLElement> xmlTree ) {
		List<TreeItem> treeItems = new ArrayList<TreeItem>();
		if (xmlTree.isEmpty()) {
			return treeItems;
		}
		// Discard tag <ARIANE>.
		List<XMLElement> xmlContents = xmlTree.get(0).getChildren();
		if (xmlContents == null) {
			return treeItems;
		}
		for (XMLElement element : xmlContents) {
			if ("OperatorLibrary".equalsIgnoreCase(element.getName())
					|| "Library".equalsIgnoreCase(element.getName())
					|| "Bibliotheque".equalsIgnoreCase(element.getName())) {
				for (XMLElement item : element.getChildren()) {
					String tag = item.getName();
					if ("DocPath".equalsIgnoreCase(tag)) {
						Ariane.setLibraryDocumentationPath(item.getValue());
					} else if ("CommandStatus".equalsIgnoreCase(tag)) {
						UserPreferences.setCommandStatus(item.getValue());
					} else if ("CommandVisu".equalsIgnoreCase(tag)) {
						UserPreferences.setCommandVisu(item.getValue());
					} else if ("Version".equalsIgnoreCase(tag)) {
						// kept for sake of compatibility
						// Gets the version number of library.
						((MutableTreeItem)getRoot()).setUserObject(item.getValue());
						reload();
						if (item.getValue().startsWith("Pandore")) {
							UserPreferences.setCommandStatus("pstatus");
							UserPreferences.setCommandVisu("pvisu");
						}
					} else if ("Index".equalsIgnoreCase(tag)) {
						// kept for sake of compatibility
						_pathToHTMLLibraryDocumentationFile = item.getValue();
					} else if ("Operator".equalsIgnoreCase(tag) || "Operateur".equalsIgnoreCase(tag)) {
						treeItems.add(new OperatorTreeItem(item.getChildren()));
					}
				}
				return treeItems;
			} else if ("SubroutineLibrary".equalsIgnoreCase(element.getName())) {
				for (XMLElement item : element.getChildren()) {
					if ("Subroutine".equalsIgnoreCase(item.getName())) {
						treeItems.add(new SubroutineTreeItem(item.getChildren()));
					}
				}
				return treeItems;
			}
		}
		return treeItems;
	}
}