/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.tree;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.List;

import fr.greyc.pantheon.ariane.model.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.item.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.model.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;

/**
 * A <code>TreeItem</code> is an item that can be inserted in
 * the tree of the explorer.
 */
public final class OperatorTreeItem extends TreeItem {
	/** The magic number for transferable item. */
	public static final DataFlavor[] FLAVORS = { new DataFlavor("application/x-java-serialized-item-tree", "TreeItem Flavor") };

	public OperatorTreeItem( Item typeItem, String name, String docURL, int inputDataSlots, int inputParameterSlots, int outputDataSlots, String signature ) {
		super(typeItem, name, docURL, inputDataSlots, inputParameterSlots, outputDataSlots, signature);
	}

	public OperatorTreeItem( List<XMLElement> xmlTree ) {
		super(xmlTree);
	}

	@Override
	public DataFlavor[] getTransferDataFlavors( ) {
		return FLAVORS.clone();
	}

	@Override
	public Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (isDataFlavorSupported(flavor)) {
			return this;
		} else {
			throw new UnsupportedFlavorException(flavor);
		}
	}

	@Override
	public boolean isDataFlavorSupported( DataFlavor flavor ) {
		return flavor.equals(FLAVORS[0]);
	}

	@Override
	public GraphItem createGraphItem( Graph graph ) {
		try {
			switch (getType()) {
			case OPERATOR: return new OperatorItem(this, graph);
			case LOADER: return new LoaderItem(this, graph);
			case SAVER: return new SaverItem(this, graph);
			case VIEWER: return new ViewerItem(this, graph);
			case FILENAME: return new FileNameItem(this, graph);
			case CONSTANT: return new ConstantItem(this, graph);
			case MACRO_OPERATOR: return new MacroOperatorItem(this, graph);
			case EVAL: return new EvalItem(this, graph);
			case FORMAT: return new FormatItem(this, graph);
			case IF_CONTROL: return new IfControlStructureItem(this, graph);
			case SELECT_CONTROL: return new SelectControlStructureItem(this, graph);
			case FOR_CONTROL: return new ForControlStructureItem(this, graph);
			case WHILE_CONTROL: return new WhileControlStructureItem(this, graph);
			case MERGE_CONTROL: return new MergeControlStructureItem(this, graph);
			case NOTE: return new NoteItem(this, graph);
			case SYSTEM: return new SystemItem(this, graph);
			case SCRIPT: return new ScriptItem(this, graph);
			default: return null;
			}
		} catch (InterruptedException e) {
			return null;
		}
	}

	///// 
	///// I N P U T  /  O U T P U T
	/////

	@Override
	public XMLElement convertToXMLTree( ) throws IOException {
		if (getType() == Item.OPERATOR) {
			XMLElement element = new XMLElement("Operator");
			element.addChildren(new XMLElement("Category", getCategory()));
			element.addChildren(new XMLElement("Name", getName()));
			element.addChildren(new XMLElement("Documentation", getDocumentationURL()));
			element.addChildren(new XMLElement("Binary", getBinaryPath()));
			element.addChildren(new XMLElement("Parameters", Integer.toString(getNumberOfParameterSlots())));
			element.addChildren(new XMLElement("Signature", getSignature()));
			element.addChildren(new XMLElement("Inputs", Integer.toString(getNumberOfInputDataSlots())));
			element.addChildren(new XMLElement("Outputs", Integer.toString(getNumberOfOutputDataSlots())));
			return element;
		} else {
			return null;
		}
	}

	@Override
	protected void loadItemFromXMLTree( List<XMLElement> xmlelements ) {
		setType(Item.OPERATOR);

		for (XMLElement item : xmlelements ) {
			if ("Category".equalsIgnoreCase(item.getName()) || "Categorie".equalsIgnoreCase(item.getName())) {
				setCategory(item.getValue());
			} else if ("Name".equalsIgnoreCase(item.getName()) || "Nom".equalsIgnoreCase(item.getName())) {
				setName(item.getValue());
			} else if ("Documentation".equalsIgnoreCase(item.getName())) {
				setDocumentationURL(item.getValue());
			} else if ("Binary".equalsIgnoreCase(item.getName()) || "Binaire".equalsIgnoreCase(item.getName())) {
				setBinaryPath(item.getValue());
			} else if ("Parameters".equalsIgnoreCase(item.getName()) || "Parametre".equalsIgnoreCase(item.getName())) {
				setNumberOfParameterSlots(Integer.parseInt(item.getValue()));
			} else if ("Signature".equalsIgnoreCase(item.getName())) {
				setSignature(item.getValue());
			} else if ("Inputs".equalsIgnoreCase(item.getName()) || "Entree".equalsIgnoreCase(item.getName())) {
				setNumberOfInputDataSlots(Integer.parseInt(item.getValue()));
			} else if ("Outputs".equalsIgnoreCase(item.getName()) || "Sortie".equalsIgnoreCase(item.getName())) {
				setNumberOfOutputDataSlots(Integer.parseInt(item.getValue()));
			}
		}
	}
}