/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.xml;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

/**
 * An <code>XMLElement</code> defines a pair tag-value
 * that can be read or write in an XML file.
 */
public final class XMLElement implements Transferable, java.io.Serializable {
	private static final long serialVersionUID = 42L; 

	private final String _tag;

	private String _value = "";

	private List<XMLElement> _children;

	// Transferable
	public static final DataFlavor FLAVOR = new DataFlavor(fr.greyc.pantheon.ariane.model.xml.XMLElement.class, "Ariane graph");

	private static final DataFlavor[] FLAVORS = { FLAVOR, };

	/**
	 * Creates a new XMLElement with the specified tag name.
	 * 
	 * @param tag the tag name.
	 */
	public XMLElement( String tag ) {
		_tag = tag;
	}

	/**
	 * Creates a new XMLElement with the specified value.
	 * 
	 * @param tag the tag name.
	 * @param value the value.
	 */
	public XMLElement( final String tag, final String value ) {
		_tag = tag;
		setValue(value);
	}

	/**
	 * @return the name of the tag.
	 */
	public String getName( ) {
		return _tag;
	}

	/**
	 * Sets the new value to the tag.
	 * 
	 * @param value the value.
	 */
	public void setValue( String value ) {
		if (value != null) {
			_value = value;
		}
	}

	public String getValue( ) {
		return _value;
	}

	/**
	 * @return the list of all children.
	 */
	public List<XMLElement> getChildren( ) {
		return _children;
	}

	/**
	 * Adds a subtree as the child of the 
	 * specified XML element.
	 * 
	 * @param element the XML element.
	 */
	public void addChildren( XMLElement element ) {
		if (element != null) {
			if (_children == null) {
				_children = new LinkedList<XMLElement>();
			}
			_children.add(element);
		}
	}

	/**
	 * Adds a subtree as the child of the specified XML tree.
	 * 
	 * @param xmlTree the subtree to be added.
	 */
	public void addChildren( List<XMLElement> xmlTree ) {
		if (xmlTree == null) {
			return;
		}
		if (_children == null) {
			_children = xmlTree;
		} else { // Catenate 2 linked-lists.
			while (!xmlTree.isEmpty()) {
				_children.add(xmlTree.remove(0));
			}
		}
	}

	@Override
	public DataFlavor[] getTransferDataFlavors() {
		return FLAVORS.clone();
	}

	@Override
	public boolean isDataFlavorSupported( DataFlavor flavor ) {
		return flavor.equals(FLAVORS[0]);
	}

	@Override
	public Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (flavor.equals(FLAVORS[0])) {
			return this;
		}
		throw new UnsupportedFlavorException(flavor);
	}
}