/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.model.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import fr.greyc.pantheon.ariane.Ariane;

/**
 * <i>XMLFileInputStream</i> is a parser of XML tags.
 * It reads and build an XML tree of elements (tag, value).
 */
public final class XMLFileInputStream extends DefaultHandler {
	/** The input stream where to read the tags. */
	private FileInputStream _ifs;

	/** The last read value in the input stream. */
	private final StringBuffer _readValue = new StringBuffer(255);

	/** The list of all read elements. */
	private List<XMLElement> _finalElements;

	/** The current read element. */
	private Stack<XMLElement> _currentElements;

	/**
	 * Creates and opens an input stream from the given filename.
	 * 
	 * @param filename the input filename.
	 * @throws IOException any IO exceptions (in the current language).
	 */
	public XMLFileInputStream( final String filename ) throws IOException {
		try {
			_ifs = new FileInputStream(filename);
		} catch (IOException e) {
			throw new IOException(Ariane.getString("XMLFileInputStream.error1") + filename, e);
		}
	}

	/**
	 * Creates and opens an input stream from the given <code>File</code>.
	 * 
	 * @param filename the input File.
	 * @throws IOException any IO exceptions (in the current language).
	 */
	public XMLFileInputStream( final File filename ) throws IOException {
		try {
			_ifs = new FileInputStream(filename);
		} catch (IOException e) {
			throw new IOException(Ariane.getString("XMLFileInputStream.error1") + filename, e);
		}
	}

	/**
	 * Parses the input stream and builds the XML tree.
	 * 
	 * @throws Exception bad XML input file.
	 */
	public void parse( ) throws SAXException, ParserConfigurationException, IOException {
		InputSource inputStream = new InputSource(_ifs);

		SAXParserFactory spf = SAXParserFactory.newInstance();
		// Set namespaceAware to true to get a default parser (SAX2, JAXP 1.0=false!)
		spf.setNamespaceAware(true);
		SAXParser saxParser = spf.newSAXParser();
		XMLReader xmlReader = saxParser.getXMLReader();
		
		xmlReader.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
		xmlReader.setFeature("http://xml.org/sax/features/validation", false);
		
		xmlReader.setContentHandler(this);
		xmlReader.setErrorHandler(this);
		// Do parsing
		xmlReader.parse(inputStream);
	}

	/**
	 * Receives notification of the beginning of the document.
	 * 
	 * @throws SAXException any SAX exception.
	 */
	@Override
	public void startDocument( ) throws SAXException {
		_currentElements = new Stack<XMLElement>();
		_finalElements = new LinkedList<XMLElement>();
	}

	/**
	 * Receives notification of the end of the document.
	 * 
	 * @throws SAXException	any SAX exception.
	 */
	@Override
	public void endDocument( ) throws SAXException {
		_currentElements = null;
	}

	/**
	 * Receives notification of the start of an element.
	 * 
	 * @param uri the name space URI mapped to the prefix.
	 * @param tag the local name (without prefix), or the empty string.
	 * @param qName the qualified name (with prefix), or the empty string.
	 * @param atts the specified or defaulted attributes.
	 * @throws SAXException any SAX exception.
	 */
	@Override
	public void startElement( String uri, String tag, String qName, Attributes atts ) throws SAXException {
		XMLElement element = new XMLElement(tag);

		if (_currentElements.empty()) { // It is a sub-item.
			_finalElements.add(element);
		} else {
			XMLElement element1 = _currentElements.peek();
			element1.addChildren(element);
		}
		_currentElements.push(element);
		_readValue.setLength(0);
	}

	/**
	 * Receives notification of the end of an element.
	 * 
	 * @param uri the name space URI mapped to the prefix.
	 * @param tag the local name (without prefix), or the empty string.
	 * @param qName the qualified name (with prefix), or the empty string.
	 * @throws SAXException any SAX exception.
	 */
	@Override
	public void endElement( String uri, String tag, String qName ) throws SAXException {
		XMLElement element = _currentElements.pop();
		element.setValue(_readValue.toString());
	}

	/**
	 * Receives notification of character data inside an element.
	 * @throws SAXException Any SAX exception.
	 */
	@Override
	public void characters( char[] characters, int start, int length ) throws SAXException {
		_readValue.append(characters , start , length);
	}

	/**
	 * Returns the reading result as an XML tree.
	 * 
	 * @return the XML tree.
	 */
	public List<XMLElement> getElements( ) {
		return _finalElements;
	}

	/**
	 * Closes the current input stream.
	 * 
	 * @throws IOException any IO exception.
	 */
	public void close( ) throws IOException {
		_ifs.close();
	}
}