/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.presenter;

import java.awt.Point;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JMenuBar;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.generator.CodeGenerator;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileInputStream;
import fr.greyc.pantheon.ariane.view.LibraryPathChooser;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.ariane.view.UserPreferencesChooser;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.explorer.ItemTreeView;
import fr.greyc.pantheon.ariane.view.utils.ScriptNameChooser;

public final class Presenter { // TODO
	private static int _currentReadPasteSession;
	
	private Workspace _workspace;
	
	private ItemTreeModel _itemTreeModel;

	public void buildWorkspace( ItemTreeModel itemTreeModel ) {
		_itemTreeModel = itemTreeModel;
		_workspace = new Workspace(this, new ItemTreeView(this, itemTreeModel));
	}

	///// 
	///// W O R K S P A C E
	///// 

	public void generatesScriptCodeFromGraph( Graph graph ) {
		CodeGenerator generator = selectCodeGenerator(graph);
		if (generator != null) {
			generator.generate(this, graph.getListOfGraphItems());
		}
	}

	private static CodeGenerator selectCodeGenerator( Graph graph ) {
		String[] chosenFile = ScriptNameChooser.chooseFilename(graph.getName(), CodeGenerator.AVAILABLE_LANGUAGES);
		if (chosenFile.length > 0) {
			return CodeGenerator.createCodeGenerator(new File(chosenFile[0]), chosenFile[1]);
		} else {
			return null;
		}
	}

	private void changeContent( String pathName ) {
		if (!UserPreferences.isOperatorLibraryUsed()) {
			return;
		}
		String libraryFilename = pathName + File.separator + "lib" + File.separator + Ariane.LIBRARY_CONFIG_FILENAME;
		try {
			List<XMLElement> configHandler = Ariane.readXMLFile(new File(libraryFilename));
			ItemTreeModel operatorTree = _itemTreeModel;
			operatorTree.setCommonOperators(configHandler);
			Ariane.setLibraryPath(pathName);
		} catch (IOException e) {
			Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
		}
	}

	public void chooseOperatorLibrary( ) {
		LibraryPathChooser library = new LibraryPathChooser();
		library.setVisible(true);
		String result = library.getOperatorLibraryPath();
		if (result != null) {
			changeContent(result);
			if (UserPreferences.saveInUserPreferenceFile()) {
				_workspace.repaintAllGraphs();
			}
		}
	}

	public void openPreferencesChooser( ) {
		UserPreferencesChooser preferenceChooser = new UserPreferencesChooser();
		preferenceChooser.setVisible(true);
		if (preferenceChooser.isValidated()) {
			_workspace.repaintAllGraphs();
			UserPreferences.saveInUserPreferenceFile();
		}
	}

	public void restoreSplitPanesLocation() {
		_workspace.restoreSplitPanesLocation();
	}

	public void display(List<String> graphFilenames ) {
		List<Graph> graphs = new ArrayList<>();
		for (String graphFilename : graphFilenames) {
			graphs.add(loadGraph(new File(graphFilename)));
		}
		_workspace.display(graphs);
	}

	public static int getCurrentReadOrPasteSessionId() {
		return _currentReadPasteSession;
	}

	public JMenuBar buildMenuBar() {
		return _workspace.buildMenuBar();
	}

	public void initContent() {
		_workspace.initContent();
	}

	public Workspace getWorkspace() {
		return _workspace;
	}

	///// 
	///// G R A P H
	///// 

	public void openFile( String filename ) {
		if (!_workspace.isGraphAlreadyLoaded(filename)) {
			Graph graph = loadGraph(new File(filename));
			if (graph != null) {
				_workspace.addGraph(graph);
			}
		}
	}

	public void createItemsFromClipboard( Graph graph, Transferable contents ) {
		try {
			_currentReadPasteSession++;
			XMLElement result = (XMLElement)contents.getTransferData(XMLElement.FLAVOR);
			graph.pasteSelectedItems(result);
		} catch (UnsupportedFlavorException e) {
		} catch (IOException e) {
		}
	}

	public Graph loadGraph( File file ) {
		_currentReadPasteSession++;
		try {
			XMLFileInputStream convert = new XMLFileInputStream(file);
			convert.parse();
			convert.close();
			Graph graph = new RootGraph(file);
			graph.createFromXMLTree(convert.getElements());
			addRecentProjects(file.getAbsolutePath());
			return graph;
		} catch (IOException | SAXException | ParserConfigurationException ex) {
			printErrorMessage(Ariane.getString("Workspace.loading.error1") + file.getName() + " " + ex.getLocalizedMessage());
//			ex.printStackTrace();
			return null;
		}
	}

	public void addNewItemInGraph( TreeItem treeItem, Point coordinate ) throws InstantiationException {
		_workspace.addNewItem(treeItem, coordinate);
	}

	///// 
	///// C O N S O L E
	///// 

	public void printOutputMessage( String message ) {
		_workspace.printOutputMessage(message);
	}

	public void printSystemMessage( String mesg ) {
		_workspace.printSystemMessage(mesg);
	}

	public void printWarning( String mesg ) {
		_workspace.printWarning(mesg);
	}

	public void printErrorMessage( String mesg ) {
		_workspace.printErrorMessage(mesg);
	}

	public void printErrorMessage( Ariane.ErrorCode errorCode, String item, int itemId, String mesg ) {
		_workspace.printErrorMessage(errorCode, item, itemId, mesg);
	}

	public void displayDocumentationPage( String title, String page, String index ) {
		_workspace.displayHTMLManualPage(title, page, index);
	}

	public void copy() {
		_workspace.copy();
	}

	public void cut() {
		_workspace.cut();
	}

	public void paste() {
		_workspace.paste();
	}

	public ItemTreeModel getTree() {
		return _itemTreeModel;
	}

	public void displayPropertyPanel( PropertyPanel propertyPanel ) {
		_workspace.displayPropertyPanel(propertyPanel);
	}

	public void addGraph( SubGraph subGraph ) {
		_workspace.addGraph(subGraph);
	}

	public void renameGraph( Graph graph, String name ) {
		_workspace.renameGraph(graph, name);
	}

	public void setExecution( boolean isRunning ) {
		_workspace.setExecution(isRunning);
	}

	public void update() {
		_workspace.update();
	}

	public void updateRecentFilesListMenu() {
		_workspace.updateRecentlyClosedWorkspacesListMenu();
	}

	public void removeGraph( Graph graph ) {
		_workspace.removeGraph(graph);
	}

	public void addRecentProjects( String absolutePath ) {
		_workspace.addRecentlyClosedWorkspace(absolutePath);
	}

	public boolean isPastable() {
		return _workspace.isPastable();
	}

	public void deleteSelectedItems() {
		_workspace.deleteSelectedItems();
	}

	public void setModifiedSaveB( boolean value ) {
		_workspace.setModifiedSaveB(value);
	}

	public void changeLanguage() {
		_workspace.changeLanguage();
	}

	public int getSplitH1() {
		return _workspace.getSplitH1();
	}

	public int getSplitH2() {
		return _workspace.getSplitH2();
	}

	public int getSplitV1() {
		return _workspace.getSplitV1();
	}

	public int getSplitV2() {
		return _workspace.getSplitV2();
	}

	public void displayLibraryDocumentation() {
		displayDocumentationPage("Index", _itemTreeModel.getPathToLibraryDocumentationFile(), null);
	}
}