/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view;

import fr.greyc.pantheon.ariane.Ariane;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.WindowConstants;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 * A <code>HTMLDocumentViewers</code> displays documentation
 * as HTML pages.
 */
public final class HTMLDocumentViewer extends JFrame implements HyperlinkListener {
	private static final String ERROR_PAGE = "item.html";
	private final JEditorPane _webEditorPane;

	public HTMLDocumentViewer( String title, URI page ) {
		_webEditorPane = new JEditorPane();
		_webEditorPane.addHyperlinkListener(this);
		getContentPane().add(new JScrollPane(_webEditorPane));
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		setBounds(200, 25, 800, 600);
		setPage(title, page);
		setLocationRelativeTo(Ariane.getFrame());
		setVisible(true);
	}

	/**
	 * Finds the absolute address of the given document either in the
	 * operators library documentation or in the Ariane manual pages.
	 * 
	 * @param document The name or path of an HTML page.
	 * @param fragment the index in the page as "#name".
	 * @return the address or null.
	 */
	public static URI getURI( String document, String fragment ) {
		if (document.isEmpty()) {
			try {
				return Ariane.getDocumentationURI(ERROR_PAGE).toURI();
			} catch (URISyntaxException e ) {
				return null;
			}
		}
		
		try {
			return directURI(document, fragment);
		} catch (IOException | URISyntaxException | IllegalArgumentException e) { }

		try {
			return relativeURIToLibrary(document, fragment);
		} catch (IOException | URISyntaxException e) { }

		try {
			return deprecatedURI1(document, fragment); // Kept for the sake of compatibility
		} catch (IOException | URISyntaxException e) {
		}

		try {
			return deprecatedURI2(document, fragment);  // Kept for the sake of compatibility
		} catch (URISyntaxException e) {
		}
		return null;
	}

	private static URI deprecatedURI2( String document, String fragment )
			throws URISyntaxException {
		URI page;
		// Ariane manual pages
		URL address = Ariane.getDocumentationURI(document);
		// The last chance: the operator page.
		if (address == null) {
			return Ariane.getDocumentationURI(ERROR_PAGE).toURI();
		} else {
			page = Ariane.getDocumentationURI(document).toURI();
			return new URI(page.toString() + ((fragment == null) ? "" : fragment));
		}
	}

	private static URI deprecatedURI1( String document, String fragment )
			throws IOException, URISyntaxException, MalformedURLException {
		URI page;
		File file = new File(Ariane.getLibraryPath() + File.separator + document);
		// Check whether the file exists.
		String address = file.getCanonicalPath();
		page = new URI("file", address, fragment);
		InputStream stream = page.toURL().openStream(); // try to open it.
		stream.close();
		return page;
	}

	private static URI relativeURIToLibrary( String document, String fragment )
			throws IOException, URISyntaxException, MalformedURLException {
		URI page;
		File file = new File(Ariane.getLibraryPath() + Ariane.getLibraryDocumentationPath() + File.separator + document);
		String address = file.getCanonicalPath();
		// Library documentation?
		page = new URI("file", address, fragment);
		// Check whether the file exists.
		InputStream stream = page.toURL().openStream(); // try to open it.
		stream.close();
		return page;
	}

	private static URI directURI( String document, String fragment )
			throws URISyntaxException, IOException, MalformedURLException {
		URI page;
		if (fragment == null ) {
			page = new URI(document);
		} else {
			page = new URI(document + fragment);
		}
		// Check whether the file exists.
		InputStream stream = page.toURL().openStream(); // try to open it.
		stream.close();
		return page;
	}

	private void setPage( String title, URI page ) {
		setTitle(title);
		if (page == null) {
			displayErrorPage(title);
		} else {
			try {
				_webEditorPane.setPage(page.toURL());
				_webEditorPane.setEditable(false);
			} catch (IOException e) {
				displayErrorPage(title);
			}
		}
	}

	@Override
	public void hyperlinkUpdate( HyperlinkEvent event ) {
		HyperlinkEvent.EventType eventtype = event.getEventType();
		if (eventtype == HyperlinkEvent.EventType.ACTIVATED) {
			URL url = event.getURL();
			try {
				_webEditorPane.setPage(url);
			} catch (IOException e) {
				displayErrorPage(url.toString());
			}
		}
	}

	private void displayErrorPage( String title ) {
		_webEditorPane.setContentType("text/html");
		_webEditorPane.setText("<html><body>" + Ariane.getString("HTMLViewer.unknownpage") + title + "</body></html>");
	}
}
