/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.EmptyBorder;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputParameterSlot;

/**
 * A <code>PropertyViewer</code> provides a dialog box to display
 * graph item's properties.
 */
public final class PropertyPanel extends JPanel {
	private JLabel _resultLabel;
	private final List<JButton> _inputButtons = new ArrayList<JButton>();
	private final List<JButton> _outputButtons = new ArrayList<JButton>();

	private static final int DEFAULT_ROW_HEIGHT = 25;

	private ResourceBundle _currentLanguage;

	private final List<PropertyCouple> _properties = new ArrayList<PropertyCouple>();
	private final List<PropertyCouple> _inputsoutputs = new ArrayList<PropertyCouple>();

	public PropertyPanel() {
		setOpaque(false);
	}

	public boolean isCorrectlyLocalized() {
		return _currentLanguage == Ariane.getLanguage();
	}

	public void update( GraphItem item ) {
		boolean updated = false;
		if (_resultLabel != null) {
			updated = true;
			if (item.getExecutionStatus() == GraphItem.Status.EXECUTED) {
				_resultLabel.setText(item.getOutputResultSlot().getValue());
			} else {
				_resultLabel.setText(Ariane.getString("PropertyViewer.result.unbound"));
			}
		}
		// Input buttons
		if (!_inputButtons.isEmpty()) {
			for (int i = 0; i < item.getInputDataSlots().size(); i++) {
				updated = true;
				if (item.getExecutionStatus() == GraphItem.Status.EXECUTED
						&& item.getInputDataSlots().get(i).isValidateValue()) {
					_inputButtons.get(i).setEnabled(true);
				} else {
					_inputButtons.get(i).setEnabled(false);
				}
			}
		}
		// Output buttons
		if (!_outputButtons.isEmpty()) {
			for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
				updated = true;
				if (item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getOutputDataSlots().get(i).isValidate()) {
					_outputButtons.get(i).setEnabled(true);
				} else {
					_outputButtons.get(i).setEnabled(false);
				}
			}
		}
		if (updated) {
			revalidate();
			repaint();
		}
	}
	
	public PropertyPanel( final ResourceBundle language, final GraphItem item ) {
		_currentLanguage = language;
		setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.getString("PropertyViewer.title") + " (" + item.getId() + ")"));
		setLayout(new BorderLayout());
		setBackground(getBackground().darker());

		buildPropertyList(item);

		JPanel contents = initPanelContents(item);
	
		JScrollPane scrollPane = new JScrollPane(contents, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(null);
		add(scrollPane, BorderLayout.CENTER);
	}

	private JPanel initPanelContents( final GraphItem item ) {
		JPanel contents = new JPanel();
		contents.setLayout(new BoxLayout(contents, BoxLayout.Y_AXIS));
		contents.setBackground(getBackground().darker());

		Dimension dimension = new Dimension(10000, DEFAULT_ROW_HEIGHT);
		for (PropertyCouple property : _properties) {
			JPanel panel = new JPanel(new GridLayout(1, 2));
			JLabel title = new JLabel(property._title);
			title.setOpaque(true);
			title.setBorder(new EmptyBorder(0, 2, 0, 0));
			title.setFont(UserPreferences.getDefaultFont());
			panel.add(title);

			JComponent component = property._value;
			adjustComponentSize(component);
			panel.add(property._value);
			
			panel.setMaximumSize(dimension);
			contents.add(panel);
		}

		for (PropertyCouple property : _inputsoutputs) {
			JPanel panel = new JPanel(new BorderLayout());
			JLabel title = new JLabel(property._title);
			title.setOpaque(true);
			title.setBorder(new EmptyBorder(0, 2, 0, 0));
			title.setFont(UserPreferences.getDefaultFont());
			panel.add(title, BorderLayout.NORTH);

			JComponent component = property._value;
			adjustComponentSize(component);
			panel.add(property._value, BorderLayout.CENTER);
			
			panel.setMaximumSize(dimension);
			contents.add(panel);
		}

		JPanel buttonPanel = new JPanel(new GridLayout(1, 2));
		JButton buttonOK = new JButton(Ariane.getString("Graph.Menu.execute"));
		buttonOK.setFont(UserPreferences.getDefaultFont());
		buttonOK.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				item.getGraph().startExecution();
			}
		});
		buttonPanel.add(buttonOK);

		JButton buttonReset = new JButton(Ariane.getString("Graph.Menu.reset"));
		buttonReset.setFont(UserPreferences.getDefaultFont());
		buttonReset.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				item.resetAll();
			}
		});
		dimension = buttonPanel.getPreferredSize();
		dimension.width = 10000;
		buttonPanel.setMaximumSize(dimension);
		buttonPanel.add(buttonReset);
		contents.add(buttonPanel);
		contents.add(Box.createVerticalGlue());
		return contents;
	}

	private void adjustComponentSize( final JComponent component ) {
		int height = component.getPreferredSize().height;
		Dimension dimension = new Dimension(0, height);
		component.setPreferredSize(dimension);
	}

	private void buildPropertyList( final GraphItem item ) {
		createNameElement(item);
		createCategoryElement(item);
		createBinaryElement(item);
		createManualElement(item);
		createFormulaElement(item);
		createSpecificElement(item);
		createParametersElement(item);
		createInputElement(item);
		createOutputElement(item);
		createResultElement(item);
	}

	private void createResultElement( final GraphItem item ) {
		if (item.getOutputResultSlots().size() > 0) {
			if (item.getExecutionStatus() == GraphItem.Status.EXECUTED) {
				_resultLabel = new JLabel(item.getOutputResultSlot().getValue());
			} else {
				_resultLabel = new JLabel(Ariane.getString("PropertyViewer.result.unbound"));
			}
			_properties.add(new PropertyCouple(Ariane.getString("PropertyViewer.result"), _resultLabel));
		}
	}

	private void createOutputElement( final GraphItem item ) {
		for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
			JButton button = (JButton)item.getOutputDataSlots().get(i).getView().getValueEditor();
			if (item.getExecutionStatus() == GraphItem.Status.EXECUTED && item.getOutputDataSlots().get(i).isValidate()) {
				button.setEnabled(true);
			} else {
				button.setEnabled(false);
			}
			_outputButtons.add(button);
			_inputsoutputs.add(new PropertyCouple(Ariane.getString("PropertyViewer.output")
					+ " #"
					+ i
					+ " (ctrl-"
					+ i + ")", button));
		}
	}

	private void createInputElement( final GraphItem item ) {
		for (int i = 0; i < item.getInputDataSlots().size(); i++) {
			JButton button = (JButton)item.getInputDataSlots().get(i).getView().getValueEditor();
			if (item.getExecutionStatus() == GraphItem.Status.EXECUTED
					&& item.getInputDataSlots().get(i).isValidateValue()) {
				button.setEnabled(true);
			} else {
				button.setEnabled(false);
			}
			_inputButtons.add(button);
			_inputsoutputs.add(new PropertyCouple(Ariane.getString("PropertyViewer.input") + " #" + i, button));
		}
	}

	private void createParametersElement( final GraphItem item ) {
		for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
			InputParameterSlot isp = item.getInputParameterSlots().get(i);
			JComponent component = isp.getView().getValueEditor();
			_properties.add(new PropertyCouple(item.getInputParameterSlots().get(i).getName(), component));
		}
	}

	private void createSpecificElement( final GraphItem item ) {
		List<PropertyCouple> additionalProperties = item.getView().getSpecificProperties();
		if (additionalProperties != null) {
			_properties.addAll(additionalProperties);
		}
	}

	private void createFormulaElement( final GraphItem item ) {
		if (item.getFormula() != null && item.isDisplayableFormula()) {
			JComponent component = item.getView().getFormulaEditor();
			_properties.add(new PropertyCouple(item.getFormulaLabel(), component));
		}
	}

	private void createManualElement( final GraphItem item ) {
		JButton jbuttonman;
		jbuttonman = new JButton(item.getName());
		jbuttonman.setFont(UserPreferences.getDefaultFont());
		jbuttonman.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				item.getView().displayDocumentationManual();
			}
		});
		_properties.add(new PropertyCouple(Ariane.getString("PropertyViewer.documentation"), jbuttonman));
	}

	private void createBinaryElement( final GraphItem item ) {
		JComponent editor = item.getView().getBinaryEditor();
		if (editor != null) {
			_properties.add(new PropertyCouple(Ariane.getString("PropertyViewer.binary"), editor));
		}
	}

	private void createCategoryElement( final GraphItem item ) {
		if (item.getCategory() != null) {
			JTextField component = new JTextField(item.getCategory());
			component.setEnabled(false);
			_properties.add(new PropertyCouple(Ariane.getString("PropertyViewer.category"), component));
		}
	}

	private void createNameElement( final GraphItem item ) {
		_properties.add(new PropertyCouple(Ariane.getString("PropertyViewer.name"), item.getView().getNameEditor()));
	}

	public static final class PropertyCouple {
		private final String _title;
		private final JComponent _value;

		public PropertyCouple( String title, JComponent value ) {
			_title = title;
			value.setFont(UserPreferences.getDefaultFont());
			_value = value;
		}
	}
}