/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.item.DataFile;
import fr.greyc.pantheon.ariane.model.tree.MutableTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.Graph;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.presenter.RootGraph;
import fr.greyc.pantheon.ariane.view.explorer.Explorer;
import fr.greyc.pantheon.ariane.view.explorer.GraphEditor;
import fr.greyc.pantheon.ariane.view.explorer.ItemTreeView;
import fr.greyc.pantheon.ariane.view.explorer.OperatorTreeEditor;
import fr.greyc.pantheon.ariane.view.explorer.SubroutineTreeEditor;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.ariane.view.utils.FadingEffects;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.KeyStroke;
import javax.swing.UIManager;

/**
 * The <i>Workspace</i> is the main window of the software.
 * It holds an editor of graphs, an explorer and a console.
 */
public final class Workspace extends JPanel implements ClipboardOwner { // TODO
	private static final int MAXIMUM_RECENT_FILES = 10;

	private final Presenter _presenter;
	private final Explorer _itemExplorer;
	private final GraphEditor _graphEditor;
	private final Console _console;
	private final JPanel _itemPropertyPanel;

	private static List<String> _recentlyClosedWorkspaceFilenames = new ArrayList<String>();

	private Graph _currentGraph;
	
	private final Toolbar _toolBar;
	private final MenuBar _menuBar;

	private final Map<String, Action> _actions = new HashMap<String, Action>();

	/** The vertical split pane ratio (splitPaneHV2 + splitPaneV2). */
	private JSplitPane _splitPaneV1;
	/** The vertical split pane ratio (splipaneH  property). */
	private JSplitPane _splitPaneV2;
	/** The current horizontal split pane ratio (explorer + terminal). */
	private JSplitPane _splitPaneH1;
	/** The current horizontal split pane ratio (graph + console). */
	private JSplitPane _splitPaneH2;

	public Workspace( Presenter presenter, ItemTreeView operatorsTree ) {
		_presenter = presenter;
		buildActions();
		_itemPropertyPanel = new JPanel(new BorderLayout());
		_itemPropertyPanel.setMinimumSize(new Dimension(70, 60));
		_console = new Console();
		_graphEditor = new GraphEditor(this);
		_itemExplorer = new Explorer(_presenter, operatorsTree);
		_toolBar = new Toolbar(_actions);
		_menuBar = new MenuBar(_actions, _itemExplorer.getSubMenu(), _recentlyClosedWorkspaceFilenames, presenter);
	}

	public void initContent( ) {
		// Organization of the workspace in three panes.
		// H: (editor + console)
		// V: Explorer V.
		JPanel contentPane = new JPanel();

		// Terminal + Explorer
		_splitPaneH1 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _itemExplorer, new Terminal(_presenter));
		_splitPaneH1.setResizeWeight(1.0);
		_splitPaneH1.setDividerSize(5);

		// Graph + console
		_splitPaneH2 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _graphEditor, new JScrollPane(_console));
		_splitPaneH2.setResizeWeight(1.0);
		_splitPaneH2.setDividerSize(5);

		// V2: tree | (editor + property)
		_splitPaneV2 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH2, _itemPropertyPanel);
		_splitPaneV2.setDividerSize(5);
		_splitPaneV2.setResizeWeight(1.0);

		// V1: tree | ((terminal explorer) + (editor+console+property)
		_splitPaneV1 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH1, _splitPaneV2);
		_splitPaneV1.setDividerSize(5);

		contentPane.setLayout(new BorderLayout());

		contentPane.add(_toolBar, BorderLayout.NORTH);
		contentPane.add(_splitPaneV1, BorderLayout.CENTER);

		setLayout(new BorderLayout());
		add(contentPane, BorderLayout.CENTER);

		changeLanguage();
		setCurrentGraph(null);

		InputMap inputMap = getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		ActionMap actionMap = getActionMap();
		displayShortcutForOutputDisplay(inputMap, actionMap, 0);
		displayShortcutForOutputDisplay(inputMap, actionMap, 1);
		displayShortcutForOutputDisplay(inputMap, actionMap, 2);
		displayShortcutForOutputDisplay(inputMap, actionMap, 3);
		displayShortcutForOutputDisplay(inputMap, actionMap, 4);
		displayShortcutForOutputDisplay(inputMap, actionMap, 5);
	}

	private void displayShortcutForOutputDisplay( InputMap inputMap, ActionMap actionMap, final int outputIndex ) {
		inputMap.put(KeyStroke.getKeyStroke(new Character((char)('0' + outputIndex)), 0), "Display" + outputIndex);
		actionMap.put("Display" + outputIndex, new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				if (_currentGraph != null) {
					_currentGraph.displayOutputs(outputIndex);
				}
			}
		});
	}

	private int getSplitPaneRatio( JSplitPane splitPaneH1, int splitPaneH1Ratio ) {
		if (Ariane.getFrame().getExtendedState() ==  Frame.MAXIMIZED_BOTH) {
			if (splitPaneH1.getLastDividerLocation() < 0) {
				return splitPaneH1Ratio;
			}
			return splitPaneH1.getLastDividerLocation();
		}
		return splitPaneH1.getDividerLocation();
	}

	public int getSplitH1( ) {
		return getSplitPaneRatio(_splitPaneH1, UserPreferences.getSplitPaneH1Ratio());
	}

	public int getSplitH2( ) {
		return getSplitPaneRatio(_splitPaneH2, UserPreferences.getSplitPaneH2Ratio());
	}

	public int getSplitV1( ) {
		return getSplitPaneRatio(_splitPaneV1, UserPreferences.getSplitPaneV1Ratio());
	}

	public int getSplitV2( ) {
		return getSplitPaneRatio(_splitPaneV2, UserPreferences.getSplitPaneV2Ratio());
	}

	public void restoreSplitPanesLocation() {
		_splitPaneH1.setDividerLocation(UserPreferences.getSplitPaneH1Ratio());
		_splitPaneH2.setDividerLocation(UserPreferences.getSplitPaneH2Ratio());
		_splitPaneV1.setDividerLocation(UserPreferences.getSplitPaneV1Ratio());
		_splitPaneV2.setDividerLocation(UserPreferences.getSplitPaneV2Ratio());
	}

	public static void setRecentProjectFilenames( List<String> filenames ) {
		_recentlyClosedWorkspaceFilenames = filenames;
	}

	public static List<String> getRecentProjectFilenames( ) {
		return _recentlyClosedWorkspaceFilenames;
	}

	public void addRecentlyClosedWorkspace( String filename ) {
		if (_recentlyClosedWorkspaceFilenames.contains(filename)) {
			_recentlyClosedWorkspaceFilenames.remove(filename);
			_recentlyClosedWorkspaceFilenames.add(0, filename);
		} else {
			if (_recentlyClosedWorkspaceFilenames.size() > MAXIMUM_RECENT_FILES) {
				_recentlyClosedWorkspaceFilenames.remove(MAXIMUM_RECENT_FILES);
			}
			_recentlyClosedWorkspaceFilenames.add(0, filename);
		}
	}

	public void updateRecentlyClosedWorkspacesListMenu() {
		_menuBar.updateRecentlyClosedWorkspacesListMenu(_recentlyClosedWorkspaceFilenames, _presenter);
	}

	/////
	///// M E N U  &  T O O L B A R
	/////

	private void buildFileActions() {
		_actions.put("new", new AbstractAction(null, Ariane.getIcon("new.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				createNewGraph();
			}
		});
		_actions.put("open", new AbstractAction(null, Ariane.getIcon("open.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				openNewGraph();
			}
		});
		_actions.put("save", new AbstractAction(null, Ariane.getIcon("save.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				saveFileAction();
			}
		});
		_actions.put("saveAs", new AbstractAction(null, Ariane.getIcon("saveas.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				saveAsFileAction();
			}
		});
		_actions.put("print", new AbstractAction(null, Ariane.getIcon("print.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				print();
			}
		});
		_actions.put("importSubroutine", new AbstractAction() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				importSubroutinesAction();
			}
		});
		_actions.put("exportSubroutine", new AbstractAction() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				exportSubroutinesAction();
			}
		});
		_actions.put("close", new AbstractAction() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				close();
			}
		});
		_actions.put("closeAll", new AbstractAction() {
			@Override
			public void actionPerformed( ActionEvent event ) {
				closeAll();
			}
		});
		_actions.put("quit", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				quitAction();
			}
		});
	}

	private void buildEditActions() {
		_actions.put("undo", new AbstractAction(null, Ariane.getIcon("undo.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				undo();
			}
		});
		_actions.put("redo", new AbstractAction(null, Ariane.getIcon("redo.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				redo();
			}
		});
		_actions.put("copy", new AbstractAction(null, Ariane.getIcon("copy.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				copy();
			}
		});
		_actions.put("cut", new AbstractAction(null, Ariane.getIcon("cut.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				cut();
			}
		});
		_actions.put("paste", new AbstractAction(null, Ariane.getIcon("paste.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				_presenter.paste();
			}
		});
		_actions.put("erase", new AbstractAction(null, Ariane.getIcon("delete.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) { 
				deleteSelectedItems();
			}
		});
		_actions.put("select", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				selectAll();
			}
		});
	}

	private void buildSearchActions() {
		_actions.put("conf", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				_presenter.chooseOperatorLibrary();
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
		_actions.put("preferences", new AbstractAction(null, Ariane.getIcon("preferences.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				_presenter.openPreferencesChooser();
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
		_actions.put("operatorTree", new AbstractAction(null, Ariane.getIcon("operatortree.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				new OperatorTreeEditor(_presenter.getTree().getUserOperatorTree()).setVisible(true);
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
		_actions.put("subroutineTree", new AbstractAction(null, Ariane.getIcon("subroutinetree.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				new SubroutineTreeEditor(_presenter.getTree().getSubroutineTree()).setVisible(true);
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
	}

	private void buildGraphActions() {
		_actions.put("stop", new AbstractAction(null, Ariane.getIcon("stop.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				stopExecution();
			}
		});
		_actions.get("stop").setEnabled(false);
		_actions.put("exec", new AbstractAction(null, Ariane.getIcon("execute.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				execute();
			}
		});
		_actions.get("exec").setEnabled(false);
		_actions.put("reset", new AbstractAction(null, null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				reset();
			}
		});
		_actions.put("generate", new AbstractAction(null, Ariane.getIcon("generation.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				_presenter.generatesScriptCodeFromGraph(_currentGraph.getRoot());
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
	}

	private void buildHelpActions() {
		_actions.put("about", new AbstractAction(null) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				FadingEffects.fadeIn(Ariane.getFrame());
				new AboutBox(Ariane.getFrame()).setVisible(true);
				FadingEffects.fadeOut(Ariane.getFrame());
			}
		});
		_actions.put("help", new AbstractAction(null, Ariane.getIcon("help.png")) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				displayHTMLManualPage("Ariane", "manual.html", null);
			}
		});
	}

	private void buildActions() {
		buildFileActions();
		buildEditActions();
		buildSearchActions();
		buildGraphActions();
		buildHelpActions();
	}

	public JMenuBar buildMenuBar( ) {
		return _menuBar;
	}

	private void setInternationalText( ) {
		// Change all text of the JColorChooser.
		UIManager.put("ColorChooser.cancelText", Ariane.getString("ColorChooser.cancelText"));
		UIManager.put("ColorChooser.okText", Ariane.getString("ColorChooser.okText"));
		UIManager.put("ColorChooser.resetText", Ariane.getString("ColorChooser.resetText"));
		UIManager.put("ColorChooser.previewText", Ariane.getString("ColorChooser.previewText"));
		UIManager.put("ColorChooser.sampleText", Ariane.getString("ColorChooser.sampleText"));
		UIManager.put("ColorChooser.swatchesNameText", Ariane.getString("ColorChooser.swatchesNameText"));
		UIManager.put("ColorChooser.swatchesRecentText", Ariane.getString("ColorChooser.swatchesRecentText"));
		UIManager.put("ColorChooser.rgbNameText", Ariane.getString("ColorChooser.rgbNameText"));
		UIManager.put("ColorChooser.rgbRedText", Ariane.getString("ColorChooser.rgbRedText"));
		UIManager.put("ColorChooser.rgbGreenText", Ariane.getString("ColorChooser.rgbGreenText"));
		UIManager.put("ColorChooser.rgbBlueText", Ariane.getString("ColorChooser.rgbBlueText"));
		UIManager.put("ColorChooser.hsbNameText", Ariane.getString("ColorChooser.hsbNameText"));
		UIManager.put("ColorChooser.hsbRedText", Ariane.getString("ColorChooser.hsbRedText"));
		UIManager.put("ColorChooser.hsbGreenText", Ariane.getString("ColorChooser.hsbGreenText"));
		UIManager.put("ColorChooser.hsbBlueText", Ariane.getString("ColorChooser.hsbBlueText"));
		UIManager.put("ColorChooser.hsbHueText", Ariane.getString("ColorChooser.hsbHueText"));
		UIManager.put("ColorChooser.hsbSaturationText", Ariane.getString("ColorChooser.hsbSaturationText"));
		UIManager.put("ColorChooser.hsbBrightnessText", Ariane.getString("ColorChooser.hsbBrightnessText"));

		// Change all text of the JFileChooser.
		UIManager.put("FileChooser.fileNameLabelText", Ariane.getString("FileChooser.fileNameLabelText"));
		UIManager.put("FileChooser.homeFolderToolTipText", Ariane.getString("FileChooser.homeFolderToolTipText"));
		UIManager.put("FileChooser.newFolderToolTipText", Ariane.getString("FileChooser.newFolderToolTipText"));
		UIManager.put("FileChooser.listViewButtonToolTipText", Ariane.getString("FileChooser.listViewButtonToolTipText"));
		UIManager.put("FileChooser.detailsViewButtonToolTipText", Ariane.getString("FileChooser.detailsViewButtonToolTipText"));
		UIManager.put("FileChooser.saveButtonText" , Ariane.getString("FileChooser.saveButtonText"));
		UIManager.put("FileChooser.openButtonText", Ariane.getString("FileChooser.openButtonText"));
		UIManager.put("FileChooser.cancelButtonText", Ariane.getString("FileChooser.cancelButtonText"));
		UIManager.put("FileChooser.updateButtonText", Ariane.getString("FileChooser.updateButtonText"));
		UIManager.put("FileChoosre.saveButtonToolTipText", Ariane.getString("FileChooser.saveButtonToolTipText"));
		UIManager.put("FileChooser.openButtonToolTipText", Ariane.getString("FileChooser.openButtonToolTipText"));
		UIManager.put("FileChooser.cancelButtonToolTipText", Ariane.getString("FileChooser.cancelButtonToolTipText"));
		UIManager.put("FileChooser.updateButtonToolTipText", Ariane.getString("FileChooser.updateButtonToolTipText"));
		UIManager.put("FileChooser.lookInLabelText", Ariane.getString("FileChooser.lookInLabelText"));
		UIManager.put("FileChooser.saveInLabelText", Ariane.getString("FileChooser.saveInLabelText"));
		UIManager.put("FileChooser.filesOfTypeLabelText", Ariane.getString("FileChooser.filesOfTypeLabelText"));
		UIManager.put("FileChooser.upFolderToolTipText", Ariane.getString("FileChooser.upFolderToolTipText"));
		UIManager.put("FileChooser.acceptAllFileFilterText", Ariane.getString("FileChooser.acceptAllFileFilterText"));

		UIManager.put("OptionPane.cancelButtonText", Ariane.getString("OptionPane.cancelButtonText"));
		UIManager.put("OptionPane.noButtonText", Ariane.getString("OptionPane.noButtonText"));
		UIManager.put("OptionPane.okButtonText", Ariane.getString("OptionPane.okButtonText"));
		UIManager.put("OptionPane.yesButtonText", Ariane.getString("OptionPane.yesButtontext"));
		_toolBar.setInternationalText();
		_menuBar.setInternationalText();
	}

	public void changeLanguage( ) {
		setInternationalText();
		_itemExplorer.setInternationalText();
	}

	public boolean isGraphAlreadyLoaded(String filename) {
		return _graphEditor.exists(filename);
	}

	public void display( List<Graph> graphs ) {
		for (Graph graph : graphs) {
			if (graph != null) {
				addGraph(graph);
			}
		}
	}

	public void displayPropertyPanel( PropertyPanel propertyPanel ) {
		_itemPropertyPanel.removeAll();
		if (propertyPanel != null) {
			_itemPropertyPanel.add(propertyPanel, BorderLayout.CENTER);
		}
		_itemPropertyPanel.revalidate();
		_itemPropertyPanel.repaint();
	}

	///// 
	///// C O N S O L E
	///// 

	public void printOutputMessage( String mesg ) {
		_console.displayOutputMessage(mesg);
	}

	public void printSystemMessage( String mesg ) {
		_console.displaySystemMessage(mesg);
	}

	public void printWarning( String mesg ) {
		_console.displayErrorMessage(Ariane.getString("warning") + mesg);
	}

	public void printErrorMessage( String mesg ) {
		_console.displayErrorMessage(Ariane.getString("error") + mesg);
	}

	public void printErrorMessage( Ariane.ErrorCode errorCode, String item, int itemId, String mesg ) {
		_console.displayErrorMessage(item + " (" + itemId + "): " + Ariane.getString("error") + "#" + errorCode.ordinal() + ": " + mesg);
	}

	///// 
	/////  A C T I O N S
	///// 

	/**
	 * Displays the specified HTML 'page' as a manual page
	 * for the word 'title'. Pages are stored either
	 * <ol>
	 * <li>in the operators library package.
	 * <li>in the Ariane jar file.
	 * </ol>
	 * 
	 * @param title the title of the manual.
	 * @param page the URL address of the manual content.
	 * @param index index in the page as "#name".
	 */
	public void displayHTMLManualPage( String title, String page, String index ) {
		URI uri = HTMLDocumentViewer.getURI(page, index);
		if (uri == null) {
			return;
		}
		String command = UserPreferences.getWebBrowserPath();
		if (command != null && command.length() > 0) {
			try {
				Runtime.getRuntime().exec(command + " " + uri);
				return;
			} catch (IOException e1) {
				printErrorMessage(Ariane.getString("Workspace.browser.error") + e1.getMessage());
			}
		}
		new HTMLDocumentViewer(title, uri);
	}

	///// 
	/////  H I S T O R Y 
	/////

	private void undo( ) {
		if (_currentGraph != null) {
			_currentGraph.undo();
		}
	}

	private void redo( ) {
		if (_currentGraph != null) {
			_currentGraph.redo();
		}
	}

	public void update( ) {
		_actions.get("save").setEnabled(_currentGraph.isModified());
		_actions.get("undo").setEnabled(_currentGraph.canUndo());
		_actions.get("redo").setEnabled(_currentGraph.canRedo());
	}

	public void setModifiedSaveB( boolean status ) {
		_actions.get("save").setEnabled(status);
	}

	public void setCurrentGraph( Graph graph ) {
		_currentGraph = graph;
		if (_currentGraph == null) {
			_actions.get("close").setEnabled(false);
			_actions.get("closeAll").setEnabled(false);
			_actions.get("save").setEnabled(false);
			_actions.get("undo").setEnabled(false);
			_actions.get("redo").setEnabled(false);
			_actions.get("saveAs").setEnabled(false);
			_actions.get("print").setEnabled(false);
			_actions.get("copy").setEnabled(false);
			_actions.get("cut").setEnabled(false);
			_actions.get("paste").setEnabled(false);
			_actions.get("erase").setEnabled(false);
			_actions.get("generate").setEnabled(false);
			_actions.get("select").setEnabled(false);
		} else {
			_actions.get("close").setEnabled(true);
			_actions.get("closeAll").setEnabled(true);
			_actions.get("saveAs").setEnabled(true);
			if (!_actions.get("stop").isEnabled()) {
				_actions.get("exec").setEnabled(true);
			}
			_actions.get("print").setEnabled(true);
			_actions.get("copy").setEnabled(true);
			_actions.get("cut").setEnabled(true);
			_actions.get("paste").setEnabled(true);
			_actions.get("erase").setEnabled(true);
			_actions.get("generate").setEnabled(true);
			_actions.get("select").setEnabled(true);
			update();
		}
	}

	private void selectAll( ) {
		if (_currentGraph != null) {
			_currentGraph.selectAll();
		}
	}

	///// 
	///// C O P Y  /  P A S T E
	///// 

	@Override
	public void lostOwnership( Clipboard clipboard, Transferable contents ) { /* empty */ }

	/**
	 * Copies the selected items in the clipboard. Each item
	 * are copied in the clipboard.
	 * <b>Warning</b>: connections between copied item
	 * are kept but connection between other items are deleted.
	 */
	public void copy( ) {
		if (_currentGraph != null) {
			Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
			clipboard.setContents(_currentGraph.copySelectedItems(), this);
		}
	}

	/**
	 * Copies the selected items in the clipboard.
	 * <b>Warning</b>: connections between copied item
	 * are kept but connection between other items are deleted.
	 */
	public void cut( ) {
		copy();
		deleteSelectedItems();
	}

	public void paste( ) {
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		Transferable contents = clipboard.getContents(null);
		if (_currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR)) {
			_presenter.createItemsFromClipboard(_currentGraph, contents);
		}
	}

	public boolean isPastable( ) {
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		Transferable contents = clipboard.getContents(null);
		return _currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR);
	}

	public void deleteSelectedItems( ) {
		if (_currentGraph != null) {
			_currentGraph.deleteSelectedItems();
		}
	}

	public void addNewItem( TreeItem treeItem, Point coordinates ) throws InstantiationException {
		if (_currentGraph == null) {
			createNewGraph();
		}
		_currentGraph.addItem(treeItem, coordinates);
	}

	///// 
	/////  P R I NT I N G
	/////

	private void print( ) {
		if (_currentGraph != null) {
			// Unselect all items.
			_currentGraph.unSelectAll();

			// Set up defaults preferences.
			PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
			aset.add(new JobName("Ariane", UserPreferences.getCurrentLocale()));
			aset.add(OrientationRequested.LANDSCAPE);
			aset.add(MediaSizeName.ISO_A4);

			// Print it
			PrinterJob printJob = PrinterJob.getPrinterJob();
			printJob.setPrintable(_currentGraph.getView());
			if (printJob.printDialog(aset)) {
				try {
					printJob.print(aset);
				} catch (PrinterException e) {
					printErrorMessage(Ariane.getString("Workspace.printing.error1") + e);
				}
			}
		}
	}

	public void repaintAllGraphs( ) {
		_graphEditor.repaintAllGraphs();
	}

	public void addGraph( Graph graph ) {
		_graphEditor.add(graph);
	}

	public void renameGraph( Graph graph, String name ) {
		_graphEditor.setTitle(graph, name);
	}

	public void removeGraph( Graph graph ) {
		_graphEditor.remove(graph);
	}

	public GraphView createNewGraph( ) {
		Graph graph = new RootGraph(Ariane.getString("defaultfilename"));
		addGraph(graph);
		return graph.getView();
	}

	private void close( ) {
		if (_currentGraph != null) {
			_currentGraph.close();
			updateRecentlyClosedWorkspacesListMenu();
		}
	}

	private void closeAll( ) {
		_graphEditor.closeAllGraphs();
	}

	private void quitAction( ) {
		if (isQuitConfirmed()) {
			System.exit(0);
		}
	}

	private void execute( ) {
		if (_currentGraph != null) {
			_currentGraph.startExecution();
		}
	}

	private void reset( ) {
		if (_currentGraph != null) {
			_currentGraph.resetSelectedItems();
		}
	}

	private void stopExecution( ) {
		if (_currentGraph != null && Graph.isExecutionRunning()) {
			_currentGraph.stopExecution();
			printSystemMessage(Ariane.getString("Workspace.execution.stop"));
		}
	}

	public void setExecution( boolean isRunning ) {
		if (isRunning) {
			_console.clear();
			_actions.get("exec").setEnabled(false);
			_actions.get("stop").setEnabled(true);
		} else {
			_actions.get("exec").setEnabled(true);
			_actions.get("stop").setEnabled(false);
		}
	}

	protected boolean isQuitConfirmed( ) {
		boolean quit = _graphEditor.closeAllGraphs();
		if (quit) {
			UserPreferences.deleteTemporaryFiles();
			UserPreferences.saveInUserPreferenceFile();
		}
		return quit;
	}

	private void openNewGraph( ) {
		File currentFileName = new File(UserPreferences.getCurrentDirectory());

		JFileChooser fileChooser = new JFileChooser(currentFileName);
		fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleOpenGraph"));
		fileChooser.setMultiSelectionEnabled(true);
		fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
			@Override
			public boolean accept( File file ) {
				String filename = file.getName();
				return filename.endsWith(Ariane.DEFAULT_ARIANE_FILE_EXT) || file.isDirectory();
			}
			@Override
			public String getDescription() {
				return "Ariane (*" + Ariane.DEFAULT_ARIANE_FILE_EXT + ")";
			}
		});

		FadingEffects.fadeIn(Ariane.getFrame());
		if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			for (File file : fileChooser.getSelectedFiles()) {
				_presenter.openFile(file.getAbsolutePath());
			}
			UserPreferences.setCurrentDirectory(fileChooser.getSelectedFiles()[0].getParent());
		}
		FadingEffects.fadeOut(Ariane.getFrame());
	}

	private void saveFileAction( ) {
		if (_currentGraph != null) {
			_currentGraph.getRoot().save();
		}
	}

	private void saveAsFileAction( ) {
		if (_currentGraph != null) {
			_currentGraph.getRoot().saveAs();
		}
	}

	private void importSubroutinesAction() {
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleOpenFile"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileChooser.setSelectedFile(new File(Ariane.getLibraryPath()));
		int returnVal = fileChooser.showOpenDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile();
			try {
				MutableTreeItem tree = _presenter.getTree().addInSubroutines(Ariane.readXMLFile(file));
				SubroutineTreeEditor.save(tree);
				printSystemMessage(Ariane.getString("Workspace.addsubroutine.message"));
			} catch (IOException e) {
				printErrorMessage(Ariane.getString("Workspace.addsubroutine.error") + " " + file.getName() + ". " + e.getLocalizedMessage());
			}
		}
	}

	private void exportSubroutinesAction() {
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setDialogTitle(Ariane.getString("FileChooser.titleSaveFile"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		fileChooser.setSelectedFile(new File(Ariane.getLibraryPath()));
		if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile();
			if (file.exists() && JOptionPane.showConfirmDialog(this,
					Ariane.getString("Workspace.dialog.message.part1")
					+ file.getName()
					+ Ariane.getString("Workspace.dialog.message.part2"),
					Ariane.getString("Workspace.dialog.title"),
					JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
				return;
			}
			try {
				DataFile.copyFile(Ariane.getSubroutineFilename(), file.getAbsolutePath());
				printSystemMessage(file.getName()
						+ " " + Ariane.getString("Workspace.saving.success"));
			} catch (IOException e ) {
				printErrorMessage(Ariane.getString("PreferenceChooser.error.saved")
						+ " " + Ariane.getPreferencesFilename() + " " + e.getLocalizedMessage());
			}
		}
	}
}
