/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.explorer;

import java.awt.Cursor;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.ItemTreeModel;
import fr.greyc.pantheon.ariane.model.tree.MutableTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.view.WidgetLocalizable;

public final class ItemTreeView extends JPanel implements DragSourceListener, DragGestureListener, WidgetLocalizable {
	private static final Cursor DRAG_AND_DROP_CURSOR;

	private final ItemTreeModel _itemTreeModel;

	private final DragSource _dragSource;

	private final JTree _jtree;

	static {
		Point hotSpot = new Point(15, 15);
		DRAG_AND_DROP_CURSOR = Toolkit.getDefaultToolkit().createCustomCursor(Ariane.getIcon("itemcursor.png").getImage(), hotSpot, "itemCursor");
	}

	public ItemTreeView( final Presenter presenter, ItemTreeModel itemTree ) {
		_itemTreeModel = itemTree;
		_itemTreeModel.setInternationalText();
		
		_jtree = new JTree(itemTree);
		_jtree.setFont(UserPreferences.getDefaultFont());
		_jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
		_jtree.putClientProperty("JTree.lineStyle", "Angled");
		final DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
		renderer.setLeafIcon(Ariane.getIcon("item.png"));
		renderer.setOpenIcon(null);
		renderer.setClosedIcon(null);
		_jtree.setCellRenderer(renderer);
		_jtree.addMouseListener(new MouseAdapter() {
			// Case of double-click -> display manual.
			@Override
			public void mousePressed( MouseEvent event ) {
				if (event.getClickCount() > 1) {
					TreePath selectedPath = _jtree.getPathForLocation(event.getX(), event.getY());
					TreeItem selectedTreeItem = getTreeItemAtTheSpecifiedPath(selectedPath);
					if (selectedTreeItem != null && selectedTreeItem.getDocumentationURL() != null) {
						presenter.displayDocumentationPage(Ariane.getString("HTMLViewer.manual") + " " + selectedTreeItem, selectedTreeItem.getDocumentationURL(), null);
					}
				}
			}
		});
		// Discard key CTRL-A (used by the Graph).
		_jtree.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_MASK), "none");

		setLayout(new GridLayout());
		setBackground(_jtree.getBackground());
		add(_jtree);

		_dragSource = DragSource.getDefaultDragSource();
		_dragSource.createDefaultDragGestureRecognizer(_jtree, DnDConstants.ACTION_MOVE, this);
	}

	@Override
	public void setInternationalText() {
		_itemTreeModel.setInternationalText();
	}

	///// 
	///// D R A G   &   D R O P
	/////

	@Override
	public void dragDropEnd( DragSourceDropEvent event ) {
		// Do nothing
	}

	@Override
	public void dragEnter( DragSourceDragEvent event ) {
		// Do nothing
	}

	@Override
	public void dragExit( DragSourceEvent event ) {
		// Do nothing
	}

	@Override
	public void dragOver( DragSourceDragEvent event ) {
		// Do nothing
	}

	@Override
	public void dropActionChanged( DragSourceDragEvent event ) {
		// Do nothing
	}

	@Override
	public void dragGestureRecognized( DragGestureEvent event ) {
		TreePath selectedPathInTheTree = _jtree.getSelectionPath();
		TreeItem selectedTreeItem = getTreeItemAtTheSpecifiedPath(selectedPathInTheTree);
		if (selectedTreeItem != null) {
			_dragSource.startDrag(event, DRAG_AND_DROP_CURSOR, selectedTreeItem, this);
		}
	}

	private TreeItem getTreeItemAtTheSpecifiedPath( TreePath selectedPathInTheTree ) {
		if (selectedPathInTheTree != null) {
			MutableTreeItem selectedNode = (MutableTreeItem)selectedPathInTheTree.getLastPathComponent();
			if (selectedNode.isLeaf() && selectedNode.getUserObject() instanceof TreeItem) {
				return (TreeItem)selectedNode.getUserObject();
			}
		}
		return null;
	}

	protected boolean highlightItemsThatContainTheSpecifiedPattern( final String pattern ) {
		List<TreePath> paths = _itemTreeModel.getItemsWithTheSpecifiedNameInTree(pattern);
		if (!paths.isEmpty()) {
			TreePath[] pathsTab = paths.toArray(new TreePath[paths.size()]);
			_jtree.setSelectionPaths(pathsTab);
			return true;
		}
		return false;
	}
}
