/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.FocusTraversalPolicy;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Traversal policy of the components in a graph.
 */
public final class GraphFocusTraversalPolicy extends FocusTraversalPolicy {
	private final Comparator<? super Component> _comparator;
	private Container _cachedRoot;
	private List<Component> _cachedCycle;

	public GraphFocusTraversalPolicy() {
		// the order is based on coordinates.
		_comparator = new Comparator<Component>() {
			@Override
			public int compare( Component component1, Component component2 ) {
				if (component1.getBounds().x < component2.getBounds().x) {
					return -1;
				} else if (component1.getBounds().x > component2.getBounds().x) {
					return 1;
				} else if (component1.getBounds().y < component2.getBounds().y) {
					return -1;
				}
				return 1;
			}
		};
	}

	private void enumerateAndSortCycle( Container focusCycleRoot, List<Component> cycle ) {
		if (!focusCycleRoot.isShowing() || !focusCycleRoot.isVisible() || !focusCycleRoot.isDisplayable()) {
			return;
		}
		Component[] components = focusCycleRoot.getComponents();
		for (int i = 0; i < components.length; i++) {
			if (components[i].isFocusable()) {
				cycle.add(components[i]);
			}
		}
		// Sort the Components in the cycle
		Collections.sort(cycle, _comparator);
	}

	@Override
	public Component getComponentAfter( Container aContainer, Component aComponent ) {
		List<Component> cycle = new ArrayList<Component>();
		enumerateAndSortCycle(aContainer, cycle);

		try {
			int index = Collections.binarySearch(cycle, aComponent, _comparator);

			if (index < 0) {
				int newIndex = cycle.indexOf(aComponent);
				if (newIndex >= 0) {
					index = newIndex;
				} else {
					index = -index - 2;
				}
			}

			while (true) {
				index++;

				if (index >= cycle.size()) {
					if (aContainer.isFocusCycleRoot()) {
						_cachedRoot = aContainer;
						_cachedCycle = cycle;

						Component retval = getFirstComponent(aContainer);

						_cachedRoot = null;
						_cachedCycle = null;
						return retval;
					} else {
						return null;
					}
				} else {
					return cycle.get(index);
				}
			}
		} catch (ClassCastException e) {
			return getFirstComponent(aContainer);
		}
	}

	@Override
	public Component getComponentBefore( Container aContainer, Component aComponent ) {
		List<Component> cycle = new ArrayList<Component>();
		enumerateAndSortCycle(aContainer, cycle);

		try {
			int index = Collections.binarySearch(cycle, aComponent, _comparator);

			if (index < 0) {
				index = -index - 2;
			} else {
				index--;
			}

			if (index < 0) {
				_cachedRoot = aContainer;
				_cachedCycle = cycle;

				Component retval = getLastComponent(aContainer);

				_cachedRoot = null;
				_cachedCycle = null;

				return retval;
			} else {
				return  cycle.get(index);
			}
		} catch (ClassCastException e) {
			return getLastComponent(aContainer);
		}
	}

	@Override
	public Component getFirstComponent( Container aContainer ) {
		List<Component> cycle;

		if (aContainer.equals(_cachedRoot)) {
			cycle = _cachedCycle;
		} else {
			cycle = new ArrayList<Component>();
			enumerateAndSortCycle(aContainer, cycle);
		}

		int size = cycle.size();
		if (size == 0) {
			return null;
		}
		return cycle.get(0);
	}

	@Override
	public Component getLastComponent( Container aContainer ) {
		List<Component> cycle;

		if (aContainer.equals(_cachedRoot)) {
			cycle = _cachedCycle;
		} else {
			cycle = new ArrayList<Component>();
			enumerateAndSortCycle(aContainer, cycle);
		}

		int size = cycle.size();
		if (size == 0) {
			return null;
		}
		return cycle.get(cycle.size() - 1);
	}

	@Override
	public Component getDefaultComponent( Container aContainer ) {
		return getFirstComponent(aContainer);
	}
}