/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.awt.Rectangle;
import java.util.Hashtable;
import java.util.Map;

/**
 * Defines a class that knows how to lay out a graph.
 * A component is added with a location constraints.
 */
public final class XYLayout implements LayoutManager2 {
	private final Map<Component, XYConstraint> _componentConstraints = new Hashtable<Component, XYConstraint>();
	private static final int ICON_WIDTH = 12;
	private static final int ICON_HEIGHT = 12;
	private static final int OFFSET = 60;

	public enum Orientation {
		NORTHWEST,
		NORTHEAST,
		SOUTHWEST,
		SOUTHEAST,
		EASTSOUTH,
		WESTNORTH
	}

	@Override
	public void addLayoutComponent( String name, Component component ) {
		// Do nothing
	}

	@Override
	public void addLayoutComponent( Component component, Object constraints ) {
		Object innerConstraints;
		if (constraints == null) {
			innerConstraints = new XYConstraint();
		} else {
			innerConstraints = constraints;
		}
		synchronized (component.getTreeLock()) {
			if (innerConstraints instanceof XYConstraint) {
				XYConstraint constraint = (XYConstraint)innerConstraints;
				_componentConstraints.put(component, (XYConstraint)innerConstraints);
				if (constraint.getRank() < 0) {
					component.setBounds(constraint.getLocation().x, constraint.getLocation().y, 0, 0);
				}
			}
		}
	}

	@Override
	public void removeLayoutComponent( Component component ) {
		synchronized (component.getTreeLock()) {
			_componentConstraints.remove(component);
		}
	}

	@Override
	public Dimension preferredLayoutSize( Container container ) {
		return minimumLayoutSize(container);
	}

	@Override
	public Dimension minimumLayoutSize( Container container ) {
		synchronized (container.getTreeLock()) {
			Dimension dimension = new Dimension(0, 0);
			final Dimension containerDim = container.getSize();
			for (Component component : container.getComponents()) {
				if (component.isVisible() && _componentConstraints.get(component).getRank() < 0) {
					Rectangle rectangle = getComponentBounds(component, containerDim);
					dimension.width  = Math.max(dimension.width , rectangle.x + rectangle.width + 50);
					dimension.height = Math.max(dimension.height, rectangle.y + rectangle.height + 50);
				}
			}
			return dimension;
		}
	}

	@Override
	public Dimension maximumLayoutSize( Container container ) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	@Override
	public void layoutContainer( Container container ) {
		synchronized (container.getTreeLock()) {
			final Dimension containerDim = container.getSize();
			for (Component component : container.getComponents()) {
				if (component.isVisible()) {
					component.setBounds(getComponentBounds(component, containerDim));
				}
			}
		}
	}

	@Override
	public float getLayoutAlignmentX( Container container ) {
		return 0.0f; // Prefer centered alignment.
	}

	@Override
	public float getLayoutAlignmentY( Container container) {
		return 0.0f; // Prefer centered alignment.
	}

	@Override
	public void invalidateLayout( Container container ) {
		// Do nothing
	}

	private Rectangle getComponentBounds( Component component, Dimension size ) {
		XYConstraint constraints = _componentConstraints.get(component);

		if (isConstrainedByOrientation(constraints)) {
			final int rank = constraints.getRank() + 1;
			switch (constraints.getOrientation()) {
			case WESTNORTH:
				return new Rectangle(0, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			case NORTHWEST:
				return new Rectangle(OFFSET + rank * 20, 0, ICON_WIDTH, ICON_HEIGHT);
			case SOUTHEAST:
				return new Rectangle(200 + OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
			case SOUTHWEST:
				return new Rectangle(OFFSET + rank * 20, size.height - ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT);
			case NORTHEAST:
				return new Rectangle(size.width - ICON_WIDTH, OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			case EASTSOUTH:
				return new Rectangle(size.width - ICON_WIDTH, 200 + OFFSET + rank * 20, ICON_WIDTH, ICON_HEIGHT);
			default:
				return null;
			}
		} else { // Constraints by coordinates.
			Rectangle rectangle = component.getBounds();
			Dimension dimension = component.getPreferredSize();
			rectangle.width = dimension.width;
			rectangle.height = dimension.height;
			return rectangle;
		}
	}

	private boolean isConstrainedByOrientation( XYConstraint constraints ) {
		return constraints.getRank() != -1;
	}
}