/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.graph.item;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.StringTokenizer;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.UserPreferences;
import fr.greyc.pantheon.ariane.presenter.command.ChangeTextEdit;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.ariane.view.graph.item.etc.NoteColor;

public final class NoteItemView extends GraphItemView {
	private static final transient PropertyPanel BLANK_PROPERTY_PANEL = new PropertyPanel();

	private JTextArea _noteText;

	@Override
	public void draw( ) {
		removeAll();
		setBackground(getBackgroundColor());
		setToolTipText(toString());

		GridBagConstraints constraint = new GridBagConstraints();
		constraint.insets = new Insets(5, 2, 2, 2);
		constraint.gridheight = 1;
		constraint.gridwidth = 1;
		constraint.gridy++; constraint.gridx = 0;
		constraint.anchor = GridBagConstraints.CENTER;
		if (_noteText == null) {
			_noteText = new JTextArea(getFormula(), 1, 2);
			_noteText.setFont(UserPreferences.getDefaultFont());
			_noteText.setBackground(Color.WHITE);
			_noteText.setForeground(Color.BLACK);
			_noteText.setTabSize(2);
		}
		// Mouse events.
		_noteText.addMouseListener(new MouseAdapter( ) {
			@Override
			public void mousePressed( MouseEvent event ) {
				if (getModel().isSelected()) {
					_noteText.getCaret().setVisible(true);
				} else {
					onMouseReleased(event);
				}
			}
		});

		_noteText.addFocusListener(this);
		_noteText.setEditable(true);
		_noteText.getCaret().setVisible(false);

		// The lights ramp that shows the current state.
		constraint.gridwidth = GridBagConstraints.REMAINDER;
		constraint.gridy++; constraint.gridx = 0;
		add(_noteText, constraint);
	}

	@Override
	public void printComponent( Graphics graphicContext ) {
		super.printComponent(graphicContext);
		FontMetrics fontMetric;
		fontMetric = graphicContext.getFontMetrics();
		int row = fontMetric.getHeight();
		super.centerText(getName(), graphicContext, row, false);
		row += fontMetric.getHeight() + 4;
		graphicContext.setFont(UserPreferences.getDefaultFont());
		if (getFormula() != null) {
			StringTokenizer stringTokenizer = new StringTokenizer(getFormula(), "\n");
			while (stringTokenizer.hasMoreTokens()) {
				graphicContext.drawString(stringTokenizer.nextToken(), 5, row);
				row += fontMetric.getHeight();
			}
		}
	}

	@Override
	protected void onMouseDoubleClicked( MouseEvent event ) {
		displayDocumentationManual();
	}

	@Override
	protected void focusGainedHook( ) {
		_noteText.moveCaretPosition(_noteText.getText().length());
	}

	@Override
	public void focusLost( FocusEvent event ) {
		if (event.getSource() == _noteText) {
			String newFormula = _noteText.getText();
			if (!newFormula.equals(getFormula()) && isShowing()) {
				getGraph().validateOneAction(new ChangeTextEdit(getModel(), _noteText, newFormula, getFormula()));
				setFormula(newFormula);
			}
		}
		_noteText.getCaret().setVisible(false);
		_noteText.setCaretPosition(0);
	}

	@Override
	protected PropertyPanel createPropertyPanel() {
		return BLANK_PROPERTY_PANEL;
	}

	///// 
	///// C O L O R
	/////

	@Override
	public Color getBackgroundColor( ) {
		return NoteColor.getColor();
	}

	///// 
	///// M E N U 
	/////

	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent event ) {
				getModel().delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);
		return popupMenu;
	}
}