/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.graph.slot;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.border.LineBorder;

import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.Slot;

public abstract class SlotView extends JLabel implements MouseListener, MouseMotionListener {
	private static final int LINE_WIDTH = 1;
	private static final transient LineBorder ACTIVE_BORDER = (LineBorder)BorderFactory.createLineBorder(Color.RED, LINE_WIDTH);
	private static final transient LineBorder NORMAL_BORDER = (LineBorder)BorderFactory.createLineBorder(new Color(0, 0, 0, 0), LINE_WIDTH);

	private Slot _slot;

	private static Slot _targetSlot;
	private static Slot _sourceSlot;

	private boolean _selected;

	public SlotView() {
		setNormalBorder();
		setFocusable(false);
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	public final void setModel( Slot slot ) {
		_slot = slot;
	}

	protected final Slot getModel() {
		return _slot;
	}

	public final boolean isSelected() {
		return _selected;
	}

	public final void setSelected( boolean isSelected ) {
		if (getModel().isConnected()) {
			highlightSlot(isSelected);
			for (Slot slot : getModel().getConnectedSlots()) {
				slot.getView().highlightSlot(isSelected);
			}
		}
	}

	public final void highlightSlot( boolean isSelected ) {
		_selected = isSelected;
		if (isSelected) {
			setActiveBorder();
		} else {
			setNormalBorder();
		}
	}

	public final void setActiveBorder( ) {
		setBorder(ACTIVE_BORDER);
	}

	public final void setNormalBorder( ) {
		setBorder(NORMAL_BORDER);
	}

	public abstract void setConnectedIcon( );

	public abstract void setDisconnectedIcon( );

	public abstract Point getAnchor();

	@Override
	public final void print( Graphics gc ) {
		gc.fillRect(0, 0, getWidth(), getHeight());
	}

	protected abstract Color getLinkColor( );

	public final void drawLink( Graphics gc ) {
		if (getModel().isConnected()) {
			((InputSlot)getModel()).getLink().draw(gc, getLinkColor(), isSelected());
		}
	}

	public abstract JComponent getValueEditor();

	/////////
	///////// MOUSE
	/////////

	private void startConnection() {
		synchronized (this) {
			_targetSlot = null;
			_sourceSlot = getModel();
		}
	}

	private void dropIntoGraph( ) {
		if (_targetSlot == null) {
			getModel().getGraph().disconnectSlot(getModel());
		} else {
			setNormalBorder();
			getModel().connectWith(_targetSlot);
		}
	}

	@Override
	public final void mouseReleased( MouseEvent event ) {
		setCursor(null);
		getModel().getGraph().getView().stopElasticConnection();
		setSelected(false);
		dropIntoGraph();
		getModel().getGraph().getView().requestFocusInWindow();
	}

	@Override
	public final void mousePressed( MouseEvent event ) {
		startConnection();
		setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
		getModel().getGraph().getView().startElasticConnection(getLocationOnScreen().x + 4, getLocationOnScreen().y + 4);
	}

	@Override
	public final void mouseExited( MouseEvent event ) {
		if (_sourceSlot != null) {
			if (!isSelected()) {
				setNormalBorder();
			}
			_targetSlot = null;
		}
	}

	@Override
	public final void mouseEntered( MouseEvent event ) {
		if (_sourceSlot != null) {
			_targetSlot = getModel();
			// Accepts the drop action for a move action.
			if (_sourceSlot.isAcceptableDropSite(_targetSlot)) {
				setActiveBorder();
			}
		}
	}

	@Override
	public final void mouseClicked( MouseEvent event ) {
		/* empty */
	}

	@Override
	public final void mouseMoved( MouseEvent event ) {
		/* empty */
	}

	@Override
	public final void mouseDragged( MouseEvent event ) {
		getModel().getGraph().getView().drawElasticConnection(event.getLocationOnScreen());
	}
}
