/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.utils;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.Timer;

/**
 * Implementation of fade in and fade out effects inspired from gamification.
 * This is a static class that defines the methods : fadeIn() and fadeOut().
 * Example of use: 
 *		FadingEffects.fadeIn(frame);
 *		JOptionPane.showMessageDialog(frame, "alert", "alert", JOptionPane.ERROR_MESSAGE);
 *		FadingEffects.fadeOut(frame);
 */
public final class FadingEffects {
	private static final int DEFAULT_DELAY = 50;
	private static final int DEFAULT_ITERATIONS = 10;
	private static final float MAX_OPACITY = 0.6f;

	private static Component _glassPane;
	private static int _delay;
	private static float _iterationIncrement;

	private FadingEffects() { }

	public static void fadeIn( final JFrame frame ) {
		fadeIn(frame, DEFAULT_DELAY, DEFAULT_ITERATIONS);
	}

	private static void fadeIn( final JFrame frame, final int delay, final int iterations ) {
		_delay = delay;
		_iterationIncrement = MAX_OPACITY / iterations;
		_glassPane = frame.getGlassPane();
		final FadingMirror glassPane = new FadingMirror();
		frame.setGlassPane(glassPane);
		glassPane.setVisible(true);
		glassPane.beginFade();
	}

	public static void fadeOut( final JFrame frame ) {
		frame.getGlassPane().setVisible(false);
		frame.setGlassPane(_glassPane);
	}

	/**
	 * The panel used as glass pane.
	 */
	private static class FadingMirror extends JPanel implements ActionListener {
		private float _opacity;
		private Timer _fadeTimer;

		public FadingMirror() {
			setBackground(Color.BLACK);
			setOpaque(false);
		}

		public void beginFade() {
			_fadeTimer = new Timer(_delay, this);
			_fadeTimer.setInitialDelay(0);
			_fadeTimer.start();
		}

		@Override
		public void actionPerformed( final ActionEvent event ) {
			_opacity += _iterationIncrement;
			if (_opacity > MAX_OPACITY) {
				_opacity = MAX_OPACITY;
				_fadeTimer.stop();
			}
			repaint();
		}

		@Override
		public void paintComponent( final Graphics graphics ) {
			((Graphics2D) graphics).setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, _opacity));
			graphics.setColor(getBackground());
			graphics.fillRect(0, 0, getWidth(), getHeight());
		}
	}
}
