/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Ariane
 */

package fr.greyc.pantheon.ariane.view.utils;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;

import fr.greyc.pantheon.ariane.model.ImageDecoder;
import fr.greyc.pantheon.ariane.model.UserPreferences;

/**
 * An implementation of image previewer that is used in FileChooser.
 * Allows the visualization in thumbnail image for supported image format,
 * such as GIF or JPEG or PANDORE.
 */
public final class FileImagePreviewer extends JPanel implements PropertyChangeListener {
	private static final int MAXSIZE = 120;
	private static final int PREVIEW_LENGTH = 300;
	private ThumbnailReader _reader;

	public FileImagePreviewer( JFileChooser fileChooser ) {
		super(new BorderLayout());
		setPreferredSize(new Dimension(200, 50));
		setBorder(new BevelBorder(BevelBorder.LOWERED));
		fileChooser.addPropertyChangeListener(this);
	}

	private void displayTextPreview( File selectedFile ) {
		add(loadText(selectedFile), BorderLayout.CENTER);
		revalidate();
		repaint();
	}

	private void displayImagePreview( File selectedFile ) {
		if (_reader != null) {
			_reader.kill();
		}
		_reader = new ThumbnailReader(this, selectedFile);
		_reader.start();
	}

	@Override
	public void propertyChange( PropertyChangeEvent event ) {
		if (event.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
			File selectedFile = (File) event.getNewValue();
			removeAll();
			if (selectedFile != null) {
				if (isText(selectedFile, false)) {
					displayTextPreview(selectedFile);
				} else {
					displayImagePreview(selectedFile);
				}
			}
		}
	}

	private class ThumbnailReader extends Thread {
		private final JComponent _component;
		private final File _selectedFile;
		private Boolean _isKilled = false;

		public ThumbnailReader( JComponent component, File selectedFile ) {
			_component = component;
			_selectedFile = selectedFile;
		}

		public final void kill() {
			_isKilled = true;
		}

		private boolean isKilled() {
			return _isKilled;
		}
		
		@Override
		public final void run() {
			_component.getParent().getParent().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			ImageIcon imageIcon = loadImage(_selectedFile);
			if (isKilled()) {
				_component.getParent().getParent().setCursor(null);
				imageIcon = null;
				return;
			}
			JLabel label = new JLabel(imageIcon);
			if (isKilled()) {
				_component.getParent().getParent().setCursor(null);
				label = null;
				return;
			}
			_component.add(label, BorderLayout.CENTER);
			if (isKilled()) {
				_component.getParent().getParent().setCursor(null);
				return;
			}
			_component.getParent().getParent().setCursor(null);
			revalidate();
			repaint();
		}
	}

	private ImageIcon loadImage( File selectedFile ) {
		if (selectedFile != null) {
			return ImageDecoder.getThumbnail(selectedFile.getPath(), MAXSIZE);
		}
		return null;
	}

	private JTextArea loadText( File selectedFile ) {
		if (selectedFile != null) {
			try {
				FileInputStream fin = new FileInputStream(selectedFile);
				final byte[] data = new byte[PREVIEW_LENGTH];
				int bytesRead = 0;
				for (int i = 0; i < PREVIEW_LENGTH; i++) {
					int character = fin.read();
					if (character == -1) {
						break;
					}
					bytesRead++;
					data[i] = (byte) character;
				}
				String preview = new String(data, 0, bytesRead);
				fin.close();
				JTextArea textArea = new JTextArea(preview);
				textArea.setEditable(false);
				textArea.setBorder(new EmptyBorder(5, 5, 5, 5));
				textArea.setFont(UserPreferences.getFont(8));
				return textArea;
			} catch (IOException ex) {
				return null;
			}
		}
		return null;
	}

	public static boolean isText(final File file, boolean acceptUTF8encoding) {
		if (!file.exists()) {
			return false;
		}
		int bufferSize = 1024;
		boolean isText = true;
		byte[] buffer = new byte[bufferSize];
		try {
			final RandomAccessFile fis = new RandomAccessFile(file, "r");
			try {
				fis.seek(0);
				final int read = fis.read(buffer);
				int lastByteTranslated = 0;
				for (int i = 0; i < read && isText; i++) {
					final byte character = buffer[i];
					int ub = character & 0xff;  // unsigned
					int utf8value = lastByteTranslated + ub;
					lastByteTranslated = ub << 8;

					if (!(     ub == 0x09 /*(tab)*/
							|| ub == 0x0A /*(line feed)*/
							|| ub == 0x0C /*(form feed)*/
							|| ub == 0x0D /*(carriage return)*/
							|| ub >= 0x20 && ub <= 0x7E /* Letters, Numbers and other "normal synbols" */
							|| ub >= 0xA0 && ub <= 0xEE /* Symbols of Latin-1 */
							|| acceptUTF8encoding && utf8value >= 0x2E2E && utf8value <= 0xC3BF)) {
						isText = false;
						break;
					}
				}
			} finally {
				fis.close();
			}

			return isText;
		} catch (IOException e) {
			return false;
		}
	}
}